/*
 * Copyright 2011 Leiden University. All rights reserved.
 */

#include <isa/adg.h>

/* Return a C representation of the variable.
 * The pointer is only valid during the lifetime of the object
 * or until the next time get_expr() is called.
 *
 * The pointer is saved in this->expr so that it can be free
 * when the object is destroyed.
 */
__isl_keep const char *adg_var::get_expr()
{
	isl_printer *p;

	if (!access)
		return NULL;

	free(expr);

	p = isl_printer_to_str(isl_pw_multi_aff_get_ctx(access));
	p = isl_printer_set_output_format(p, ISL_FORMAT_C);
	p = isl_printer_print_pw_multi_aff(p, access);
	expr = isl_printer_get_str(p);
	isl_printer_free(p);

	return expr;
}

/* Return an adg_expr that corresponds to the affine expression "aff".
 * If such an adg_expr already exists in controls, then we return that.
 * Otherwise, we construct a new one with a unique name.
 */
adg_expr *adg::to_control(__isl_take isl_aff *aff)
{
	int i;
	adg_expr *expr;
	isl_pw_aff *pa;

	pa = isl_pw_aff_from_aff(aff);

	for (i = 0; i < controls.size(); ++i)
		if (isl_pw_aff_plain_is_equal(pa, controls[i]->expr))
			break;

	if (i < controls.size()) {
		isl_pw_aff_free(pa);
		expr = controls[i];
	} else {
		char buf[10];

		expr = new adg_expr;
		snprintf(buf, sizeof(buf), "e%zd", controls.size());
		expr->name = isl_id_alloc(isl_pw_aff_get_ctx(pa), buf, NULL);
		expr->expr = pa;
		controls.push_back(expr);
	}

	return new adg_expr(*expr);
}

/* Find and return a pointer to the node that has the give name.
 */
adg_node *adg::node_by_id(__isl_take isl_id *id)
{
	int i;

	for (i = 0; i < nodes.size(); ++i)
		if (nodes[i]->name == id)
			break;
	isl_id_free(id);
	return i < nodes.size() ? nodes[i] : NULL;
}

int adg_var::is_equal(adg_var *var2)
{
	if (!isl_pw_multi_aff_plain_is_equal(access, var2->access))
		return 0;
	if (type != var2->type)
		return 0;
	if (!expr != !var2->expr)
		return 0;
	if (expr && strcmp(expr, var2->expr))
		return 0;

	return 1;
}

int adg_expr::is_equal(adg_expr *expr2)
{
	if (name != expr2->name)
		return 0;
	if (!isl_pw_aff_plain_is_equal(expr, expr2->expr))
		return 0;

	return 1;
}

int adg_domain::is_equal(adg_domain *domain2)
{
	if (!isl_set_is_equal(bounds, domain2->bounds))
		return 0;

	if (filters.size() != domain2->filters.size())
		return 0;
	for (int i = 0; i < filters.size(); ++i)
		if (!filters[i]->is_equal(domain2->filters[i]))
			return 0;

	return 1;
}

int adg_gate::is_equal(adg_gate *gate2)
{
	if (name != gate2->name)
		return 0;
	if (node_name != gate2->node_name)
		return 0;
	if (!in->is_equal(gate2->in))
		return 0;
	if (!out->is_equal(gate2->out))
		return 0;
	if (!domain->is_equal(gate2->domain))
		return 0;

	return 1;
}

int adg_port::is_equal(adg_port *port2)
{
	if (name != port2->name)
		return 0;
	if (node_name != port2->node_name)
		return 0;
	if (edge_name != port2->edge_name)
		return 0;

	if (vars.size() != port2->vars.size())
		return 0;
	for (int i = 0; i < vars.size(); ++i)
		if (!vars[i]->is_equal(port2->vars[i]))
			return 0;

	if (!domain->is_equal(port2->domain))
		return 0;

	return 1;
}

int adg_function::is_equal(adg_function *function2)
{
	if (name != function2->name)
		return 0;
	if (!domain->is_equal(function2->domain))
		return 0;

	if (in.size() != function2->in.size())
		return 0;
	for (int i = 0; i < in.size(); ++i)
		if (!in[i]->is_equal(function2->in[i]))
			return 0;

	if (out.size() != function2->out.size())
		return 0;
	for (int i = 0; i < out.size(); ++i)
		if (!out[i]->is_equal(function2->out[i]))
			return 0;

	if (ctrl.size() != function2->ctrl.size())
		return 0;
	for (int i = 0; i < ctrl.size(); ++i)
		if (!ctrl[i]->is_equal(function2->ctrl[i]))
			return 0;

	return 1;
}

int adg_node::is_equal(adg_node *node2)
{
	if (name != node2->name)
		return 0;

	if (input_ports.size() != node2->input_ports.size())
		return 0;
	for (int i = 0; i < input_ports.size(); ++i)
		if (!input_ports[i]->is_equal(node2->input_ports[i]))
			return 0;

	if (gates.size() != node2->gates.size())
		return 0;
	for (int i = 0; i < gates.size(); ++i)
		if (!gates[i]->is_equal(node2->gates[i]))
			return 0;

	if (output_ports.size() != node2->output_ports.size())
		return 0;
	for (int i = 0; i < output_ports.size(); ++i)
		if (!output_ports[i]->is_equal(node2->output_ports[i]))
			return 0;

	if (expressions.size() != node2->expressions.size())
		return 0;
	for (int i = 0; i < expressions.size(); ++i)
		if (!expressions[i]->is_equal(node2->expressions[i]))
			return 0;

	if (!function->is_equal(node2->function))
		return 0;
	if (!domain->is_equal(node2->domain))
		return 0;
	if (!isl_map_is_equal(schedule, node2->schedule))
		return 0;
	
	return 1;
}

int adg_arg::is_equal(adg_arg *arg2)
{
	if (type != arg2->type)
		return 0;
	if (!var != !arg2->var)
		return 0;
	if (var && !var->is_equal(arg2->var))
		return 0;

	return 1;
}

int adg_edge::is_equal(adg_edge *edge2)
{
	if (name != edge2->name)
		return 0;
	if (from_node_name != edge2->from_node_name)
		return 0;
	if (from_port_name != edge2->from_port_name)
		return 0;
	if (to_node_name != edge2->to_node_name)
		return 0;
	if (to_port_name != edge2->to_port_name)
		return 0;
	if (type != edge2->type)
		return 0;
	if (!isl_multi_aff_plain_is_equal(map, edge2->map))
		return 0;
	if (isl_int_ne(value_size, edge2->value_size))
		return 0;

	return 1;
}

int adg_param::is_equal(adg_param *param2)
{
	if (name != param2->name)
		return 0;
	if (!isl_aff_plain_is_equal(lb, param2->lb))
		return 0;
	if (!isl_aff_plain_is_equal(ub, param2->ub))
		return 0;
	if (!isl_aff_plain_is_equal(val, param2->val))
		return 0;

	return 1;
}

int adg::is_equal(adg *adg2)
{
	if (name != adg2->name)
		return 0;
	if (!isl_set_is_equal(context, adg2->context))
		return 0;
	if (!isl_map_is_equal(iterator_map, adg2->iterator_map))
		return 0;

	if (params.size() != adg2->params.size())
		return 0;
	for (int i = 0; i < params.size(); ++i)
		if (!params[i]->is_equal(adg2->params[i]))
			return 0;

	if (nodes.size() != adg2->nodes.size())
		return 0;
	for (int i = 0; i < nodes.size(); ++i)
		if (!nodes[i]->is_equal(adg2->nodes[i]))
			return 0;

	if (edges.size() != adg2->edges.size())
		return 0;
	for (int i = 0; i < edges.size(); ++i)
		if (!edges[i]->is_equal(adg2->edges[i]))
			return 0;

	if (all_iterators.size() != adg2->all_iterators.size())
		return 0;
	for (int i = 0; i < all_iterators.size(); ++i)
		if (all_iterators[i] != adg2->all_iterators[i])
			return 0;

	if (iterators.size() != adg2->iterators.size())
		return 0;
	for (int i = 0; i < iterators.size(); ++i)
		if (iterators[i] != adg2->iterators[i])
			return 0;

	return 1;
}
