/*
 * Copyright 2011 Leiden University. All rights reserved.
 */

#include <assert.h>
#include "adg_parse.h"
#include "adg2dot_options.h"

/* Add a connection from the element called "id" to the input port
 * "port" provided it reads from the variable "name".
 */
void add_connections_in(FILE *out, __isl_keep isl_id *id, const char *name,
	adg_port *port)
{
	const char *name2;

	for (int j = 0; j < port->domain->filters.size(); ++j) {
		name2 = isl_pw_multi_aff_get_tuple_name(
			port->domain->filters[j]->access, isl_dim_out);
		if (name != name2)
			continue;
		fprintf(out, "\t%s -> %s;\n",
			isl_id_get_name(id), isl_id_get_name(port->name));
		break;
	}
}

/* Add a connection from the element called "id" to the output port
 * "port" provided it reads from the variable "name".
 */
void add_connections_out(FILE *out, __isl_keep isl_id *id, const char *name,
	adg_port *port)
{
	const char *name2;

	for (int i = 0; i < port->vars.size(); ++i) {
		name2 = isl_pw_multi_aff_get_tuple_name(
				port->vars[i]->access, isl_dim_out);
		if (name != name2)
			continue;
		fprintf(out, "\t%s -> %s;\n",
			isl_id_get_name(id), isl_id_get_name(port->name));
		break;
	}
}

/* Add a connection from the element called "id" to "gate" provided
 * it reads from the variable "name".
 */
void add_connections(FILE *out, __isl_keep isl_id *id, const char *name,
	adg_gate *gate)
{
	const char *name2;

	for (int j = 0; j < gate->domain->filters.size(); ++j) {
		name2 = isl_pw_multi_aff_get_tuple_name(
			gate->domain->filters[j]->access, isl_dim_out);
		if (name != name2)
			continue;
		fprintf(out, "\t%s -> %s;\n",
			isl_id_get_name(id), isl_id_get_name(gate->name));
		break;
	}

	name2 = isl_pw_multi_aff_get_tuple_name(gate->in->access, isl_dim_out);
	if (name != name2)
		return;
	fprintf(out, "\t%s -> %s;\n",
		isl_id_get_name(id), isl_id_get_name(gate->name));
}

/* Add connections from the element called "id" to "fn" provided
 * it reads from the variable "name".
 */
void add_connections(FILE *out, __isl_keep isl_id *id, const char *name,
	__isl_keep isl_id *node_id, adg_function *fn)
{
	const char *name2;

	for (int i = 0; i < fn->in.size(); ++i) {
		name2 = isl_pw_multi_aff_get_tuple_name(fn->in[i]->var->access,
							isl_dim_out);
		if (name != name2)
			continue;
		fprintf(out, "\t%s -> %s;\n",
			isl_id_get_name(id), isl_id_get_name(node_id));
		return;
	}
	for (int i = 0; i < fn->domain->filters.size(); ++i) {
		name2 = isl_pw_multi_aff_get_tuple_name(
				fn->domain->filters[i]->access, isl_dim_out);
		if (name != name2)
			continue;
		fprintf(out, "\t%s -> %s;\n",
			isl_id_get_name(id), isl_id_get_name(node_id));
		return;
	}
}

/* Add connections from the element called "id" to elements of "node" that
 * refer to the variable "name".
 */
void add_connections(FILE *out, __isl_keep isl_id *id, const char *name,
	adg_node *node)
{
	for (int k = 0; k < node->input_ports.size(); ++k)
		add_connections_in(out, id, name, node->input_ports[k]);
	for (int j = 0; j < node->gates.size(); ++j)
		add_connections(out, id, name, node->gates[j]);
	add_connections(out, id, name, node->name, node->function);
	for (int k = 0; k < node->output_ports.size(); ++k)
		add_connections_out(out, id, name, node->output_ports[k]);
}

/* Add connections from "port" to elements of "node" that refer to the
 * same variable.
 */
void add_connections(FILE *out, adg_port *port, adg_node *node)
{
	for (int i = 0; i < port->vars.size(); ++i) {
		const char *name;
		name = isl_pw_multi_aff_get_tuple_name(
				port->vars[i]->access, isl_dim_out);
		add_connections(out, port->name, name, node);
	}
}

/* Add connections from "fn" to elements of "node" that refer to the
 * same variable.
 */
void add_connections(FILE *out, adg_function *fn, adg_node *node)
{
	for (int i = 0; i < fn->out.size(); ++i) {
		const char *name;
		name = isl_pw_multi_aff_get_tuple_name(
				fn->out[i]->var->access, isl_dim_out);
		add_connections(out, node->name, name, node);
	}
	for (int i = 0; i < fn->ctrl.size(); ++i) {
		const char *name;
		name = isl_id_get_name(fn->ctrl[i]->name);
		add_connections(out, node->name, name, node);
	}
}

void print(FILE *out, adg_node *node)
{
	fprintf(out, "subgraph cluster_%s {\n", isl_id_get_name(node->name));
	if (node->function)
		fprintf(out, "\t%s [label = \"%s\"];\n",
			isl_id_get_name(node->name),
			isl_id_get_name(node->function->name));
	else
		fprintf(out, "%s;\n", isl_id_get_name(node->name));
	for (int j = 0; j < node->input_ports.size(); ++j) {
		adg_port *port = node->input_ports[j];
		fprintf(out, "\t%s [label = \"%s\"];\n",
			isl_id_get_name(port->name),
			isl_pw_multi_aff_get_tuple_name(
				    port->vars[0]->access, isl_dim_out));
		add_connections(out, port, node);
	}
	for (int j = 0; j < node->gates.size(); ++j) {
		adg_gate *gate = node->gates[j];
		fprintf(out, "\t%s [label = \"%s\"];\n",
			isl_id_get_name(gate->name),
			isl_pw_multi_aff_get_tuple_name(gate->out->access,
							isl_dim_out));
		fprintf(out, "\t%s -> %s;\n",
			isl_id_get_name(gate->name),
			isl_id_get_name(node->name));
	}
	for (int j = 0; j < node->output_ports.size(); ++j) {
		adg_port *port = node->output_ports[j];
		fprintf(out, "\t%s [label = \"%s\"];\n",
			isl_id_get_name(port->name),
			isl_pw_multi_aff_get_tuple_name(
				    port->vars[0]->access, isl_dim_out));
	}
	add_connections(out, node->function, node);
	fprintf(out, "}\n");
}

void print(adg *graph, FILE *out)
{
	fprintf(out, "digraph dummy {\n");
	for (int i = 0; i < graph->nodes.size(); ++i)
		print(out, graph->nodes[i]);
	for (int i = 0; i < graph->edges.size(); ++i) {
		adg_edge *edge = graph->edges[i];
		fprintf(out, "%s -> %s [label = \"%s\"];\n",
			isl_id_get_name(edge->from_port_name),
			isl_id_get_name(edge->to_port_name),
			isl_id_get_name(edge->name));
	}
	fprintf(out, "}\n");
}

static void set_output_name(isl_ctx *ctx, struct options *options)
{
	int len;
	const char *suffix = ".dot";

	if (!options->input || !strcmp(options->input, "-"))
		return;
	if (options->output)
		return;

	len = strlen(options->input);
	if (len > 5 && !strcmp(options->input + len - 5, ".yaml"))
		len -= 5;
	options->output = isl_alloc_array(ctx, char, len + strlen(suffix) + 1);
	assert(options->output);
	strncpy(options->output, options->input, len);
	strcpy(options->output + len, suffix);
}

int main(int argc, char *argv[])
{
	isl_ctx *ctx;
	struct options *options = options_new_with_defaults();
	adg *graph;
	FILE *in = stdin, *out = stdout;

	ctx = isl_ctx_alloc();
	argc = options_parse(options, argc, argv, ISL_ARG_ALL);

	set_output_name(ctx, options);

	if (options->input && strcmp(options->input, "-")) {
		in = fopen(options->input, "r");
		assert(in);
	}
	if (options->output && strcmp(options->output, "-")) {
		out = fopen(options->output, "w");
		assert(out);
	}

	graph = adg_parse(ctx, in);
	assert(graph);

	print(graph, out);

	delete graph;

	isl_ctx_free(ctx);

	return 0;
}
