/*
 * Copyright 2011 Leiden University. All rights reserved.
 */

#include "adg_parse.h"

static int extract_isl_int(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node, isl_int &v)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return -1);

	isl_int_read(v, (char *) node->data.scalar.value);

	return 0;
}

static __isl_give isl_id *extract_id(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return NULL);

	return isl_id_alloc(ctx, (char *) node->data.scalar.value, NULL);
}

static __isl_give isl_set *extract_set(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return NULL);

	return isl_set_read_from_str(ctx, (char *) node->data.scalar.value);
}

static __isl_give isl_map *extract_map(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return NULL);

	return isl_map_read_from_str(ctx, (char *) node->data.scalar.value);
}

static __isl_give isl_aff *extract_aff(isl_ctx *ctx,
	yaml_document_t *document, yaml_node_t *node)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return NULL);

	return isl_aff_read_from_str(ctx, (char *) node->data.scalar.value);
}

static __isl_give isl_pw_aff *extract_pw_aff(isl_ctx *ctx,
	yaml_document_t *document, yaml_node_t *node)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return NULL);

	return isl_pw_aff_read_from_str(ctx, (char *) node->data.scalar.value);
}

static __isl_give isl_multi_aff *extract_multi_aff(isl_ctx *ctx,
	yaml_document_t *document, yaml_node_t *node)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return NULL);

	return isl_multi_aff_read_from_str(ctx,
					    (char *) node->data.scalar.value);
}

static __isl_give isl_pw_multi_aff *extract_pw_multi_aff(isl_ctx *ctx,
	yaml_document_t *document, yaml_node_t *node)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return NULL);

	return isl_pw_multi_aff_read_from_str(ctx,
					    (char *) node->data.scalar.value);
}

static int extract_id_list(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node, std::vector<isl_id *> &list)
{
	yaml_node_item_t *item;

	if (node->type != YAML_SEQUENCE_NODE)
		isl_die(ctx, isl_error_invalid, "expecting sequence",
			return -1);

	for (item = node->data.sequence.items.start;
	     item < node->data.sequence.items.top; ++item) {
		yaml_node_t *n;
		isl_id *id;

		n = yaml_document_get_node(document, *item);
		id = extract_id(ctx, document, n);
		if (!id)
			return -1;
		list.push_back(id);
	}

	return 0;
}

static adg_var *extract_var(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg_var *var;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	var = new adg_var;
	if (!var)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "access"))
			var->access = extract_pw_multi_aff(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "type"))
			var->type = extract_id(ctx, document, value);
	}

	return var;
error:
	if (var)
		delete var;
	return NULL;
}

static adg_expr *extract_expr(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg_expr *expr;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	expr = new adg_expr;
	if (!expr)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "name"))
			expr->name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "expr"))
			expr->expr = extract_pw_aff(ctx, document, value);
	}

	return expr;
error:
	if (expr)
		delete expr;
	return NULL;
}

static int extract_var_list(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node, std::vector<adg_var *> &list)
{
	yaml_node_item_t *item;

	if (node->type != YAML_SEQUENCE_NODE)
		isl_die(ctx, isl_error_invalid, "expecting sequence",
			return -1);

	for (item = node->data.sequence.items.start;
	     item < node->data.sequence.items.top; ++item) {
		yaml_node_t *n;
		adg_var *var;

		n = yaml_document_get_node(document, *item);
		var = extract_var(ctx, document, n);
		if (!var)
			return -1;
		list.push_back(var);
	}

	return 0;
}

static enum adg_arg_type extract_arg_type(isl_ctx *ctx,
	yaml_document_t *document, yaml_node_t *node)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return adg_arg_unknown);

	if (!strcmp((char *) node->data.scalar.value, "value"))
		return adg_arg_value;
	if (!strcmp((char *) node->data.scalar.value, "reference"))
		return adg_arg_reference;
	if (!strcmp((char *) node->data.scalar.value, "return"))
		return adg_arg_return;

	return adg_arg_unknown;
}

static adg_arg *extract_arg(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg_arg *arg;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	arg = new adg_arg;
	if (!arg)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "var"))
			arg->var = extract_var(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "type"))
			arg->type = extract_arg_type(ctx, document, value);
	}

	return arg;
error:
	if (arg)
		delete arg;
	return NULL;
}

static int extract_arg_list(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node, std::vector<adg_arg *> &list)
{
	yaml_node_item_t *item;

	if (node->type != YAML_SEQUENCE_NODE)
		isl_die(ctx, isl_error_invalid, "expecting sequence",
			return -1);

	for (item = node->data.sequence.items.start;
	     item < node->data.sequence.items.top; ++item) {
		yaml_node_t *n;
		adg_arg *arg;

		n = yaml_document_get_node(document, *item);
		arg = extract_arg(ctx, document, n);
		if (!arg)
			return -1;
		list.push_back(arg);
	}

	return 0;
}

static int extract_expr_list(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node, std::vector<adg_expr *> &list)
{
	yaml_node_item_t *item;

	if (node->type != YAML_SEQUENCE_NODE)
		isl_die(ctx, isl_error_invalid, "expecting sequence",
			return -1);

	for (item = node->data.sequence.items.start;
	     item < node->data.sequence.items.top; ++item) {
		yaml_node_t *n;
		adg_expr *expr;

		n = yaml_document_get_node(document, *item);
		expr = extract_expr(ctx, document, n);
		if (!expr)
			return -1;
		list.push_back(expr);
	}

	return 0;
}

static adg_domain *extract_domain(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg_domain *domain;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	domain = new adg_domain;
	if (!domain)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "bounds"))
			domain->bounds = extract_set(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "controls") &&
		    extract_expr_list(ctx, document, value, domain->controls) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "filters") &&
		    extract_var_list(ctx, document, value, domain->filters) < 0)
			goto error;
	}

	return domain;
error:
	if (domain)
		delete domain;
	return NULL;
}

static adg_function *extract_function(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg_function *fn;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	fn = new adg_function;
	if (!fn)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "name"))
			fn->name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "in") &&
		    extract_arg_list(ctx, document, value, fn->in) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "out") &&
		    extract_arg_list(ctx, document, value, fn->out) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "ctrl") &&
		    extract_expr_list(ctx, document, value, fn->ctrl) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "domain"))
			fn->domain = extract_domain(ctx, document, value);
	}

	return fn;
error:
	if (fn)
		delete fn;
	return NULL;
}

static adg_port *extract_port(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg_port *port;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	port = new adg_port;
	if (!port)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "name"))
			port->name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "node"))
			port->node_name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "edge"))
			port->edge_name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "vars") &&
		    extract_var_list(ctx, document, value, port->vars) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "domain"))
			port->domain = extract_domain(ctx, document, value);
	}

	return port;
error:
	if (port)
		delete port;
	return NULL;
}

static int extract_port_list(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node, std::vector<adg_port *> &list)
{
	yaml_node_item_t *item;

	if (node->type != YAML_SEQUENCE_NODE)
		isl_die(ctx, isl_error_invalid, "expecting sequence",
			return -1);

	for (item = node->data.sequence.items.start;
	     item < node->data.sequence.items.top; ++item) {
		yaml_node_t *n;
		adg_port *port;

		n = yaml_document_get_node(document, *item);
		port = extract_port(ctx, document, n);
		if (!port)
			return -1;
		list.push_back(port);
	}

	return 0;
}

static adg_gate *extract_gate(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg_gate *gate;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	gate = new adg_gate;
	if (!gate)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "name"))
			gate->name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "node"))
			gate->node_name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "in"))
			gate->in = extract_var(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "out"))
			gate->out = extract_var(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "domain"))
			gate->domain = extract_domain(ctx, document, value);
	}

	return gate;
error:
	if (gate)
		delete gate;
	return NULL;
}

static int extract_gate_list(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node, std::vector<adg_gate *> &list)
{
	yaml_node_item_t *item;

	if (node->type != YAML_SEQUENCE_NODE)
		isl_die(ctx, isl_error_invalid, "expecting sequence",
			return -1);

	for (item = node->data.sequence.items.start;
	     item < node->data.sequence.items.top; ++item) {
		yaml_node_t *n;
		adg_gate *gate;

		n = yaml_document_get_node(document, *item);
		gate = extract_gate(ctx, document, n);
		if (!gate)
			return -1;
		list.push_back(gate);
	}

	return 0;
}

static adg_node *extract_node(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg_node *anode;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	anode = new adg_node;
	if (!anode)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "name"))
			anode->name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "function"))
			anode->function = extract_function(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "domain"))
			anode->domain = extract_domain(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "schedule"))
			anode->schedule = extract_map(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "input_ports") &&
		    extract_port_list(ctx, document, value,
						anode->input_ports) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "output_ports") &&
		    extract_port_list(ctx, document, value,
						anode->output_ports) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "gates") &&
		    extract_gate_list(ctx, document, value, anode->gates) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "expressions") &&
		    extract_expr_list(ctx, document, value,
					anode->expressions) < 0)
			goto error;
	}

	return anode;
error:
	if (anode)
		delete anode;
	return NULL;
}

static int extract_node_list(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node, std::vector<adg_node *> &list)
{
	yaml_node_item_t *item;

	if (node->type != YAML_SEQUENCE_NODE)
		isl_die(ctx, isl_error_invalid, "expecting sequence",
			return -1);

	for (item = node->data.sequence.items.start;
	     item < node->data.sequence.items.top; ++item) {
		yaml_node_t *n;
		adg_node *node;

		n = yaml_document_get_node(document, *item);
		node = extract_node(ctx, document, n);
		if (!node)
			return -1;
		list.push_back(node);
	}

	return 0;
}

static enum adg_edge_type extract_type(isl_ctx *ctx,
	yaml_document_t *document, yaml_node_t *node)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return adg_edge_error);

	if (!strcmp((char *) node->data.scalar.value, "fifo"))
		return adg_edge_fifo;
	if (!strcmp((char *) node->data.scalar.value, "sticky fifo"))
		return adg_edge_sticky_fifo;
	if (!strcmp((char *) node->data.scalar.value, "shift register"))
		return adg_edge_shift_register;
	if (!strcmp((char *) node->data.scalar.value, "broadcast"))
		return adg_edge_broadcast;
	if (!strcmp((char *) node->data.scalar.value, "generic"))
		return adg_edge_generic;

	return adg_edge_error;
}

static adg_edge *extract_edge(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg_edge *edge;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	edge = new adg_edge;
	if (!edge)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "name"))
			edge->name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "map"))
			edge->map = extract_multi_aff(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "from_node"))
			edge->from_node_name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "from_port"))
			edge->from_port_name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "to_node"))
			edge->to_node_name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "to_port"))
			edge->to_port_name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "value_size") &&
		    extract_isl_int(ctx, document, value, edge->value_size) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "type"))
			edge->type = extract_type(ctx, document, value);
	}

	return edge;
error:
	if (edge)
		delete edge;
	return NULL;
}

static int extract_edge_list(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node, std::vector<adg_edge *> &list)
{
	yaml_node_item_t *item;

	if (node->type != YAML_SEQUENCE_NODE)
		isl_die(ctx, isl_error_invalid, "expecting sequence",
			return -1);

	for (item = node->data.sequence.items.start;
	     item < node->data.sequence.items.top; ++item) {
		yaml_node_t *n;
		adg_edge *edge;

		n = yaml_document_get_node(document, *item);
		edge = extract_edge(ctx, document, n);
		if (!edge)
			return -1;
		list.push_back(edge);
	}

	return 0;
}

static adg_param *extract_param(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg_param *param;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	param = new adg_param;
	if (!param)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "name"))
			param->name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "lb"))
			param->lb = extract_aff(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "ub"))
			param->ub = extract_aff(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "val"))
			param->val = extract_aff(ctx, document, value);
	}

	return param;
error:
	if (param)
		delete param;
	return NULL;
}

static int extract_param_list(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node, std::vector<adg_param *> &list)
{
	yaml_node_item_t *item;

	if (node->type != YAML_SEQUENCE_NODE)
		isl_die(ctx, isl_error_invalid, "expecting sequence",
			return -1);

	for (item = node->data.sequence.items.start;
	     item < node->data.sequence.items.top; ++item) {
		yaml_node_t *n;
		adg_param *param;

		n = yaml_document_get_node(document, *item);
		param = extract_param(ctx, document, n);
		if (!param)
			return -1;
		list.push_back(param);
	}

	return 0;
}

static adg *extract_adg(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	adg *adg;
	yaml_node_pair_t *pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	adg = new ::adg;
	if (!adg)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				goto error);
		if (!strcmp((char *) key->data.scalar.value, "name"))
			adg->name = extract_id(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "context"))
			adg->context = extract_set(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "iterator_map"))
			adg->iterator_map = extract_map(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value, "iterators") &&
		    extract_id_list(ctx, document, value, adg->iterators) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "all_iterators") &&
		    extract_id_list(ctx, document, value, adg->all_iterators) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "nodes") &&
		    extract_node_list(ctx, document, value, adg->nodes) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "edges") &&
		    extract_edge_list(ctx, document, value, adg->edges) < 0)
			goto error;
		if (!strcmp((char *) key->data.scalar.value, "params") &&
		    extract_param_list(ctx, document, value, adg->params) < 0)
			goto error;
	}

	return adg;
error:
	if (adg)
		delete adg;
	return NULL;
}

/* Extract an adg from the YAML description in "in".
 */
adg *adg_parse(isl_ctx *ctx, FILE *in)
{
	adg *adg = NULL;
	yaml_parser_t parser;
	yaml_node_t *root;
	yaml_document_t document = { 0 };

	yaml_parser_initialize(&parser);

	yaml_parser_set_input_file(&parser, in);

	if (!yaml_parser_load(&parser, &document))
		goto error;

	root = yaml_document_get_root_node(&document);

	adg = extract_adg(ctx, &document, root);

	yaml_document_delete(&document);

	yaml_parser_delete(&parser);

	return adg;
error:
	yaml_parser_delete(&parser);
	return NULL;
}
