/*
 * Copyright 2011 Leiden University. All rights reserved.
 */

#include <isa/adg.h>

static int init(yaml_emitter_t *emitter, FILE *out)
{
	yaml_event_t event;

	yaml_emitter_initialize(emitter);

	yaml_emitter_set_output_file(emitter, out);

	yaml_stream_start_event_initialize(&event, YAML_UTF8_ENCODING);
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (!yaml_document_start_event_initialize(&event, NULL, NULL, NULL, 1))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;
	return 0;
}

static int fini(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_document_end_event_initialize(&event, 1))
		goto error;
	if (!yaml_emitter_emit(emitter, &event))
		goto error;

	yaml_stream_end_event_initialize(&event);
	if (!yaml_emitter_emit(emitter, &event))
		goto error;

	yaml_emitter_delete(emitter);
	return 0;
error:
	yaml_emitter_delete(emitter);
	return -1;
}

static int emit_string(yaml_emitter_t *emitter, const char *str)
{
	yaml_event_t event;

	if (!yaml_scalar_event_initialize(&event, NULL, NULL,
				    (yaml_char_t *) str, strlen(str),
				    1, 1, YAML_PLAIN_SCALAR_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

static int emit_int(yaml_emitter_t *emitter, int i)
{
	char buffer[40];

	snprintf(buffer, sizeof(buffer), "%d", i);
	return emit_string(emitter, buffer);
}

static int emit_named_int(yaml_emitter_t *emitter, const char *name, int i)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (emit_int(emitter, i) < 0)
		return -1;
	return 0;
}

static int emit_isl_int(yaml_emitter_t *emitter, isl_int i)
{
	char *s;
	isl_int_print_gmp_free_t gmp_free;
	int r = 0;

	s = isl_int_get_str(i);
	if (emit_string(emitter, s) < 0)
		r = -1;

	mp_get_memory_functions(NULL, NULL, &gmp_free);
	(*gmp_free)(s, strlen(s) + 1);

	return r;
}

static int emit_named_isl_int(yaml_emitter_t *emitter, const char *name,
	isl_int i)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (emit_isl_int(emitter, i) < 0)
		return -1;
	return 0;
}

static int emit_named_id(yaml_emitter_t *emitter, const char *name,
	__isl_keep isl_id *id)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (emit_string(emitter, isl_id_get_name(id)) < 0)
		return -1;
	return 0;
}

static int emit_set(yaml_emitter_t *emitter, __isl_keep isl_set *set)
{
	isl_ctx *ctx = isl_set_get_ctx(set);
	isl_printer *p;
	char *str;
	int r;

	p = isl_printer_to_str(ctx);
	p = isl_printer_print_set(p, set);
	str = isl_printer_get_str(p);
	isl_printer_free(p);
	r = emit_string(emitter, str);
	free(str);
	return r;
}

static int emit_map(yaml_emitter_t *emitter, __isl_keep isl_map *map)
{
	isl_ctx *ctx = isl_map_get_ctx(map);
	isl_printer *p;
	char *str;
	int r;

	p = isl_printer_to_str(ctx);
	p = isl_printer_print_map(p, map);
	str = isl_printer_get_str(p);
	isl_printer_free(p);
	r = emit_string(emitter, str);
	free(str);
	return r;
}

static int emit_named_set(yaml_emitter_t *emitter, const char *name,
	__isl_keep isl_set *set)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (emit_set(emitter, set) < 0)
		return -1;
	return 0;
}

static int emit_named_map(yaml_emitter_t *emitter, const char *name,
	__isl_keep isl_map *map)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (emit_map(emitter, map) < 0)
		return -1;
	return 0;
}

static int emit_aff(yaml_emitter_t *emitter, __isl_keep isl_aff *aff)
{
	isl_ctx *ctx = isl_aff_get_ctx(aff);
	isl_printer *p;
	char *str;
	int r;

	p = isl_printer_to_str(ctx);
	p = isl_printer_print_aff(p, aff);
	str = isl_printer_get_str(p);
	isl_printer_free(p);
	r = emit_string(emitter, str);
	free(str);
	return r;
}

static int emit_named_aff(yaml_emitter_t *emitter, const char *name,
	__isl_keep isl_aff *aff)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (emit_aff(emitter, aff) < 0)
		return -1;
	return 0;
}

static int emit_pw_aff(yaml_emitter_t *emitter, __isl_keep isl_pw_aff *pa)
{
	isl_ctx *ctx = isl_pw_aff_get_ctx(pa);
	isl_printer *p;
	char *str;
	int r;

	p = isl_printer_to_str(ctx);
	p = isl_printer_print_pw_aff(p, pa);
	str = isl_printer_get_str(p);
	isl_printer_free(p);
	r = emit_string(emitter, str);
	free(str);
	return r;
}

static int emit_named_pw_aff(yaml_emitter_t *emitter, const char *name,
	__isl_keep isl_pw_aff *pwaff)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (emit_pw_aff(emitter, pwaff) < 0)
		return -1;
	return 0;
}

static int emit_multi_aff(yaml_emitter_t *emitter,
	__isl_keep isl_multi_aff *maff)
{
	isl_ctx *ctx = isl_multi_aff_get_ctx(maff);
	isl_printer *p;
	char *str;
	int r;

	p = isl_printer_to_str(ctx);
	p = isl_printer_print_multi_aff(p, maff);
	str = isl_printer_get_str(p);
	isl_printer_free(p);
	r = emit_string(emitter, str);
	free(str);
	return r;
}

static int emit_named_multi_aff(yaml_emitter_t *emitter, const char *name,
	__isl_keep isl_multi_aff *maff)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (emit_multi_aff(emitter, maff) < 0)
		return -1;
	return 0;
}

static int emit_pw_multi_aff(yaml_emitter_t *emitter,
	__isl_keep isl_pw_multi_aff *maff)
{
	isl_ctx *ctx = isl_pw_multi_aff_get_ctx(maff);
	isl_printer *p;
	char *str;
	int r;

	p = isl_printer_to_str(ctx);
	p = isl_printer_print_pw_multi_aff(p, maff);
	str = isl_printer_get_str(p);
	isl_printer_free(p);
	r = emit_string(emitter, str);
	free(str);
	return r;
}

static int emit_named_pw_multi_aff(yaml_emitter_t *emitter, const char *name,
	__isl_keep isl_pw_multi_aff *maff)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (emit_pw_multi_aff(emitter, maff) < 0)
		return -1;
	return 0;
}

int adg_var::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_FLOW_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (access && emit_named_pw_multi_aff(emitter, "access", access))
		return -1;
	if (type && emit_named_id(emitter, "type", type))
		return -1;

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

static int emit_var_list(yaml_emitter_t *emitter,
	const char *name, std::vector<adg_var *> &list)
{
	yaml_event_t event;

	if (list.size() == 0)
		return 0;

	if (emit_string(emitter, name) < 0)
		return -1;

	if (!yaml_sequence_start_event_initialize(&event, NULL, NULL, 1,
					    YAML_BLOCK_SEQUENCE_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	for (int i = 0; i < list.size(); ++i)
		if (list[i]->print(emitter) < 0)
			return -1;

	if (!yaml_sequence_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

static int emit_arg_list(yaml_emitter_t *emitter,
	const char *name, std::vector<adg_arg *> &list)
{
	yaml_event_t event;

	if (list.size() == 0)
		return 0;

	if (emit_string(emitter, name) < 0)
		return -1;

	if (!yaml_sequence_start_event_initialize(&event, NULL, NULL, 1,
					    YAML_BLOCK_SEQUENCE_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	for (int i = 0; i < list.size(); ++i)
		if (list[i]->print(emitter) < 0)
			return -1;

	if (!yaml_sequence_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

static int emit_named_var(yaml_emitter_t *emitter, const char *name,
	adg_var *var)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (var->print(emitter) < 0)
		return -1;
	return 0;
}

static int emit_arg_type(yaml_emitter_t *emitter, enum adg_arg_type type)
{
	const char *s_type;

	switch (type) {
	case adg_arg_value:
		s_type = "value";
		break;
	case adg_arg_reference:
		s_type = "reference";
		break;
	case adg_arg_return:
		s_type = "return";
		break;
	case adg_arg_unknown:
		s_type = "unknown";
		break;
	default:
		return -1;
	}

	if (emit_string(emitter, "type") < 0)
		return -1;
	if (emit_string(emitter, s_type) < 0)
		return -1;

	return 0;
}

int adg_arg::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_FLOW_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (var && emit_named_var(emitter, "var", var))
		return -1;
	if (emit_arg_type(emitter, type))
		return -1;

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

int adg_expr::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_FLOW_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (name && emit_named_id(emitter, "name", name))
		return -1;
	if (expr && emit_named_pw_aff(emitter, "expr", expr))
		return -1;

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

static int emit_expr_list(yaml_emitter_t *emitter,
	const char *name, std::vector<adg_expr *> &list)
{
	yaml_event_t event;

	if (list.size() == 0)
		return 0;

	if (emit_string(emitter, name) < 0)
		return -1;

	if (!yaml_sequence_start_event_initialize(&event, NULL, NULL, 1,
					    YAML_BLOCK_SEQUENCE_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	for (int i = 0; i < list.size(); ++i)
		if (list[i]->print(emitter) < 0)
			return -1;

	if (!yaml_sequence_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

int adg_domain::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_ANY_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (bounds && emit_named_set(emitter, "bounds", bounds))
		return -1;
	if (emit_expr_list(emitter, "controls", controls))
		return -1;
	if (emit_var_list(emitter, "filters", filters))
		return -1;

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

static int emit_named_domain(yaml_emitter_t *emitter, const char *name,
	adg_domain *domain)
{
	if (emit_string(emitter, name) < 0)
		return -1;
	if (domain->print(emitter) < 0)
		return -1;
	return 0;
}

static int emit_id_list(yaml_emitter_t *emitter,
	const char *name, std::vector<isl_id *> &list)
{
	yaml_event_t event;

	if (list.size() == 0)
		return 0;

	if (emit_string(emitter, name) < 0)
		return -1;

	if (!yaml_sequence_start_event_initialize(&event, NULL, NULL, 1,
					    YAML_FLOW_SEQUENCE_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	for (int i = 0; i < list.size(); ++i)
		if (emit_string(emitter, isl_id_get_name(list[i])) < 0)
			return -1;

	if (!yaml_sequence_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

int adg_port::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_BLOCK_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (name && emit_named_id(emitter, "name", name))
		return -1;
	if (node_name && emit_named_id(emitter, "node", node_name))
		return -1;
	if (edge_name && emit_named_id(emitter, "edge", edge_name))
		return -1;
	if (emit_var_list(emitter, "vars", vars))
		return -1;
	if (domain && emit_named_domain(emitter, "domain", domain))
		return -1;

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

int adg_gate::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_BLOCK_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (name && emit_named_id(emitter, "name", name))
		return -1;
	if (node_name && emit_named_id(emitter, "node", node_name))
		return -1;
	if (in && emit_named_var(emitter, "in", in))
		return -1;
	if (out && emit_named_var(emitter, "out", out))
		return -1;
	if (domain && emit_named_domain(emitter, "domain", domain))
		return -1;

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

int adg_function::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_BLOCK_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (name && emit_named_id(emitter, "name", name))
		return -1;

	if (emit_arg_list(emitter, "in", in))
		return -1;
	if (emit_arg_list(emitter, "out", out))
		return -1;
	if (emit_expr_list(emitter, "ctrl", ctrl))
		return -1;

	if (domain && emit_named_domain(emitter, "domain", domain))
		return -1;

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

static int emit_port_list(yaml_emitter_t *emitter,
	const char *name, std::vector<adg_port *> &list)
{
	yaml_event_t event;

	if (list.size() == 0)
		return 0;

	if (emit_string(emitter, name) < 0)
		return -1;

	if (!yaml_sequence_start_event_initialize(&event, NULL, NULL, 1,
					    YAML_BLOCK_SEQUENCE_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	for (int i = 0; i < list.size(); ++i)
		if (list[i]->print(emitter) < 0)
			return -1;

	if (!yaml_sequence_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

int adg_node::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_BLOCK_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (name && emit_named_id(emitter, "name", name))
		return -1;

	if (function) {
		if (emit_string(emitter, "function") < 0)
			return -1;
		function->print(emitter);
	}

	if (domain && emit_named_domain(emitter, "domain", domain))
		return -1;
	if (schedule && emit_named_map(emitter, "schedule", schedule))
		return -1;

	if (emit_port_list(emitter, "input_ports", input_ports))
		return -1;
	if (emit_port_list(emitter, "output_ports", output_ports))
		return -1;
	if (emit_expr_list(emitter, "expressions", expressions))
		return -1;

	if (gates.size() > 0) {
		if (emit_string(emitter, "gates") < 0)
			return -1;
		if (!yaml_sequence_start_event_initialize(&event, NULL, NULL, 1,
						    YAML_BLOCK_SEQUENCE_STYLE))
			return -1;
		if (!yaml_emitter_emit(emitter, &event))
			return -1;
		for (int i = 0; i < gates.size(); ++i)
			if (gates[i]->print(emitter))
				return -1;
		if (!yaml_sequence_end_event_initialize(&event))
			return -1;
		if (!yaml_emitter_emit(emitter, &event))
			return -1;
	}

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

static int emit_type(yaml_emitter_t *emitter, enum adg_edge_type type)
{
	const char *s_type;

	switch (type) {
	case adg_edge_fifo:
		s_type = "fifo";
		break;
	case adg_edge_sticky_fifo:
		s_type = "sticky fifo";
		break;
	case adg_edge_shift_register:
		s_type = "shift register";
		break;
	case adg_edge_broadcast:
		s_type = "broadcast";
		break;
	case adg_edge_generic:
		s_type = "generic";
		break;
	default:
		return -1;
	}

	if (emit_string(emitter, "type") < 0)
		return -1;
	if (emit_string(emitter, s_type) < 0)
		return -1;

	return 0;
}

int adg_edge::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_BLOCK_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (name && emit_named_id(emitter, "name", name))
		return -1;
	if (emit_type(emitter, type))
		return -1;
	if (map && emit_named_multi_aff(emitter, "map", map))
		return -1;
	if (from_node_name &&
	    emit_named_id(emitter, "from_node", from_node_name))
		return -1;
	if (from_port_name &&
	    emit_named_id(emitter, "from_port", from_port_name))
		return -1;
	if (to_node_name && emit_named_id(emitter, "to_node", to_node_name))
		return -1;
	if (to_port_name && emit_named_id(emitter, "to_port", to_port_name))
		return -1;
	emit_named_isl_int(emitter, "value_size", value_size);

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

int adg_param::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_FLOW_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (name && emit_named_id(emitter, "name", name))
		return -1;
	if (lb && emit_named_aff(emitter, "lb", lb))
		return -1;
	if (ub && emit_named_aff(emitter, "ub", ub))
		return -1;
	if (val && emit_named_aff(emitter, "val", val))
		return -1;

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

static int emit_param_list(yaml_emitter_t *emitter,
	const char *name, std::vector<adg_param *> &list)
{
	yaml_event_t event;

	if (list.size() == 0)
		return 0;

	if (emit_string(emitter, name) < 0)
		return -1;

	if (!yaml_sequence_start_event_initialize(&event, NULL, NULL, 1,
					    YAML_ANY_SEQUENCE_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	for (int i = 0; i < list.size(); ++i)
		if (list[i]->print(emitter) < 0)
			return -1;

	if (!yaml_sequence_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

int adg::print(yaml_emitter_t *emitter)
{
	yaml_event_t event;

	if (!yaml_mapping_start_event_initialize(&event, NULL, NULL, 1,
						YAML_BLOCK_MAPPING_STYLE))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	if (name && emit_named_id(emitter, "name", name))
		return -1;
	if (context && emit_named_set(emitter, "context", context))
		return -1;
	if (iterator_map && emit_named_map(emitter, "iterator_map", iterator_map))
		return -1;
	if (emit_id_list(emitter, "iterators", iterators))
		return -1;
	if (emit_id_list(emitter, "all_iterators", all_iterators))
		return -1;
	if (emit_param_list(emitter, "params", params))
		return -1;

	if (nodes.size() > 0) {
		if (emit_string(emitter, "nodes") < 0)
			return -1;
		if (!yaml_sequence_start_event_initialize(&event, NULL, NULL, 1,
						    YAML_BLOCK_SEQUENCE_STYLE))
			return -1;
		if (!yaml_emitter_emit(emitter, &event))
			return -1;
		for (int i = 0; i < nodes.size(); ++i)
			if (nodes[i]->print(emitter))
				return -1;
		if (!yaml_sequence_end_event_initialize(&event))
			return -1;
		if (!yaml_emitter_emit(emitter, &event))
			return -1;
	}

	if (edges.size() > 0) {
		if (emit_string(emitter, "edges") < 0)
			return -1;
		if (!yaml_sequence_start_event_initialize(&event, NULL, NULL, 1,
						    YAML_BLOCK_SEQUENCE_STYLE))
			return -1;
		if (!yaml_emitter_emit(emitter, &event))
			return -1;
		for (int i = 0; i < edges.size(); ++i)
			if (edges[i]->print(emitter))
				return -1;
		if (!yaml_sequence_end_event_initialize(&event))
			return -1;
		if (!yaml_emitter_emit(emitter, &event))
			return -1;
	}

	if (!yaml_mapping_end_event_initialize(&event))
		return -1;
	if (!yaml_emitter_emit(emitter, &event))
		return -1;

	return 0;
}

/* Print a YAML serialization of this to "out".
 */
void adg_var::print(FILE *out)
{
	yaml_emitter_t emitter;

	if (init(&emitter, out) < 0) {
		yaml_emitter_delete(&emitter);
		return;
	}
	print(&emitter);
	fini(&emitter);

}

/* Print a YAML serialization of this to "out".
 */
void adg::print(FILE *out)
{
	yaml_emitter_t emitter;

	if (init(&emitter, out) < 0) {
		yaml_emitter_delete(&emitter);
		return;
	}
	print(&emitter);
	fini(&emitter);
}
