/*
 * Copyright 2011 Leiden University. All rights reserved.
 */

#include <assert.h>
#include <sstream>
#include <libxml/xmlwriter.h>
#include <isl/union_map.h>
#include <cloog/isl/cloog.h>
#include "adg_xml.h"
#include "xml_AST.h"

static void writeVar(xmlTextWriterPtr writer, adg_var *var)
{
	int rc;

	assert(var->access);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
					BAD_CAST var->get_expr());
	assert(rc >= 0);

	assert(var->type);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "dataType",
					BAD_CAST isl_id_get_name(var->type));
	assert(rc >= 0);
}

static void writeVar(xmlTextWriterPtr writer, const char *type, adg_var *var)
{
	int rc;

	rc = xmlTextWriterStartElement(writer, BAD_CAST type);
	assert(rc >= 0);

	writeVar(writer, var);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

static void writeArg(xmlTextWriterPtr writer, const char *type, adg_arg *arg)
{
	int rc;

	rc = xmlTextWriterStartElement(writer, BAD_CAST type);
	assert(rc >= 0);

	writeVar(writer, arg->var);

	switch (arg->type) {
	case adg_arg_value:
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "pass",
						 BAD_CAST "value");
		break;
	case adg_arg_return:
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "pass",
						 BAD_CAST "return_value");
		break;
	case adg_arg_reference:
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "pass",
						 BAD_CAST "reference");
		break;
	case adg_arg_unknown:
		assert(0);
		break;
	}
	assert(rc >= 0);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

/* Print "v * name" onto os.
 */
static void print_term(std::ostringstream &os, isl_int v, const char *name)
{
	if (!isl_int_is_one(v)) {
		os << isl_int_get_si(v);
		os << "*";
	}
	os << name;
}

static void print_aff_expr(std::ostringstream &os, __isl_keep isl_aff *aff);

/* Print "v * div(aff)" onto os.
 */
static void print_term(std::ostringstream &os, isl_int v,
	__isl_take isl_aff *aff)
{
	isl_int d;

	isl_int_init(d);

	os << "div(";
	print_aff_expr(os, aff);
	os << ",";
	isl_aff_get_denominator(aff, &d);
	os << isl_int_get_si(d);
	os << ")";
	isl_int_clear(d);
	isl_aff_free(aff);
}

/* Print "aff" onto os.
 */
static void print_aff_expr(std::ostringstream &os, __isl_keep isl_aff *aff)
{
	bool first = true;
	isl_int v;
	enum isl_dim_type types[] = { isl_dim_in, isl_dim_param, isl_dim_div };
	int n_type = sizeof(types)/sizeof(*types);

	isl_int_init(v);

	for (int t = 0; t < n_type; ++t) {
		for (int i = 0; i < isl_aff_dim(aff, types[t]); ++i) {
			isl_aff_get_coefficient(aff, types[t], i, &v);
			if (isl_int_is_zero(v))
				continue;
			if (!first && isl_int_is_pos(v))
				os << " + ";
			if (types[t] == isl_dim_div) {
				isl_aff *div;
				div = isl_aff_get_div(aff, i);
				print_term(os, v, div);
			} else {
				const char *name;
				name = isl_aff_get_dim_name(aff, types[t], i);
				print_term(os, v,  name);
			}
			first = false;
		}
	}

	isl_aff_get_constant(aff, &v);
	if (first || !isl_int_is_zero(v)) {
		if (!first && isl_int_is_pos(v))
			os << " + ";
		os << isl_int_get_si(v);
	}

	isl_int_clear(v);
}

struct print_pw_aff_data {
	std::ostringstream &os;
	std::vector<isl_id *> &iterators;
	bool first;

	print_pw_aff_data(std::ostringstream &os,
		std::vector<isl_id *> &iterators) :
		os(os), iterators(iterators), first(true) {}
};

extern "C" {
	static int print_pw_aff_piece(__isl_take isl_set *set,
		__isl_take isl_aff *aff, void *user);
}

/* Print a single piece of an isl_pw_aff.
 * We currently assume that there is only a single piece
 * and abort if there happens to be more than one.
 */
static int print_pw_aff_piece(__isl_take isl_set *set,
	__isl_take isl_aff *aff, void *user)
{
	struct print_pw_aff_data *data = (struct print_pw_aff_data *) user;

	assert(data->first);

	print_aff_expr(data->os, aff);

	data->first = false;

	isl_set_free(set);
	isl_aff_free(aff);
	return 0;
}

static void print_pw_aff_expr(std::ostringstream &os, __isl_keep isl_pw_aff *pa,
	std::vector<isl_id *> &iterators)
{
	struct print_pw_aff_data data(os, iterators);
	isl_pw_aff_foreach_piece(pa, &print_pw_aff_piece, &data);
}

static void writeExpr(xmlTextWriterPtr writer, const char *type,
	const char *expr_name, adg_expr *expr, adg *graph)
{
	int rc;
	std::ostringstream strm;

	rc = xmlTextWriterStartElement(writer, BAD_CAST type);
	assert(rc >= 0);

	assert(expr->name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
					BAD_CAST isl_id_get_name(expr->name));
	assert(rc >= 0);

	assert(expr->expr);
	print_pw_aff_expr(strm, expr->expr, graph->iterators);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST expr_name,
					 BAD_CAST strm.str().c_str());
	assert(rc >= 0);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

/* Write out the rows of "mat", each prefixed by the integer "ineq".
 * The rows are separated by ";".  If first is false, then we
 * also include a ";" before the first row.
 */
static void writeMatrix(std::ostream &os, __isl_keep isl_mat *mat,
	int ineq, bool first)
{
	int i, j;
	isl_int v;

	isl_int_init(v);
	for (i = 0; i < isl_mat_rows(mat); ++i) {
		if (!first)
			os << "; ";
		first = false;
		os << ineq;
		for (j = 0; j < isl_mat_cols(mat); ++j) {
			isl_mat_get_element(mat, i, j, &v);
			os << ", " << isl_int_get_si(v);
		}
	}
	isl_int_clear(v);
}

/* Find the number of dynamic control variables in "space".
 *
 * The dynamic control variables, if any, are located in the range
 * of a wrapped map, with the (possibly lifted) iteration domain
 * in the domain.  After adding the dynamic control variables, the
 * result may have been lifted as well, so we need dig through
 * any possible lifting of the form
 *
 *	D -> [D -> local[..]]
 *
 * until we arrive at the iteration domain, in which case there
 * are no dynamic control variables, or until we find the control
 * variables.
 */
static int n_dynamic_in_space(__isl_take isl_space *space)
{
	int n;

	if (!isl_space_is_wrapping(space)) {
		isl_space_free(space);
		return 0;
	}
	space = isl_space_unwrap(space);
	if (isl_space_has_tuple_id(space, isl_dim_out))
		return n_dynamic_in_space(isl_space_domain(space));
	n = isl_space_dim(space, isl_dim_out);
	isl_space_free(space);
	return n;
}

/* Find the position of the first dynamic control variable in "space".
 *
 * See n_dynamic_in_space.
 */
static int pos_dynamic_in_space(__isl_take isl_space *space)
{
	int pos;

	if (!isl_space_is_wrapping(space)) {
		isl_space_free(space);
		return 0;
	}
	space = isl_space_unwrap(space);
	if (isl_space_has_tuple_id(space, isl_dim_out))
		return pos_dynamic_in_space(isl_space_domain(space));
	pos = isl_space_dim(space, isl_dim_in);
	isl_space_free(space);
	return pos;
}

/* Write a matrix corresponding to the constraints of "bset".
 * The order of the dimensions in the matrix is a follows.
 *
 *	iterators static_controls dynamic_controls parameters constant
 *
 * Inside "bset", however, in the set dimensions, the dynamic
 * controls may appear in the middle of the static controls.
 *
 * We therefore need to move the dynamic controls.
 * "pos_dynamic" indicates the position of the first dynamic control.
 * "n_dynamic" indicates the number of dynamic controls.
 */
static void writeMatrix(xmlTextWriterPtr writer, const char *name,
	__isl_keep isl_basic_set *bset, int n_dynamic)
{
	bool first = true;
	int rc;
	std::ostringstream strm;
	isl_mat *mat;
	int n_set = isl_basic_set_dim(bset, isl_dim_set);
	int pos_dynamic = pos_dynamic_in_space(isl_basic_set_get_space(bset));

	rc = xmlTextWriterStartElement(writer, BAD_CAST name);
	assert(rc >= 0);

	strm << "[";
	mat = isl_basic_set_equalities_matrix(bset,
		    isl_dim_set, isl_dim_div, isl_dim_param, isl_dim_cst);
	mat = isl_mat_move_cols(mat, n_set - n_dynamic, pos_dynamic, n_dynamic);
	writeMatrix(strm, mat, 0, first);
	first = isl_mat_rows(mat) == 0;
	isl_mat_free(mat);

	mat = isl_basic_set_inequalities_matrix(bset,
		    isl_dim_set, isl_dim_div, isl_dim_param, isl_dim_cst);
	mat = isl_mat_move_cols(mat, n_set - n_dynamic, pos_dynamic, n_dynamic);
	writeMatrix(strm, mat, 1, first);
	isl_mat_free(mat);
	strm << "]";

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "matrix",
					 BAD_CAST strm.str().c_str());
	assert(rc >= 0);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

/* Print the coefficients of the affine expression "aff",
 * separated by commas.
 *
 * Note that the set dimensions of "aff" are given in the order
 *
 *	iterators static_and_dynamic_controls
 *
 * while the coefficients should be printed in the order
 *
 *	iterators static_controls dynamic_controls parameters constant
 */
static void print_aff_coefficients(std::ostringstream &os,
	__isl_keep isl_aff *aff)
{
	bool first = true;
	isl_int v;
	isl_space *space;
	int n_index;
	int pos_dynamic;
	int n_dynamic;

	space = isl_aff_get_domain_space(aff);
	pos_dynamic = pos_dynamic_in_space(isl_space_copy(space));
	n_dynamic = n_dynamic_in_space(space);
	n_index = isl_aff_dim(aff, isl_dim_in) - n_dynamic;

	isl_int_init(v);

	for (int i = 0; i < pos_dynamic; ++i) {
		if (!first)
			os << ", ";
		isl_aff_get_coefficient(aff, isl_dim_in, i, &v);
		os << isl_int_get_si(v);
		first = false;
	}

	for (int i = pos_dynamic; i < n_index; ++i) {
		if (!first)
			os << ", ";
		isl_aff_get_coefficient(aff, isl_dim_in, n_dynamic + i, &v);
		os << isl_int_get_si(v);
		first = false;
	}

	for (int i = 0; i < n_dynamic; ++i) {
		if (!first)
			os << ", ";
		isl_aff_get_coefficient(aff, isl_dim_in, pos_dynamic + i, &v);
		os << isl_int_get_si(v);
		first = false;
	}

	for (int i = 0; i < isl_aff_dim(aff, isl_dim_param); ++i) {
		if (!first)
			os << ", ";
		isl_aff_get_coefficient(aff, isl_dim_param, i, &v);
		os << isl_int_get_si(v);
		first = false;
	}

	if (!first)
		os << ", ";
	isl_aff_get_constant(aff, &v);
	os << isl_int_get_si(v);

	isl_int_clear(v);
}

/* Write out "map" in the form of a matrix, with each row
 * corresponding to an affine expression.
 */
static void writeMatrix(xmlTextWriterPtr writer, const char *name,
	__isl_keep isl_multi_aff *map)
{
	int rc;
	std::ostringstream strm;
	int n;

	rc = xmlTextWriterStartElement(writer, BAD_CAST name);
	assert(rc >= 0);

	strm << "[";
	n = isl_multi_aff_dim(map, isl_dim_out);
	for (int i = 0; i < n; ++i) {
		isl_aff *aff;
		if (i)
			strm << "; ";
		aff = isl_multi_aff_get_aff(map, i);
		print_aff_coefficients(strm, aff);
		isl_aff_free(aff);
	}
	strm << "]";

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "matrix",
					 BAD_CAST strm.str().c_str());
	assert(rc >= 0);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

struct write_bound_data {
	xmlTextWriterPtr writer;
	adg_domain *domain;
	adg *graph;
};

static int write_bound(__isl_take isl_basic_set *bset, void *user)
{
	int rc;
	int nparam;
	std::ostringstream strm;
	write_bound_data *data = (write_bound_data *) user;
	xmlTextWriterPtr writer = data->writer;

	rc = xmlTextWriterStartElement(writer, BAD_CAST "linearbound");
	assert(rc >= 0);

	for (int i = 0; i < data->graph->iterators.size(); ++i) {
		if (i)
			strm << ", ";
		strm << isl_id_get_name(data->graph->iterators[i]);
	}

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "index",
					 BAD_CAST strm.str().c_str());
	assert(rc >= 0);

	strm.str("");
	for (int i = 0; i < data->domain->controls.size(); ++i) {
		if (i)
			strm << ", ";
		strm << isl_id_get_name(data->domain->controls[i]->name);
	}
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "staticControl",
					 BAD_CAST strm.str().c_str());
	assert(rc >= 0);

	strm.str("");
	for (int i = 0; i < data->domain->filters.size(); ++i) {
		if (i)
			strm << ", ";
		strm << data->domain->filters[i]->get_expr();
	}
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "dynamicControl",
					 BAD_CAST strm.str().c_str());
	assert(rc >= 0);

	strm.str("");
	nparam = isl_basic_set_dim(bset, isl_dim_param);
	for (int i = 0; i < nparam; ++i) {
		if (i)
			strm << ", ";
		strm << isl_basic_set_get_dim_name(bset, isl_dim_param, i);
	}
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "parameter",
					 BAD_CAST strm.str().c_str());
	assert(rc >= 0);

	writeMatrix(writer, "constraint", bset, data->domain->filters.size());

	isl_basic_set_free(bset);

	if (data->graph->context &&
	    !isl_set_plain_is_universe(data->graph->context)) {
		assert(isl_set_n_basic_set(data->graph->context) == 1);
		bset = isl_set_simple_hull(isl_set_copy(data->graph->context));
		writeMatrix(writer, "context", bset, 0);
		isl_basic_set_free(bset);
	}

	for (int i = 0; i < data->domain->controls.size(); ++i)
		writeExpr(writer, "control", "exp", data->domain->controls[i],
			    data->graph);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);

	return 0;
}

static void writeDomain(xmlTextWriterPtr writer, adg_domain *domain, adg *graph)
{
	int rc;
	write_bound_data data = { writer, domain, graph };

	rc = xmlTextWriterStartElement(writer, BAD_CAST "domain");
	assert(rc >= 0);

	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type",
						BAD_CAST "LBS");
	assert(rc >= 0);

	rc = isl_set_foreach_basic_set(domain->bounds, &write_bound, &data);
	assert(rc >= 0);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

static void writePort(xmlTextWriterPtr writer, adg_port *port, bool output,
	adg *graph)
{
	int rc;
	const char *type;

	type = output ? "outport" : "inport";

	rc = xmlTextWriterStartElement(writer, BAD_CAST type);
	assert(rc >= 0);

	assert(port->name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
					 BAD_CAST isl_id_get_name(port->name));
	assert(port->node_name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "node",
				     BAD_CAST isl_id_get_name(port->node_name));
	assert(port->edge_name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "edge",
				     BAD_CAST isl_id_get_name(port->edge_name));

	for (int i = 0; i < port->vars.size(); ++i)
		writeVar(writer, "bindvariable", port->vars[i]);

	writeDomain(writer, port->domain, graph);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

static void writeGate(xmlTextWriterPtr writer, adg_gate *gate, adg *graph)
{
	int rc;

	rc = xmlTextWriterStartElement(writer, BAD_CAST "invar");
	assert(rc >= 0);

	assert(gate->name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
					 BAD_CAST isl_id_get_name(gate->name));

	assert(gate->in && gate->in->access);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "realName",
					 BAD_CAST gate->in->get_expr());

	assert(gate->out);
	writeVar(writer, "bindvariable", gate->out);

	writeDomain(writer, gate->domain, graph);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

static void writeFunction(xmlTextWriterPtr writer, adg_function *fn, adg *graph)
{
	int rc;

	rc = xmlTextWriterStartElement(writer, BAD_CAST "function");
	assert(rc >= 0);

	assert(fn->name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
					 BAD_CAST isl_id_get_name(fn->name));

	for (int i = 0; i < fn->in.size(); ++i)
		writeArg(writer, "inargument", fn->in[i]);
	for (int i = 0; i < fn->out.size(); ++i)
		writeArg(writer, "outargument", fn->out[i]);
	for (int i = 0; i < fn->ctrl.size(); ++i)
		writeExpr(writer, "ctrlvar", "iterator", fn->ctrl[i], graph);

	writeDomain(writer, fn->domain, graph);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

/* The gates should be printed after the input ports because
 * they may use data read through an input port.
 */
static void writeNode(xmlTextWriterPtr writer, adg_node *node, adg *graph)
{
	int rc;

	rc = xmlTextWriterStartElement(writer, BAD_CAST "node");
	assert(rc >= 0);

	assert(node->name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
					 BAD_CAST isl_id_get_name(node->name));
	assert(rc >= 0);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "levelUpNode",
					BAD_CAST "");
	assert(rc >= 0);

	for (int i = 0; i < node->input_ports.size(); ++i)
		writePort(writer, node->input_ports[i], false, graph);

	for (int i = 0; i < node->gates.size(); ++i)
		writeGate(writer, node->gates[i], graph);

	for (int i = 0; i < node->output_ports.size(); ++i)
		writePort(writer, node->output_ports[i], true, graph);

	for (int i = 0; i < node->expressions.size(); ++i)
		writeExpr(writer, "expression", "value",
				node->expressions[i], graph);

	writeFunction(writer, node->function, graph);

	writeDomain(writer, node->domain, graph);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

static void writeEdge(xmlTextWriterPtr writer, adg_edge *edge)
{
	int rc;
	std::ostringstream strm;

	rc = xmlTextWriterStartElement(writer, BAD_CAST "edge");
	assert(rc >= 0);

	assert(edge->name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
					 BAD_CAST isl_id_get_name(edge->name));
	assert(rc >= 0);

	assert(edge->from_port_name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "fromPort",
			     BAD_CAST isl_id_get_name(edge->from_port_name));
	assert(rc >= 0);

	assert(edge->from_node_name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "fromNode",
			     BAD_CAST isl_id_get_name(edge->from_node_name));
	assert(rc >= 0);

	assert(edge->to_port_name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "toPort",
			     BAD_CAST isl_id_get_name(edge->to_port_name));
	assert(rc >= 0);

	assert(edge->to_node_name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "toNode",
			     BAD_CAST isl_id_get_name(edge->to_node_name));
	assert(rc >= 0);

	strm << edge->value_size;
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "size",
						 BAD_CAST strm.str().c_str());
	assert(rc >= 0);

	rc = xmlTextWriterStartElement(writer, BAD_CAST "linearization");
	assert(rc >= 0);

	switch (edge->type) {
	case adg_edge_fifo:
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type",
						 BAD_CAST "fifo");
		break;
	case adg_edge_sticky_fifo:
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type",
						 BAD_CAST "sticky_fifo");
		break;
	case adg_edge_shift_register:
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type",
						 BAD_CAST "shift_register");
		break;
	case adg_edge_broadcast:
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type",
						 BAD_CAST "BroadcastInOrder");
		break;
	case adg_edge_generic:
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "type",
						 BAD_CAST "GenericOutOfOrder");
		break;
	case adg_edge_error:
		assert(0);
		break;
	}
	assert(rc >= 0);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);

	writeMatrix(writer, "mapping", edge->map);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

static void writeParam(xmlTextWriterPtr writer, adg_param *param, adg *graph)
{
	int rc;
	std::ostringstream strm;

	rc = xmlTextWriterStartElement(writer, BAD_CAST "parameter");
	assert(rc >= 0);

	assert(param->name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
					 BAD_CAST isl_id_get_name(param->name));
	assert(rc >= 0);

	if (param->lb) {
		strm.str("");
		print_aff_expr(strm, param->lb);
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "lb",
						 BAD_CAST strm.str().c_str());
		assert(rc >= 0);
	}

	if (param->ub) {
		strm.str("");
		print_aff_expr(strm, param->ub);
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "ub",
						 BAD_CAST strm.str().c_str());
		assert(rc >= 0);
	}

	if (param->val) {
		strm.str("");
		print_aff_expr(strm, param->val);
		rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "value",
						 BAD_CAST strm.str().c_str());
		assert(rc >= 0);
	}

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

static void writeADG(xmlTextWriterPtr writer, adg *graph)
{
	int rc;

	rc = xmlTextWriterStartElement(writer, BAD_CAST "adg");
	assert(rc >= 0);

	assert(graph->name);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "name",
					 BAD_CAST isl_id_get_name(graph->name));
	assert(rc >= 0);
	rc = xmlTextWriterWriteAttribute(writer, BAD_CAST "levelUpNode",
					BAD_CAST "");
	assert(rc >= 0);

	for (int i = 0; i < graph->params.size(); ++i)
		writeParam(writer, graph->params[i], graph);

	for (int i = 0; i < graph->nodes.size(); ++i)
		writeNode(writer, graph->nodes[i], graph);

	for (int i = 0; i < graph->edges.size(); ++i)
		writeEdge(writer, graph->edges[i]);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);
}

/* Project out all the lifted variables that used to be
 * existentially quantified.  In other words, undo the lifting.
 * Also project out the dynamic control variables.
 */
static __isl_give isl_set *unwrap(__isl_take isl_set *dom)
{
	if (!isl_set_is_wrapping(dom))
		return dom;
	return unwrap(isl_map_domain(isl_set_unwrap(dom)));
}

/* Construct a schedule by adding an extra pair of dimensions to
 * domain->schedule, fixed to "offset1" and "offset2"
 * and setting the input name to "name".
 */
static __isl_give isl_map *domain_schedule(adg_node *node, adg_domain *domain,
	int offset1, int offset2, __isl_take isl_id *name)
{
	isl_set *dom;
	isl_map *sched;
	int n_out;

	n_out = isl_map_dim(node->schedule, isl_dim_out);

	dom = isl_set_copy(domain->bounds);
	dom = unwrap(dom);

	sched = isl_map_copy(node->schedule);
	sched = isl_map_intersect_domain(sched, dom);
	sched = isl_map_add_dims(sched, isl_dim_out, 2);
	sched = isl_map_fix_si(sched, isl_dim_out, n_out, offset1);
	sched = isl_map_fix_si(sched, isl_dim_out, n_out + 1, offset2);
	sched = isl_map_set_tuple_id(sched, isl_dim_in, name);

	return sched;
}

static __isl_give isl_map *port_schedule(adg_node *node, adg_port *port,
	int offset)
{
	return domain_schedule(node, port->domain, offset, port->arg_pos,
				isl_id_copy(port->name));
}

static __isl_give isl_map *gate_schedule(adg_node *node, adg_gate *gate,
	int offset)
{
	return domain_schedule(node, gate->domain, offset, 0,
				isl_id_copy(gate->name));
}

/* Construct a schedule for all elements in the node.
 * We add an extra dimension to node->schedule to ensure that
 * the different elements are executed in the right order:
 * input ports, gates, function, output ports.
 * The input ports themselves are ordered according to the access nr.
 */
static __isl_give isl_union_map *collect_node_schedule(adg_node *node, int n)
{
	isl_map *sched_f;
	isl_union_map *sched;
	int n_out;

	n_out = isl_map_dim(node->schedule, isl_dim_out);

	sched_f = isl_map_copy(node->schedule);
	sched_f = isl_map_add_dims(sched_f, isl_dim_out, 2);
	sched_f = isl_map_fix_si(sched_f, isl_dim_out, n_out, 4 * n + 2);
	sched_f = isl_map_fix_si(sched_f, isl_dim_out, n_out + 1, 0);

	sched = isl_union_map_from_map(sched_f);

	for (int i = 0; i < node->input_ports.size(); ++i) {
		isl_map *sched_p;
		sched_p = port_schedule(node, node->input_ports[i], 4 * n + 0);
		sched = isl_union_map_add_map(sched, sched_p);
	}

	for (int i = 0; i < node->gates.size(); ++i) {
		isl_map *sched_p;
		sched_p = gate_schedule(node, node->gates[i], 4 * n + 1);
		sched = isl_union_map_add_map(sched, sched_p);
	}

	for (int i = 0; i < node->output_ports.size(); ++i) {
		isl_map *sched_p;
		sched_p = port_schedule(node, node->output_ports[i], 4 * n + 3);
		sched = isl_union_map_add_map(sched, sched_p);
	}

	return sched;
}

/* Construct a schedule for all nodes in the graph.
 */
static isl_union_map *collect_schedule(adg *graph)
{
	isl_space *space = isl_set_get_space(graph->context);
	isl_union_map *sched = isl_union_map_empty(space);

	for (int i = 0; i < graph->nodes.size(); ++i) {
		isl_union_map *sched_i;
		sched_i = collect_node_schedule(graph->nodes[i], i);
		sched = isl_union_map_union(sched, sched_i);
	}

	return sched;
}

/* Construct a schedule, convert it to a clast using CLooG and then
 * write out the AST using writeAST.
 */
static void writeAST(xmlTextWriterPtr writer, adg *graph)
{
	int rc;
	isl_ctx *ctx = isl_set_get_ctx(graph->context);
	isl_union_map *sched;
	CloogState *cloog_state;
	CloogOptions *cloog_options;
	CloogDomain *context;
	CloogUnionDomain *ud;
	CloogInput *input;

	sched = collect_schedule(graph);

	cloog_state = cloog_isl_state_malloc(ctx);
	cloog_options = cloog_options_malloc(cloog_state);
	context = cloog_domain_from_isl_set(isl_set_copy(graph->context));
	ud = cloog_union_domain_from_isl_union_map(sched);
	for (int i = 0; i < graph->all_iterators.size(); ++i) {
		const char *name = isl_id_get_name(graph->all_iterators[i]);
		ud = cloog_union_domain_set_name(ud, CLOOG_SCAT, i, name);
	}
	input = cloog_input_alloc(context, ud);
	writeAST(writer, input, graph->all_iterators.size(), cloog_options);
	cloog_options_free(cloog_options);
	cloog_state_free(cloog_state);
}

/* Write out an XML representation of "graph" to "out".
 * The adg part is obtained by scanning the data structure.
 * The ast part is obtained by running cloog based on the domain schedules.
 */
void adg_print_xml(adg *graph, FILE *out)
{
	int rc;
	xmlTextWriterPtr writer;
	xmlDocPtr doc;

	LIBXML_TEST_VERSION

	writer = xmlNewTextWriterDoc(&doc, 0);
	assert(writer);

	rc = xmlTextWriterStartDocument(writer, NULL, NULL, NULL);
	assert(rc >= 0);

	rc = xmlTextWriterStartDTD(writer, BAD_CAST "sadg",
			BAD_CAST "-//LIACS//DTD ESPAM 1//EN",
			BAD_CAST "http://www.liacs.nl/~cserc/dtd/espam_1.dtd");
	assert(rc >= 0);
	rc = xmlTextWriterEndDTD(writer);
	assert(rc >= 0);

	rc = xmlTextWriterStartElement(writer, BAD_CAST "sadg");
	assert(rc >= 0);

	writeADG(writer, graph);
	writeAST(writer, graph);

	rc = xmlTextWriterEndElement(writer);
	assert(rc >= 0);

	rc = xmlTextWriterEndDocument(writer);
	assert(rc >= 0);

	xmlFreeTextWriter(writer);

	rc = xmlDocFormatDump(out, doc, 1);
	assert(rc >= 0);
}
