#include <assert.h>
#include <stdio.h>
#include <string.h>

#include <sstream>

#include <isl/ctx.h>
#include <isl/aff.h>
#include <isl/ilp.h>
#include <isl/constraint.h>
#include <pet.h>

#include <isa/pdg.h>
#include "c2pdg_options.h"

const char dot_yaml[] = ".yaml";

static __isl_give isl_map *isl_map_equal_dims(__isl_take isl_space *dim,
	enum isl_dim_type type1, int pos1, enum isl_dim_type type2, int pos2)
{
	isl_map *map = isl_map_universe(dim);
	return isl_map_equate(map, type1, pos1, type2, pos2);
}

static __isl_give isl_map *isl_map_opposite_dims(__isl_take isl_space *dim,
	enum isl_dim_type type1, int pos1, enum isl_dim_type type2, int pos2)
{
	isl_map *map = isl_map_universe(dim);
	return isl_map_oppose(map, type1, pos1, type2, pos2);
}

static pdg::call_or_access *convert_expr(pdg::PDG *pdg,
	pdg::statement *stat, struct pet_expr *expr,
	std::map<isl_id *, pdg::array *> &id2array, __isl_keep isl_map *trans);

/* Convert the pet_expr to a pdg::access and return it.
 * The pdg::access is also added to the list of accesses in "stat".
 * The pointer to the pdg::array is taken from "id2array" based
 * on the tuple id on the access relation.
 *
 * The domain of the access relation is transformed using "trans".
 * If the access has any arguments then the domain of the access relation
 * is a wrapped mapping from the iteration space to the space of
 * argument values.  We only need to change the domain of this wrapped
 * mapping, so we extend the input transformation with an identity mapping
 * on the space of argument values.
 */
static pdg::access *extract_access(pdg::PDG *pdg, pdg::statement *stat,
	struct pet_expr *expr, std::map<isl_id *, pdg::array *> &id2array,
	__isl_keep isl_map *trans)
{
	pdg::access *access = new pdg::access;
	isl_id *id = NULL;
	isl_space *dim;
	isl_map *map;
	unsigned n_index;

	for (int i = 0; i < expr->n_arg; ++i) {
		pdg::call_or_access *coa;
		coa = convert_expr(pdg, stat, expr->args[i], id2array, trans);
		access->nested.push_back(coa);
	}

	if (expr->acc.write)
		access->type = pdg::access::write;
	else
		access->type = pdg::access::read;

	map = isl_map_copy(expr->acc.access);
	trans = isl_map_copy(trans);

	dim = isl_space_domain(isl_map_get_space(map));
	if (!isl_space_is_wrapping(dim))
		isl_space_free(dim);
	else {
		isl_map *id;
		dim = isl_space_unwrap(dim);
		dim = isl_space_range(dim);
		dim = isl_space_map_from_set(dim);
		id = isl_map_identity(dim);
		trans = isl_map_product(trans, id);
	}

	map = isl_map_apply_domain(map, trans);
	if (isl_map_has_tuple_id(map, isl_dim_out))
		id = isl_map_get_tuple_id(map, isl_dim_out);
	n_index = isl_map_dim(map, isl_dim_out);
	for (int i = n_index - 1; i >= 0; --i)
		if (!isl_map_involves_dims(map, isl_dim_out, i, 1))
			map = isl_map_project_out(map, isl_dim_out, i, 1);
		else
			break;
	if (id)
		map = isl_map_set_tuple_id(map, isl_dim_out, id);
	access->map = new pdg::IslMap(map);

	if (isl_map_has_tuple_id(expr->acc.access, isl_dim_out)) {
		id = isl_map_get_tuple_id(expr->acc.access, isl_dim_out);
		access->array = id2array[id];
		isl_id_free(id);
	}

	stat->accesses.push_back(access);

	return access;
}

static pdg::call_or_access *convert_expr(pdg::PDG *pdg,
	pdg::statement *stat, struct pet_expr *expr,
	std::map<isl_id *, pdg::array *> &id2array, __isl_keep isl_map *trans)
{
	pdg::call_or_access *coa = new pdg::call_or_access;

	if (expr->type == pet_expr_access) {
		coa->type = pdg::call_or_access::t_access;
		coa->access = extract_access(pdg, stat, expr, id2array, trans);
	} else {
		coa->type = pdg::call_or_access::t_call;
		coa->call = new pdg::function_call;

		if (expr->type == pet_expr_call)
			coa->call->name = new str(expr->name);
		else if (expr->type == pet_expr_double) {
			std::ostringstream strm;
			strm << expr->d;
			coa->call->name = new str(strm.str());
		} else if (expr->type == pet_expr_ternary)
			coa->call->name = new str("#test");
		else 
			coa->call->name = new str(pet_op_str(expr->op));

		for (int i = 0; i < expr->n_arg; ++i) {
			pdg::call_or_access *child;
			child = convert_expr(pdg, stat, expr->args[i], id2array,
						trans);
			coa->call->arguments.push_back(child);
		}
	}

	return coa;
}

static pdg::function_call *extract_top_function(pdg::PDG *pdg,
	pdg::statement *stat, struct pet_expr *expr,
	std::map<isl_id *, pdg::array *> &id2array, __isl_keep isl_map *trans)
{
	pdg::function_call *call;
	pdg::call_or_access *coa = convert_expr(pdg, stat, expr, id2array, trans);

	if (coa->type == pdg::call_or_access::t_access) {
		call = new pdg::function_call;
		call->arguments.push_back(coa);
		call->name = new str(string(""));
	} else {
		call = coa->call;
		delete coa;
	}

	return call;
}

/* For each of the filters ("arguments") of "stmt", check if the user
 * has specified any bounds on the values of the corresponding array
 * elements.  If so, apply those bounds to the filter in the iteration
 * domain "dom".
 * Return the (possibly) restricted iteration domain.
 */
static __isl_give isl_set *apply_filter_bounds(__isl_take isl_set *dom,
	struct pet_stmt *stmt, std::map<isl_id *, pdg::array *> &id2array)
{
	isl_map *map;

	if (!isl_set_is_wrapping(dom))
		return dom;

	map = isl_set_unwrap(dom);

	for (int i = 0; i < stmt->n_arg; ++i) {
		isl_id *id;
		pdg::array *array;
		pet_expr *expr = stmt->args[i];
		isl_set *bound;

		if (expr->type != pet_expr_access)
			continue;

		id = isl_map_get_tuple_id(expr->acc.access, isl_dim_out);
		array = id2array[id];
		isl_id_free(id);

		if (!array->value_bounds)
			continue;
		bound = array->value_bounds->get_isl_set();
		bound = isl_set_insert_dims(bound, isl_dim_set, 0, i);
		bound = isl_set_add_dims(bound, isl_dim_set,
						stmt->n_arg - (i + 1));
		map = isl_map_intersect_range(map, bound);
	}

	return isl_map_wrap(map);
}

/* Create and return a pdg::node correspondingto stmt.
 * Since isa uses the "prefix" field in a pdg::node to represent the
 * original schedule, we try to extract such a prefix from the schedule.
 * If one of the original for loops had a decrement, then the corresponding
 * part of the schedule will be of the form [i] -> [-i].  Since such
 * an inverse cannot be represented by "prefix", we invert the corresponding
 * dimension of the iteration domain instead, taking care to also
 * change the domains of all access relations accordingly.
 *
 * If 'stmt" has any arguments, then stmt->domain is a wrapped map
 * and the transformation on the iteration domain only applies
 * to the domain of that map.  We therefore need to combine that
 * transformation with an identity mapping on the range before
 * applying it to the domain.
 * Any arguments found are also converted into filters on the node
 * domain.
 */
static pdg::node *extract_node(pdg::PDG *pdg, struct pet_stmt *stmt,
	std::map<isl_id *, pdg::array *> &id2array)
{
	pdg::node *node = new pdg::node;
	pdg::statement *stat;
	int nr;
	int n_out;
	isl_int v;
	isl_space *dim;
	isl_set *dom;
	isl_map *trans, *trans_dom;

	isl_int_init(v);

	dim = isl_set_get_space(stmt->domain);
	if (isl_space_is_wrapping(dim))
		dim = isl_space_domain(isl_space_unwrap(dim));
	trans = isl_map_universe(isl_space_map_from_set(dim));

	n_out = isl_map_dim(stmt->schedule, isl_dim_out);
	for (int i = 0, j = 0; i < n_out; ++i) {
		if (isl_map_plain_is_fixed(stmt->schedule, isl_dim_out, i, &v))
			node->prefix.push_back(isl_int_get_si(v));
		else {
			isl_map *test;
			isl_space *dim;
			dim = isl_map_get_space(stmt->schedule);
			test = isl_map_equal_dims(dim, isl_dim_in, j,
							isl_dim_out, i);
			if (isl_map_is_subset(stmt->schedule, test)) {
				trans = isl_map_equate(trans, isl_dim_in, j,
							isl_dim_out, j);
			} else {
				isl_map_free(test);
				dim = isl_map_get_space(stmt->schedule);
				test = isl_map_opposite_dims(dim, isl_dim_in, j,
								isl_dim_out, i);
				assert(isl_map_is_subset(stmt->schedule, test));
				trans = isl_map_oppose(trans, isl_dim_in, j,
							isl_dim_out, j);
			}
			isl_map_free(test);
			j++;
			node->prefix.push_back(-1);
		}
	}

	dom = isl_set_copy(stmt->domain);
	trans_dom = isl_map_copy(trans);
	if (isl_set_is_wrapping(dom)) {
		isl_space *space = isl_set_get_space(dom);
		isl_map *id;
		space = isl_space_range(isl_space_unwrap(space));
		id = isl_map_identity(isl_space_map_from_set(space));
		trans_dom = isl_map_product(trans_dom, id);
	}
	dom = isl_set_apply(dom, trans_dom);
	dom = apply_filter_bounds(dom, stmt, id2array);
	node->source = new pdg::IslSet(dom);
	node->name = new str(isl_map_get_tuple_name(stmt->schedule,
						    isl_dim_in));

	stat = new pdg::statement;
	node->statement = stat;

	stat->line = stmt->line;

	for (int i = 0; i < stmt->n_arg; ++i) {
		pdg::call_or_access *coa;
		coa = convert_expr(pdg, stat, stmt->args[i], id2array, trans);
		node->filters.push_back(coa);
	}

	if (stmt->body->type == pet_expr_binary &&
	    stmt->body->op <= pet_op_assign &&
	    stmt->body->args[0]->type == pet_expr_access) {
		pdg::access *access;
		if (stmt->body->op != pet_op_assign) {
			access = extract_access(pdg, stat,
					stmt->body->args[0], id2array, trans);
			access->type = pdg::access::read;
			stat->accesses.push_back(access);
		}
		stat->top_function = extract_top_function(pdg, stat,
					stmt->body->args[1], id2array, trans);
		access = extract_access(pdg, stat,
					stmt->body->args[0], id2array, trans);
		stat->top_outputs.push_back(access);
	} else if (stmt->body->type == pet_expr_unary &&
		    pet_op_is_inc_dec(stmt->body->op) &&
		    stmt->body->args[0]->type == pet_expr_access) {
		pdg::access *access;
		access = extract_access(pdg, stat,
				stmt->body->args[0], id2array, trans);
		access->type = pdg::access::read;
		stat->top_function = new pdg::function_call;
		stat->top_function->name = new str(pet_op_str(stmt->body->op));
		access = extract_access(pdg, stat,
					stmt->body->args[0], id2array, trans);
		stat->top_outputs.push_back(access);
	} else
		stat->top_function = extract_top_function(pdg, stat,
					stmt->body, id2array, trans);

	nr = 0;
	for (int i = 0; i < stat->accesses.size(); ++i)
		if (stat->accesses[i]->type == pdg::access::read)
			stat->accesses[i]->nr = nr++;
	for (int i = 0; i < stat->accesses.size(); ++i)
		if (stat->accesses[i]->type != pdg::access::read)
			stat->accesses[i]->nr = nr++;

	isl_int_clear(v);

	isl_map_free(trans);

	return node;
}

/* Create and return a pdg::array corresponding to pa and add it
 * to id2array.
 */
static pdg::array *extract_array(pdg::PDG *pdg, struct pet_array *pa,
	std::map<isl_id *, pdg::array *> &id2array)
{
	isl_id *id = isl_set_get_tuple_id(pa->extent);
	pdg::array *array = new pdg::array;
	int dim = isl_set_dim(pa->extent, isl_dim_set);
	isl_int v;

	isl_int_init(v);

	array->name = new str(isl_id_get_name(id));
	array->element_type = new str(pa->element_type);
	for (int j = 0; j < dim; ++j) {
		isl_space *dim;
		isl_aff *obj;
		enum isl_lp_result res;

		dim = isl_set_get_space(pa->extent);
		obj = isl_aff_zero_on_domain(isl_local_space_from_space(dim));
		obj = isl_aff_add_coefficient_si(obj, isl_dim_in, j, 1);
		res = isl_set_max(pa->extent, obj, &v);
		isl_aff_free(obj);
		if (res == isl_lp_ok)
			array->dims.push_back(isl_int_get_si(v) + 1);
		else
			array->dims.push_back(-1);
	}
	if (pa->value_bounds)
		array->value_bounds =
			new pdg::IslSet(isl_set_copy(pa->value_bounds));
	array->uniquely_defined = pa->uniquely_defined;

	id2array[id] = array;

	isl_id_free(id);
	isl_int_clear(v);

	return array;
}

/* Assign a default value to parameter "p", which corresponds
 * to the parameter at position "i" in scop->context.
 *
 * If the parameter appears in scop->context_value and if
 * it has a fixed value, then use that value.
 * Otherwise find the minimual value of "p" in scop->context.
 * It there is no minimal value, then we currently leave p->value unset.
 */
static void set_default_parameter_value(pdg::parameter *p,
	__isl_keep isl_id *id, struct pet_scop *scop, int i)
{
	isl_space *space;
	isl_aff *obj;
	enum isl_lp_result res;
	isl_int v;
	int pos;

	isl_int_init(v);

	pos = isl_set_find_dim_by_id(scop->context_value, isl_dim_param, id);
	if (pos >= 0 &&
	    isl_set_plain_is_fixed(scop->context_value, isl_dim_param, pos, &v)) {
		p->value = new integer(isl_int_get_si(v));
		isl_int_clear(v);
		return;
	}

	space = isl_set_get_space(scop->context);
	obj = isl_aff_zero_on_domain(isl_local_space_from_space(space));
	obj = isl_aff_add_coefficient_si(obj, isl_dim_param, i, 1);
	res = isl_set_min(scop->context, obj, &v);
	isl_aff_free(obj);
	if (res == isl_lp_ok)
		p->value = new integer(isl_int_get_si(v));

	isl_int_clear(v);
}

int main(int argc, char **argv)
{
	isl_ctx *ctx;
	struct options *options = options_new_with_defaults();
	struct pet_scop *scop;
	pdg::PDG *pdg;
	int len;
	char *output_name;
	FILE *output;
	int nparam;
	char *slash;
	std::map<isl_id *, pdg::array *> id2array;

	ctx = isl_ctx_alloc_with_options(&options_args, options);
	if (!ctx) {
		fprintf(stderr, "Unable to allocate ctx\n");
		return -1;
	}

	pdg = new pdg::PDG(ctx);
	pdg->add_history_line("c2pdg", argc, argv);
	pdg->placement = new str(string("original"));

	pet_options_set_autodetect(ctx, 1);
	argc = options_parse(options, argc, argv, ISL_ARG_ALL);

	len = strlen(options->input);
	if (len > 2 && !strcmp(options->input + len - 2, ".c"))
		len -= 2;
	output_name = isl_alloc_array(ctx, char, len + sizeof(dot_yaml));
	assert(output_name);
	memcpy(output_name, options->input, len);
	memcpy(output_name + len, dot_yaml, sizeof(dot_yaml));
	output = fopen(output_name, "w");
	assert(output);

	scop = pet_scop_extract_from_C_source(ctx, options->input,
						options->function);
	scop = pet_scop_align_params(scop);
	assert(scop);

	output_name[len] = '\0';
	slash = strrchr(output_name, '/');
	if (!slash)
		slash = output_name;
	pdg->name = new str(output_name);

	nparam = isl_set_dim(scop->context, isl_dim_param);
	for (int i = 0; i < nparam; ++i) {
		isl_id *id = isl_set_get_dim_id(scop->context, isl_dim_param, i);
		pdg::parameter *p = new pdg::parameter;

		p->name = new str(isl_id_get_name(id));
		set_default_parameter_value(p, id, scop, i);

		pdg->params.push_back(p);
		isl_id_free(id);
	}

	pdg->context = new pdg::IslSet(isl_set_copy(scop->context));

	for (int i = 0; i < scop->n_array; ++i) {
		pdg::array *array;
		array = extract_array(pdg, scop->arrays[i], id2array);
		pdg->arrays.push_back(array);
	}

	for (int i = 0, nodenr = 0; i < scop->n_stmt; ++i) {
		pdg::node *node;
		if (isl_set_is_empty(scop->stmts[i]->domain))
			continue;
		node = extract_node(pdg, scop->stmts[i], id2array);
		node->nr = nodenr++;
		pdg->nodes.push_back(node);
	}

	pdg->Dump(output);
	pdg->free();
	delete pdg;

	pet_scop_free(scop);

	fclose(output);
	free(output_name);

	isl_ctx_free(ctx);
	return 0;
}
