#include <iostream>
#include <cloog/isl/cloog.h>

#include <isa/yaml.h>
#include <isa/pdg.h>
#include "clast_scanner.h"

using namespace std;
using namespace pdg;

struct core_printer : public clast_scanner {
    void print_user(clast_user_stmt *s);
};

void core_printer::print_user(clast_user_stmt *u)
{
    pdg::dependence *dep = (pdg::dependence*) u->statement->usr;
    pdg::node* from = dep->from;
    pdg::node* to = dep->to;
    int from_access = dep->from_access ? dep->from_access->nr
				       : from->statement->accesses.size();
    int to_access = dep->to_access ? dep->to_access->nr
				   : to->statement->accesses.size();

    print_indent();
    if ((u->statement->number % 2) == 0) {
	fprintf(dst, "if (++i_%d_%d_%d_%d > s_%d_%d_%d_%d)\n",
		    from->nr, from_access, to->nr, to_access,
		    from->nr, from_access, to->nr, to_access);
	print_indent();
	fprintf(dst, "    s_%d_%d_%d_%d = i_%d_%d_%d_%d;\n",
		    from->nr, from_access, to->nr, to_access,
		    from->nr, from_access, to->nr, to_access);
    } else
	fprintf(dst, "--i_%d_%d_%d_%d;\n",
		    from->nr, from_access, to->nr, to_access);
}

struct program_printer : public clast_scanner {
    PDG *pdg;

    void print_user(clast_user_stmt *s);
    void scan_program(struct clast_stmt *stmt, CloogOptions *options);
    void print_program(PDG *pdg);
};

void program_printer::print_user(clast_user_stmt *s)
{
    int maxdim = 0;
    char name[MAX_NAME];
    CloogState *state = cloog_state_malloc();
    CloogOptions *options = cloog_options_malloc(state);
    CloogDomain *context, *domain;
    CloogUnionDomain *ud;
    CloogInput *input;
    options->compilable = 1;
    options->l = pdg->dimension;

    int nparam = pdg->params.size();

    context = cloog_domain_from_isl_set(pdg->get_context_isl_set());
    ud = cloog_union_domain_alloc(nparam);

    for (int i = 0; i < pdg->dependences.size(); ++i) {
	CloogDomain *dom;
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type == pdg::dependence::uninitialized)
	    continue;
	pdg::node* from = dep->from;
	pdg::node* to = dep->to;
	int from_access = dep->from_access ? dep->from_access->nr
					   : from->statement->accesses.size();
	int to_access = dep->to_access ? dep->to_access->nr
				       : to->statement->accesses.size();
	if (!dep->size)
	    continue;
	print_indent();
	fprintf(dst, "int i_%d_%d_%d_%d = 0;\n", from->nr, from_access,
						 to->nr, to_access);
	print_indent();
	fprintf(dst, "int s_%d_%d_%d_%d = 0;\n", from->nr, from_access,
						 to->nr, to_access);
	isl_map *map = scatter_dependence(pdg, dep);
	if (isl_map_dim(map, isl_dim_in) > maxdim)
	    maxdim = isl_map_dim(map, isl_dim_in);
	if (isl_map_dim(map, isl_dim_out) > maxdim)
	    maxdim = isl_map_dim(map, isl_dim_out);
	dom = cloog_domain_from_isl_set(isl_map_range(isl_map_copy(map)));
	ud = cloog_union_domain_add_domain(ud, NULL, dom, NULL, dep);
	dom = cloog_domain_from_isl_set(isl_map_domain(map));
	ud = cloog_union_domain_add_domain(ud, NULL, dom, NULL, dep);
    }

    for (int i = 0; i < nparam; ++i) {
	const char *s = strdup(pdg->params[i]->name->s.c_str());
	ud = cloog_union_domain_set_name(ud, CLOOG_PARAM, i, s);
    }
    for (int i = 0; i < maxdim; ++i) {
	sprintf(name, "i%d", i);
	ud = cloog_union_domain_set_name(ud, CLOOG_ITER, i, name);
    }

    input = cloog_input_alloc(context, ud);
    clast_stmt *stmt = cloog_clast_create_from_input(input, options);

    core_printer cp;
    cp.indent = indent;
    cp.dst = dst;
    cp.print_iterators(((struct clast_root *)stmt)->names);
    cp.print(stmt);
    cloog_clast_free(stmt);

    cloog_options_free(options);
    cloog_state_free(state);

    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	if (dep->type == pdg::dependence::uninitialized)
	    continue;
	pdg::node* from = dep->from;
	pdg::node* to = dep->to;
	int from_access = dep->from_access ? dep->from_access->nr
					   : from->statement->accesses.size();
	int to_access = dep->to_access ? dep->to_access->nr
				       : to->statement->accesses.size();
	if (!dep->size)
	    continue;
	print_indent();
	fprintf(dst, "assert(i_%d_%d_%d_%d == 0);\n",
			    from->nr, from_access, to->nr, to_access);
	print_indent();
	fprintf(dst, 
	    "printf(\"s_%d_%d_%d_%d: %%d; %%d\\n\", s_%d_%d_%d_%d, (int)(%s));\n", 
			    from->nr, from_access, to->nr, to_access,
			    from->nr, from_access, to->nr, to_access,
			    dep->size->s->c_str());
	print_indent();
	fprintf(dst, "assert(s_%d_%d_%d_%d <= (int)(%s));\n",
			    from->nr, from_access, to->nr, to_access,
			    dep->size->s->c_str());
    }
}

void program_printer::scan_program(struct clast_stmt *stmt,
	CloogOptions *options)
{
    print_header();
    fprintf(dst, "%s", "#define max(x,y)\t" "((x)>(y)?(x):(y))" "\n");
    fprintf(dst, "\n");
    fprintf(dst, "int main()\n{\n");
    indent += 4;
    print_iterators(((struct clast_root *)stmt)->names);
    print(stmt);
    print_indent();
    fprintf(dst, "return 0;\n");
    fprintf(dst, "}\n");
}


void program_printer::print_program(PDG *pdg)
{
    isl_ctx *ctx = pdg->get_isl_ctx();

    dst = stdout;
    this->pdg = pdg;
    isl_space *dim;
    isl_basic_set *model;
    unsigned nparam;

    isl_set *context = pdg->get_context_isl_set();
    nparam = isl_set_dim(context, isl_dim_param);
    dim = isl_space_set_alloc(ctx, 0, nparam);
    model = isl_basic_set_empty(dim);
    context = isl_set_to_underlying_set(context);
    context = isl_set_from_underlying_set(context, model);

    CloogState *state = cloog_isl_state_malloc(ctx);
    CloogOptions *options = cloog_options_malloc(state);
    options->compilable = 1;

    dim = isl_space_set_alloc(ctx, 0, 0);
    isl_set *universe = isl_set_universe(dim);
    CloogDomain *cloog_context = cloog_domain_from_isl_set(universe);
    CloogDomain *domain;
    CloogUnionDomain *ud;
    CloogInput *input;
    struct clast_stmt *stmt;

    ud = cloog_union_domain_alloc(0);
    domain = cloog_domain_from_isl_set(context);
    ud = cloog_union_domain_add_domain(ud, NULL, domain, NULL, NULL);

    for (int i = 0; i < nparam; ++i) {
	const char *s = strdup(pdg->params[i]->name->s.c_str());
	ud = cloog_union_domain_set_name(ud, CLOOG_ITER, i, s);
    }

    input = cloog_input_alloc(cloog_context, ud);
    stmt = cloog_clast_create_from_input(input, options);

    scan_program(stmt, options);
    cloog_clast_free(stmt);
    cloog_options_free(options);
    cloog_state_free(state);
}

int main(int argc, char * argv[])
{
    PDG *pdg;
    isl_ctx *ctx = isl_ctx_alloc();
    pdg = PDG::Load(stdin, ctx);

    assert(pdg->context);

    program_printer pp;
    pp.print_program(pdg);

    pdg->free();
    delete pdg;
    isl_ctx_free(ctx);

    return 0;
}
