#include "cloog_util.h"

static clast_expr *clast_expr_copy(clast_expr *e)
{
    if (!e)
	return NULL;
    switch (e->type) {
    case clast_expr_name: {
	clast_name* n = (struct clast_name*) e;
	return &new_clast_name(n->name)->expr;
    }
    case clast_expr_term: {
	clast_term* t = (struct clast_term*) e;
	return &new_clast_term(t->val, clast_expr_copy(t->var))->expr;
    }
    case clast_expr_red: {
	clast_reduction *r = (struct clast_reduction*) e;
	clast_reduction *r2 = new_clast_reduction(r->type, r->n);
	for (int i = 0; i < r->n; ++i)
	    r2->elts[i] = clast_expr_copy(r->elts[i]);
	return &r2->expr;
    }
    case clast_expr_bin: {
	clast_binary *b = (struct clast_binary*) e;
	return &new_clast_binary(b->type, clast_expr_copy(b->LHS), b->RHS)->expr;
    }
    default:
	assert(0);
    }
}

static void simple_guards(clast_for *s);
static void simple_guards(clast_guard *s);

void cloog_util_simple_guards(clast_stmt *s)
{
    for ( ; s; s = s->next) {
	if (CLAST_STMT_IS_A(s, stmt_root) ||
	    CLAST_STMT_IS_A(s, stmt_ass) ||
	    CLAST_STMT_IS_A(s, stmt_user))
	    continue;
	if (CLAST_STMT_IS_A(s, stmt_for)) {
	    simple_guards((clast_for *) s);
	} else if (CLAST_STMT_IS_A(s, stmt_guard)) {
	    simple_guards((clast_guard *) s);
	} else {
	    assert(0);
	}
    }
}

void convert_minmax_to_ifs(clast_expr **rp, clast_expr *LHS, int sign,
			   clast_stmt **body)
{
    assert((*rp)->type == clast_expr_red);
    clast_reduction *r = (clast_reduction *)*rp;

    clast_stmt *child = *body;
    while (r->n > 1) {
	r->n--;
	clast_guard *newg = new_clast_guard(1);
	newg->eq[0].LHS = clast_expr_copy(LHS);
	newg->eq[0].RHS = r->elts[r->n];
	newg->eq[0].sign = sign;
	newg->then = child;
	child = &newg->stmt;
    }
    *body = child;
    *rp = r->elts[0];
    r->n--;
    free_clast_reduction(r);
}

void simple_guards(clast_for *f)
{
    if (f->LB->type == clast_expr_red) {
	clast_reduction *r = (clast_reduction *)f->LB;
	if (r->type != clast_red_sum) {
	    cloog_int_t one;
	    cloog_int_init(one);
	    cloog_int_set_si(one, 1);
	    clast_term *LHS = new_clast_term(one,
					&new_clast_name(f->iterator)->expr);
	    convert_minmax_to_ifs(&f->LB, &LHS->expr, 1, &f->body);
	    free_clast_term(LHS);
	    cloog_int_clear(one);
	}
    }
    if (f->UB && f->UB->type == clast_expr_red) {
	clast_reduction *r = (clast_reduction *)f->UB;
	if (r->type != clast_red_sum) {
	    cloog_int_t one;
	    cloog_int_init(one);
	    cloog_int_set_si(one, 1);
	    clast_term *LHS = new_clast_term(one,
					&new_clast_name(f->iterator)->expr);
	    convert_minmax_to_ifs(&f->UB, &LHS->expr, -1, &f->body);
	    free_clast_term(LHS);
	    cloog_int_clear(one);
	}
    }
    cloog_util_simple_guards(f->body);
}

void simple_guards_minmax(clast_guard *g)
{
    if (g->eq[0].RHS->type != clast_expr_red)
	return;

    clast_reduction *r = (clast_reduction *)g->eq[0].RHS;
    if (r->type == clast_red_sum)
	return;

    convert_minmax_to_ifs(&g->eq[0].RHS, g->eq[0].LHS, g->eq[0].sign, &g->then);
}

void simple_guards(clast_guard *g)
{
    clast_stmt *then = g->then;
    clast_stmt *child = g->then;

    while (g->n > 1) {
	g->n--;
	clast_guard *newg = new_clast_guard(1);
	newg->eq[0].LHS = g->eq[g->n].LHS;
	newg->eq[0].RHS = g->eq[g->n].RHS;
	newg->eq[0].sign = g->eq[g->n].sign;
	newg->then = child;
	simple_guards_minmax(newg);
	child = &newg->stmt;
    }
    g->then = child;
    simple_guards_minmax(g);

    cloog_util_simple_guards(then);
}

static int clast_equation_equal(clast_equation *e1, clast_equation *e2)
{
    return e1->sign == e2->sign &&
	   clast_expr_equal(e1->LHS, e2->LHS) &&
	   clast_expr_equal(e1->RHS, e2->RHS);
}

struct guard_info {
    struct clast_equation *eq;
    int			   first;
    int			   last;
};

/* attach tail to the last element of head and return this previously
 * last element of head.
 */
static clast_stmt *append_stmt_list(clast_stmt *head, clast_stmt *tail)
{
    assert(head);
    while (head->next)
	head = head->next;
    head->next = tail;
    return head;
}

/* Hoist info->eq from the info->first element of the next chain
 * of s to the info->last element of this chain.
 * The above subchain is replaced by the guard with info->eq as condition
 * and with the subchain as the "then" element.
 * The first element in the subchain is assumed to have a guard among
 * its nested guards with eq[0] equal to info->eq, which is rotated
 * outermost.
 * The other elements are assumed to have a guard with eq[0] equivalent
 * to info->eq and these are simply removed.
 */
static clast_stmt *hoist(struct guard_info* info, clast_stmt *s)
{
    clast_stmt **parent = &s;
    clast_stmt *sf = s, *s2, *next;
    clast_guard *hoist, *up;

    for (int i = 0; i < info->first; ++i) {
	parent = &sf->next;
	sf = sf->next;
    }

    /* Find guard and rotate out */
    assert(CLAST_STMT_IS_A(sf, stmt_guard));
    up = hoist = (clast_guard *)sf;
    while (&hoist->eq[0] != info->eq) {
	clast_stmt *s2 = hoist->then;
	assert(!s2->next);
	assert(CLAST_STMT_IS_A(s2, stmt_guard));
	up = hoist;
	hoist = (clast_guard *)s2;
    }
    if (up != hoist) {
	*parent = &hoist->stmt;
	up->then = hoist->then;
	hoist->then = sf;
	hoist->stmt.next = sf->next;
	sf->next = NULL;
    }

    /* Move other elements in the chain to the then part of the hoisted if
     * and remove the equivalent guard.
     */
    s2 = &hoist->stmt;
    s2 = append_stmt_list(hoist->then, s2->next);
    for (int i = info->first+1; i <= info->last; ++i) {
	clast_stmt **parent = &s2->next;
	clast_guard *current;
	s2 = s2->next;
	assert(CLAST_STMT_IS_A(s2, stmt_guard));
	current = (clast_guard *)s2;
	while (!clast_equation_equal(&current->eq[0], info->eq)) {
	    parent = &current->then;
	    clast_stmt *s3 = current->then;
	    assert(!s3->next);
	    assert(CLAST_STMT_IS_A(s3, stmt_guard));
	    current = (clast_guard *)s3;
	}
	if (*parent == s2)
	    s2 = append_stmt_list(current->then, s2->next);
	else
	    assert(!current->stmt.next);
	*parent = current->then;
	current->then = NULL;
	free_clast_stmt(&current->stmt);
    }
    hoist->stmt.next = s2->next;
    s2->next = NULL;

    return s;
}

/* Hoist ifs that are common to several adjacent statements.
 * 
 * We first create a list of guards and note where they first
 * and last appear in sequence.  Then we hoist the if with largest
 * sequence and repeat until the largest sequence is of length 1.
 */
static clast_stmt *do_hoist_ifs(clast_stmt *s)
{
    int n = 0, m = 0;
    for (clast_stmt *s2 = s; s2; s2 = s2->next) {
	clast_stmt *g = s2;
	int has_guard = 0;
	while (CLAST_STMT_IS_A(g, stmt_guard)) {
	    has_guard = 1;
	    n++;
	    g = ((clast_guard *) g)->then;
	    if (g->next)
		break;
	}
	if (has_guard)
	    ++m;
    }
    if (m <= 1)
	return s;

    struct guard_info info[n];
    int i = 0, j = 0, max;
    for (clast_stmt *s2 = s; s2; s2 = s2->next) {
	clast_stmt *gs = s2;
	while (CLAST_STMT_IS_A(gs, stmt_guard)) {
	    int k;
	    clast_guard *g = (clast_guard *) gs;
	    assert(g->n == 1);
	    for (k = 0; k < i; ++k)
		if (info[k].last == j-1 && 
			clast_equation_equal(info[k].eq, &g->eq[0]))
		    break;
	    if (k == i) {
		info[k].eq = &g->eq[0];
		info[k].first = j;
		i++;
	    }
	    info[k].last = j;
	    gs = g->then;
	    if (gs->next)
		break;
	}
	++j;
    }
    n = i;
    do {
	max = 0;
	for (i = 1; i < n; ++i)
	    if (info[i].last - info[i].first > 
		    info[max].last - info[max].first)
		max = i;
	if (info[max].last == info[max].first)
	    break;
	int first = info[max].first;
	int last = info[max].last;
	s = hoist(&info[max], s);
	for (int i = 0; i < n; ++i) {
	    if (info[i].last < first)
		continue;
	    if (info[i].first > last) {
		info[i].first -= last-first;
		info[i].last -= last-first;
		continue;
	    }
	    info[i--] = info[--n];
	}
    } while (n > 0);

    return s;
}

clast_stmt *cloog_util_hoist_ifs(clast_stmt *s)
{
    if (s->next)
	s = do_hoist_ifs(s);

    for (clast_stmt *s2 = s; s2; s2 = s2->next) {
	if (CLAST_STMT_IS_A(s2, stmt_root) ||
	    CLAST_STMT_IS_A(s2, stmt_ass) ||
	    CLAST_STMT_IS_A(s2, stmt_user))
	    continue;
	if (CLAST_STMT_IS_A(s2, stmt_for)) {
	    clast_for *f = (clast_for *) s2;
	    f->body = cloog_util_hoist_ifs(f->body);
	} else if (CLAST_STMT_IS_A(s2, stmt_guard)) {
	    clast_guard *g = (clast_guard *) s2;
	    g->then = cloog_util_hoist_ifs(g->then);
	} else {
	    assert(0);
	}
    }
    return s;
}
