#include <string.h>
#include "condition_tree.h"

static condition_tree_type builtin_binary(in_cal *cal)
{
	if (!cal)
		return ctt_invalid;
	if (cal->num_args() != 2)
		return ctt_invalid;
	in_ldc *ldc = dynamic_cast<in_ldc *>(cal->addr_op().instr());
	if (!ldc || !ldc->value().symbol()->is_proc())
		return ctt_invalid;
	proc_sym *proc = dynamic_cast<proc_sym *>(ldc->value().symbol());
	if (!strcmp(proc->name(), "max"))
		return ctt_max;
	if (!strcmp(proc->name(), "min"))
		return ctt_min;
	if (!strcmp(proc->name(), "ceild"))
		return ctt_ceild;
	if (!strcmp(proc->name(), "floord"))
		return ctt_floord;
	return ctt_invalid;
}

condition_tree *construct_condition_tree(operand op, int verbose)
{
	if (op.kind() == OPER_SYM)
		return new condition_tree(op.symbol());

	if (op.kind() != OPER_INSTR)
		return new condition_tree(ctt_invalid);

	return construct_condition_tree_expr(op.instr(), verbose);
}

condition_tree *construct_condition_tree_expr(instruction *ins, int verbose)
{
	switch (ins->opcode()) {
	case io_cpy:
		return construct_condition_tree(ins->src_op(0), verbose);
	case io_lod:
		return new condition_tree(ins->src_op(0));
	case io_ldc: {
		in_ldc *ic = dynamic_cast<in_ldc *>(ins);
		return new condition_tree(ic->value().integer());
	}
	case io_add: {
		in_rrr *ir = dynamic_cast<in_rrr *>(ins);
		return new condition_tree(ctt_add,
			construct_condition_tree(ir->src1_op(), verbose),
			construct_condition_tree(ir->src2_op(), verbose));
	}
	case io_sub: {
		in_rrr *ir = dynamic_cast<in_rrr *>(ins);
		return new condition_tree(ctt_sub,
			construct_condition_tree(ir->src1_op(), verbose),
			construct_condition_tree(ir->src2_op(), verbose));
	}
	case io_mul: {
		in_rrr *ir = dynamic_cast<in_rrr *>(ins);
		operand op1 = ir->src1_op();
		operand op2 = ir->src2_op();
		if (op1.kind() == OPER_INSTR && 
		    op1.instr()->opcode() == io_ldc) {
			in_ldc *ins = dynamic_cast<in_ldc *>(op1.instr());
			return new condition_tree(ctt_mul,
				new condition_tree(ins->value().integer()),
				construct_condition_tree(op2, verbose));
		} else if (op2.kind() == OPER_INSTR &&
		    op2.instr()->opcode() == io_ldc) {
			in_ldc *ins = dynamic_cast<in_ldc *>(op2.instr());
			return new condition_tree(ctt_mul,
				new condition_tree(ins->value().integer()),
				construct_condition_tree(op1, verbose));
		} else
			return new condition_tree(ctt_invalid);
	}
	case io_neg: {
		in_rrr *ir = dynamic_cast<in_rrr *>(ins);
		return new condition_tree(ctt_neg,
			construct_condition_tree(ir->src1_op(), verbose));
	}
	case io_rem: {
		in_rrr *ir = dynamic_cast<in_rrr *>(ins);
		operand op1 = ir->src1_op();
		operand op2 = ir->src2_op();
		if (op2.kind() != OPER_INSTR || op2.instr()->opcode() != io_ldc)
			return new condition_tree(ctt_invalid);
		in_ldc *ins = dynamic_cast<in_ldc *>(op2.instr());
		return new condition_tree(ctt_rem,
			construct_condition_tree(op1, verbose),
			new condition_tree(ins->value().integer()));
	}
	case io_cal: {
		condition_tree_type type;
		in_cal *cal = dynamic_cast<in_cal *>(ins);
		type = builtin_binary(cal);
		if (type == ctt_invalid)
			break;
		return new condition_tree(type,
			construct_condition_tree(cal->argument(0), verbose),
			construct_condition_tree(cal->argument(1), verbose));
	}
	}

	return new condition_tree(ctt_invalid);
}

condition_tree *construct_condition_tree(instruction *ins, int verbose)
{
	condition_tree *s1, *s2;

	switch (ins->opcode()) {
	case io_sle:
	case io_sl:
	case io_seq:
	case io_sne:
		break;
	default:
		if (verbose) {
			fprintf(stderr, "Unhandled guard:\n");
			ins->print(stderr);
		}
		return new condition_tree(ctt_invalid);
	}

	in_rrr *ir = dynamic_cast<in_rrr *>(ins);
	s1 = construct_condition_tree(ir->src1_op(), verbose);
	s2 = construct_condition_tree(ir->src2_op(), verbose);
	switch (ins->opcode()) {
	/* s2 - s1 >= 0 */ 
	case io_sle:
		s2 = new condition_tree(ctt_sub, s2, s1);
		s1 = new condition_tree(ctt_ge, s2);
		return s1;
		
	/* s2 - s1 - 1 >= 0 */ 
	case io_sl:
		return new condition_tree(ctt_ge,
			new condition_tree(ctt_add,
				new condition_tree(ctt_sub, s2, s1),
				new condition_tree(-1)));
	/* s1 - s2 == 0 */ 
	case io_seq:
		return new condition_tree(ctt_eq,
			new condition_tree(ctt_sub, s1, s2));
	/* s1 - s2 - 1 >= 0 or s2 - s1 - 1 >= 0 */
	case io_sne: {
		condition_tree *t1, *t2;
		t1 = new condition_tree(ctt_ge,
			new condition_tree(ctt_add,
			    new condition_tree(ctt_sub, s1->dup(), s2->dup()),
			    new condition_tree(-1)));
		t2 = new condition_tree(ctt_ge,
			new condition_tree(ctt_add,
			    new condition_tree(ctt_sub, s2, s1),
			    new condition_tree(-1)));
		return new condition_tree(ctt_or, t1, t2);
	}
	}
	assert(0);
}

void condition_tree::dump(FILE *out, int indent)
{
	const char *str;
	fprintf(out, "%*s", indent, "");
	switch (type) {
	case ctt_invalid: str = "invalid"; break;
	case ctt_and: str = "&&"; break;
	case ctt_or: str = "||"; break;
	case ctt_ge: str = ">="; break;
	case ctt_eq: str = "=="; break;
	case ctt_cst: str = NULL; break;
	case ctt_sym: str = NULL; break;
	case ctt_neg: str = "-"; break;
	case ctt_add: str = "+"; break;
	case ctt_sub: str = "-"; break;
	case ctt_mul: str = "*"; break;
	case ctt_rem: str = "%"; break;
	case ctt_max: str = "max"; break;
	case ctt_min: str = "min"; break;
	case ctt_ceild: str = "ceild"; break;
	case ctt_floord: str = "floord"; break;
	case ctt_lod: str = "lod"; break;
	default: assert(0);
	}
	if (!str) {
		switch (type) {
		case ctt_cst:
			fprintf(out, "%d\n", val);
			break;
		case ctt_sym:
			fprintf(out, "%s\n", var->name());
			break;
		}
		return;
	}
	fprintf(out, "%s\n", str);
	if (left)
		left->dump(out, indent + 2);
	if (right)
		right->dump(out, indent + 2);
}

/* This function moves min/max nodes up the tree by performing
 * simple transformations and then replaces the special case
 * of min(A,B) >= 0 by A >= 0 && B >= 0.
 * Other mins and maxs that are left over result in an invalid
 * tree.
 */
void condition_tree::eliminate_minmax()
{
	switch (type) {
	case ctt_invalid:
	case ctt_cst:
	case ctt_sym:
	case ctt_mul:
	case ctt_rem:
	case ctt_max:
	case ctt_min:
		return;
	case ctt_or:
	case ctt_and:
	case ctt_add:
		right->eliminate_minmax();
	case ctt_ge:
	case ctt_eq:
	case ctt_neg:
		left->eliminate_minmax();
		break;
	case ctt_sub:
		assert(0);
	}

	switch (type) {
	case ctt_neg: {
		if (left->type == ctt_min || left->type == ctt_max) {
			/* replace -max/min(A,B) by min/max(-A,-B) */
			condition_tree *tmp = left;
			left = new condition_tree(ctt_neg, tmp->left);
			right = new condition_tree(ctt_neg, tmp->right);
			type = tmp->type == ctt_min ? ctt_max : ctt_min;
			tmp->left = tmp->right = NULL;
			delete tmp;
			break;
		}
		if (left->type == ctt_floord || left->type == ctt_ceild) {
			/* replace -floord/ceild(A,B) by ceild/floord(-A,B) */
			condition_tree *tmp = left;
			left = new condition_tree(ctt_neg, tmp->left);
			right = tmp->right;
			type = tmp->type == ctt_floord ? ctt_ceild : ctt_floord;
			tmp->left = tmp->right = NULL;
			delete tmp;
			break;
		}
		break;
	}
	case ctt_add: {
		if (left->type == ctt_min || left->type == ctt_max ||
		    right->type == ctt_min || right->type == ctt_max) {
			/* if right is min/max and left is not,
			 * switch arguments
			 */
			if (left->type != ctt_min && left->type != ctt_max) {
				condition_tree *tmp = left;
				left = right;
				right = tmp;
			}
			/* replace min/max(A,B)+C by min/max(A+C,B+C) */
			condition_tree *tmp = left;
			left = new condition_tree(ctt_add,
						    tmp->left, right->dup());
			right = new condition_tree(ctt_add, tmp->right, right);
			type = tmp->type;
			tmp->left = tmp->right = NULL;
			delete tmp;
			break;
		}
		if (left->type == ctt_floord || left->type == ctt_ceild ||
		    right->type == ctt_floord || right->type == ctt_ceild) {
			/* if right is ceild/floord and left is not,
			 * switch arguments
			 */
			if (left->type != ctt_floord &&
			    left->type != ctt_ceild) {
				condition_tree *tmp = left;
				left = right;
				right = tmp;
			}
			/* replace ceild(n,d)+e by ceild(n+d*e,d) */
			condition_tree *tmp = right;
			type = left->type;
			right = left->right;
			left->type = ctt_add;
			left->right = new condition_tree(ctt_mul,
							    right->dup(), tmp);
			break;
		}
		break;
	}
	case ctt_ge: {
		if (left->type == ctt_min || left->type == ctt_max) {
			if (left->type == ctt_max) {
				/* this case not handled for now */
				delete left;
				left = NULL;
				type = ctt_invalid;
				break;
			}
			/* replace min(A,B) >= 0 by A >= 0 && B >= 0 */
			condition_tree *tmp = left;
			left = new condition_tree(ctt_ge, tmp->left);
			right = new condition_tree(ctt_ge, tmp->right);
			type = ctt_and;
			tmp->left = tmp->right = NULL;
			delete tmp;
			left->eliminate_minmax();
			right->eliminate_minmax();
			break;
		}
		if (left->type == ctt_floord || left->type == ctt_ceild) {
			if (left->type == ctt_ceild) {
				/* this case not handled for now */
				delete left;
				left = NULL;
				type = ctt_invalid;
				break;
			}
			/* replace floord(A,B) >= 0 by A >= 0 */
			condition_tree *tmp = left;
			left = tmp->left;
			tmp->left = NULL;
			delete tmp;
			left->eliminate_minmax();
			break;
		}
		break;
	}
	case ctt_eq:
		if (left->type != ctt_min &&
		    left->type != ctt_max)
			break;
		/* this case not handled for now */
		delete left;
		left = NULL;
		type = ctt_invalid;
		break;
	}
}
