#include <algorithm>
#include <vector>
#include <isl/constraint.h>
#include <omega.h>

#include "convert.h"

using namespace omega;

typedef std::vector<Variable_ID> varvector;

static void max_index(Constraint_Handle c, varvector& vv, varvector& params)
{
	for (Constr_Vars_Iter cvi(c); cvi; ++cvi) {
		Variable_ID var = (*cvi).var;
		if (find(vv.begin(), vv.end(), var) == vv.end() &&
		    find(params.begin(), params.end(), var) == params.end())
			vv.push_back(var);
	}
}

static void set_constraint(isl_constraint *c, Constraint_Handle h,
				varvector &vv, varvector &params)
{
	isl_int v;

	isl_int_init(v);

	for (int i = 0; i < vv.size(); ++i) {
		isl_int_set_si(v, h.get_coef(vv[i]));
		isl_constraint_set_coefficient(c, isl_dim_set, i, v);
	}
	for (int i = 0; i < params.size(); ++i) {
		isl_int_set_si(v, h.get_coef(params[i]));
		isl_constraint_set_coefficient(c, isl_dim_param, i, v);
	}
	isl_int_set_si(v, h.get_const());
	isl_constraint_set_constant(c, v);

	isl_int_clear(v);
}

static isl_basic_set *disjunct2basic_set(DNF_Iterator &di, isl_space *dim,
					    varvector &vv, varvector &params)
{
	int d = vv.size();
	isl_constraint *c;
	isl_basic_set *bset;
	isl_local_space *ls;

	for (EQ_Iterator ei = (*di)->EQs(); ei; ++ei)
		max_index((*ei), vv, params);
	for (GEQ_Iterator gi = (*di)->GEQs(); gi; ++gi)
		max_index((*gi), vv, params);

	dim = isl_space_add_dims(dim, isl_dim_set, vv.size() - d);
	bset = isl_basic_set_universe(isl_space_copy(dim));
	ls = isl_local_space_from_space(dim);

	for (EQ_Iterator ei = (*di)->EQs(); ei; ++ei) {
		c = isl_equality_alloc(isl_local_space_copy(ls));
		set_constraint(c, (*ei), vv, params);
		bset = isl_basic_set_add_constraint(bset, c);
	}
	for (GEQ_Iterator gi = (*di)->GEQs(); gi; ++gi) {
		c = isl_inequality_alloc(isl_local_space_copy(ls));
		set_constraint(c, (*gi), vv, params);
		bset = isl_basic_set_add_constraint(bset, c);
	}

	isl_local_space_free(ls);

	bset = isl_basic_set_project_out(bset, isl_dim_set, d, vv.size() - d);
	vv.resize(d);

	return bset;
}

__isl_give isl_map *relation2map(isl_ctx *ctx, Relation &r)
{
	varvector vv;
	varvector params;
	isl_space *dim;
	isl_space *dim_set;
	isl_map *map;

	for (int j = 1; j <= r.n_inp(); ++j)
		vv.push_back(r.input_var(j));
	for (int j = 1; j <= r.n_out(); ++j)
		vv.push_back(r.output_var(j));

	const Variable_ID_Tuple *globals = r.global_decls();
	dim = isl_space_alloc(ctx, globals->size(), r.n_inp(), r.n_out());
	for (int i = 0; i < globals->size(); ++i) {
		Variable_ID v = r.get_local((*globals)[i+1]);
		params.push_back(v);
		dim = isl_space_set_dim_name(dim, isl_dim_param, i, v->base_name);
	}

	dim_set = isl_space_set_alloc(ctx, params.size(), vv.size());
	map = isl_map_empty(isl_space_copy(dim));

	for (DNF_Iterator di(r.query_DNF()); di; ++di) {
		isl_basic_set *bset;
		isl_basic_map *bmap;

		bset = disjunct2basic_set(di, isl_space_copy(dim_set), vv, params);
		bmap = isl_basic_map_from_basic_set(bset, isl_space_copy(dim));
		map = isl_map_union(map, isl_map_from_basic_map(bmap));
	}

	isl_space_free(dim_set);
	isl_space_free(dim);

	return map;
}

__isl_give isl_set *relation2set(isl_ctx *ctx, Relation &r)
{
	varvector vv;
	varvector params;
	isl_space *dim;
	isl_set *set;

	for (int j = 1; j <= r.n_set(); ++j)
		vv.push_back(r.set_var(j));

	const Variable_ID_Tuple *globals = r.global_decls();
	dim = isl_space_set_alloc(ctx, globals->size(), r.n_set());
	for (int i = 0; i < globals->size(); ++i) {
		Variable_ID v = r.get_local((*globals)[i+1]);
		params.push_back(v);
		dim = isl_space_set_dim_name(dim, isl_dim_param, i, v->base_name);
	}

	set = isl_set_empty(isl_space_copy(dim));

	for (DNF_Iterator di(r.query_DNF()); di; ++di) {
		isl_basic_set *bset;

		bset = disjunct2basic_set(di, isl_space_copy(dim), vv, params);
		set = isl_set_union(set, isl_set_from_basic_set(bset));
	}

	isl_space_free(dim);

	return set;
}
