#include <stdio.h>
#include <string.h>
#include <iostream>
#include <map>

#include "version.h"

#include <isa/yaml.h>
#include <isa/pdg.h>
#include "da.h"

#include <isl/set.h>
#include <isl/map.h>

#include "da_options.h"

using pdg::PDG;
using namespace std;
using namespace da;

/* Return an input array that is otherwise a copy of a */
pdg::array *input_copy(PDG *pdg, pdg::array *a)
{
	pdg::array *copy = new pdg::array;
	copy->id = a->id;
	copy->name = a->name;
	copy->dims = a->dims;
	copy->element_type = a->element_type;
	copy->type = pdg::array::input;
	pdg->arrays.push_back(copy);
	return copy;
}

/* Determine the type (input/output/temp) of each array in the program.
 * An array is considered
 *	- input, if some element is read without being written
 *	  If so, the corresponding access are changed to read from
 *	  a copy of the array marked as input.  The original array
 *	  is then marked temporary or output, as appropriate.
 *	- output, if some element is written without being subsequently read
 *	  or overwritten
 *	- temporary, otherwhise
 */
void determine_array_types(PDG *pdg, bool verbose)
{
	struct isl_ctx *ctx = pdg->get_isl_ctx();
	isl_printer *prn;
	std::map<pdg::node *, isl_set *> exposed;
	std::map<pdg::array *, pdg::array *> shadow_input;

	unsigned nparam = pdg->params.size();
	for (int j = 0; j < pdg->arrays.size(); ++j)
		pdg->arrays[j]->type = pdg::array::temp;
	for (int j = 0; j < pdg->dependences.size(); ++j) {
		pdg::dependence *dep = pdg->dependences[j];
		pdg::array *a = dep->array;
		if (dep->type != pdg::dependence::uninitialized)
			continue;
		if (shadow_input.find(a) == shadow_input.end())
			shadow_input[a] = input_copy(pdg, a);
		dep->array = shadow_input[a];
	}
	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		pdg::statement *s = node->statement;

		isl_set *context = pdg->get_context_isl_set();
		exposed[node] = node->source->get_isl_set(ctx);
		exposed[node] = isl_set_intersect_params(exposed[node],
							    context);
	}
	for (int i = 0; i < pdg->dependences.size(); ++i) {
		pdg::dependence *dep = pdg->dependences[i];
		if (dep->type != pdg::dependence::flow &&
		    dep->type != pdg::dependence::output)
			continue;
		pdg::node *node = dep->from;
		isl_map *rel = dep->relation->get_isl_map(ctx);
		exposed[node] = isl_set_subtract(exposed[node],
						isl_map_domain(rel));
	}
	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		pdg::statement *s = node->statement;

		int is_empty = isl_set_is_empty(exposed[node]);
		assert(is_empty >= 0);
		if (verbose && !is_empty) {
			fprintf(stderr,
				"exposed write domain for node %d, line %d\n",
				node->nr, node->statement->line);
			prn = isl_printer_to_file(ctx, stderr);
			prn = isl_printer_set_indent(prn, 4);
			prn = isl_printer_print_set(prn, exposed[node]);
			prn = isl_printer_end_line(prn);
			isl_printer_free(prn);
		}
		isl_set_free(exposed[node]);
		if (is_empty)
			continue;

		for (int j = 0; j < s->accesses.size(); ++j) {
			if (s->accesses[j]->type != pdg::access::write)
				continue;
			pdg::array *a = s->accesses[j]->array;
			a->type = pdg::array::output;
		}
	}
}

int main(int argc, char * argv[])
{
    isl_ctx *ctx;
    FILE *in = stdin, *out = stdout;
    PDG *pdg;
    struct options *options = options_new_with_defaults();

    argc = options_parse(options, argc, argv, ISL_ARG_ALL);

    ctx = isl_ctx_alloc_with_options(&options_args, options);

    if (options->input && strcmp(options->input, "-")) {
	in = fopen(options->input, "r");
	assert(in);
	if (!options->output) {
	    int len = strlen(options->input);
	    if (len > 5 && !strcmp(options->input+len-5, ".yaml"))
		len -= 5;
	    options->output = (char *)malloc(len+9+1);
	    strncpy(options->output, options->input, len);
	    strcpy(options->output+len, "_da.yaml");
	}
    }

    if (options->array_types)
	options->types |= TYPE_FLOW;

    pdg = PDG::Load(in, ctx);
    assert(pdg);

    compute_filter_sources(pdg);
    for (int i = 0; i < pdg->arrays.size(); ++i) {
	if (options->types & (TYPE_FLOW))
	    find_deps(pdg, pdg->arrays[i], flow);
	if (options->types & (TYPE_ANTI))
	    find_deps(pdg, pdg->arrays[i], anti);
	if (options->types & (TYPE_REUSE_PAIR))
	    find_deps(pdg, pdg->arrays[i], reuse_pair);
	if (options->types & (TYPE_OUTPUT))
	    find_deps(pdg, pdg->arrays[i], output);
    }

    if (options->array_types)
	determine_array_types(pdg, options->verbose);

    if (options->output && strcmp(options->output, "-")) {
	out = fopen(options->output, "w");
	assert(out);
    }

    pdg->add_history_line("da", argc, argv);

    pdg->Dump(out);
    pdg->free();
    delete pdg;

    isl_ctx_free(ctx);

    return 0;
}
