#ifndef DEPENDENCE_GRAPH_H
#define DEPENDENCE_GRAPH_H

#include <vector>
#include <isl/map.h>
#include <isl/set.h>
#include <isa/pdg.h>

struct computation;

struct edge {
	computation		*source;
	isl_map			*relation;
	unsigned		 pos;
	enum type {
		normal,
		expansion,
	} type;

	edge() : type(normal) { }
	~edge() {
		isl_map_free(relation);
	}

	void dump(void);
};

struct computation {
	isl_set			*domain;
	const char		*operation;
	unsigned		 arity;
	int			 location;
	int			 dim;
	unsigned		 input : 1;
	unsigned		 is_expanded : 1;
	computation		*original;

	std::vector<edge *>	 edges;

	computation() : input(0), is_expanded(0), original(NULL), dim(-1) {}

	~computation() {
		for (int i = 0; i < edges.size(); ++i)
			delete edges[i];
		isl_set_free(domain);
		free((char *)operation);
	}

	bool has_same_source(computation *other) const {
		if (this == other)
			return 1;
		if (this->original == other)
			return 1;
		if (this == other->original)
			return 1;
		if (this->original && this->original == other->original)
			return 1;
		return 0;
	}

	bool is_input() const {
		return input;
	}

	bool is_copy() const {
		return !input && !is_expanded &&
		       !strcmp(operation, "") && arity == 1;
	}

	void dump(void);
};

struct input_array : computation {
	input_array(const char *name, __isl_take isl_space *dims) {
		dim = isl_space_dim(dims, isl_dim_set);
		if (*name)
			dims = isl_space_set_tuple_name(dims,
							isl_dim_set, name);
		domain = isl_set_universe(dims);
		operation = strdup(name);
		arity = 0;
		location = -1;
		input = 1;
	}
};

struct dependence_graph {
	isl_set				*context;
	std::vector<char *>	 	 output_arrays;
	std::vector<unsigned>	 	 output_array_dims;
	std::vector<computation *>	 vertices;
	std::vector<computation *>	 input_computations;
	computation 			*out;

	dependence_graph() : out(NULL), context(NULL) {}
	~dependence_graph() {
		isl_set_free(context);
		for (int i = 0; i < output_arrays.size(); ++i)
			free(output_arrays[i]);
		for (int i = 0; i < vertices.size(); ++i)
			delete vertices[i];
		delete out;
	}

	void splice(computation *comp, edge *e);
	computation *split_comp(computation *comp, edge *e);
	void split_edges(computation *comp_orig, computation *comp_dup);
	void split_edges(computation *comp,
			 computation *comp_orig, computation *comp_dup);
	computation *associative_node(edge **e,
		const std::vector<const char *> &associative);
	void flatten_associative_operators(
		const std::vector<const char *> &associative);
	void dump(void);
};

dependence_graph *pdg_to_dg(pdg::PDG *pdg, unsigned out_dim,
	__isl_take isl_set *context);

#endif
