#include <isl/set.h>
#include <isl/map.h>
#include <isl/union_map.h>
#include "version.h"
#include <isa/pdg.h>
#include "da.h"
#include "dependence_graph_relation_options.h"

/* Print the dependence graph of the input program as a single relation.
 * If type is "mem" (the default), then memory based dependences are
 * computed, i.e., there is a dependence between two statement iterations
 * if they (may) access the same memory location, where at least one
 * of the accesses is a write, and if the first statement iteration is
 * executed before the second.
 * If type is "val", then value based dependence are computed.
 * If type is "flow", then value based flow dependence are computed.
 */

using pdg::PDG;
using namespace da;

/* Compute and return the maximal iteration domain dimension.
 */
int get_max_dim(PDG *pdg)
{
	int max_dim = -1;
	isl_ctx *ctx = pdg->get_isl_ctx();

	for (int i = 0; i < pdg->nodes.size(); ++i) {
		isl_set *dom;
		int dim;

		dom = pdg->nodes[i]->source->get_isl_set(ctx);

		dim = isl_set_dim(dom, isl_dim_set);
		if (dim > max_dim)
			max_dim = dim;

		isl_set_free(dom);
	}

	return max_dim;
}

/* Return 2 * d + b, where d is the shared nesting level and b
 * is true is first textually appears before second.
 */
static int precedes_level_nodes(pdg::node *first, pdg::node *second)
{
	int d = 0;
	int cmp = 0;
	for (int i = 0; i < first->prefix.size(); ++i) {
		if (i >= second->prefix.size())
			break;
		cmp = first->prefix[i] - second->prefix[i];
		if (cmp)
			break;
		if (first->prefix[i] == -1)
			++d;
	}
	return 2 * d + (cmp < 0);
}

/* Pad either the input or the output dimension of the map to max_dim
 * and add an extra dimension with as fixed value "nr".
 */
isl_map *normalize_dim(isl_map *map, enum isl_dim_type type, int nr,
	int max_dim, struct options *options)
{
	if (options->named) {
		char s[20];
		sprintf(s, "N_%d", nr);
		map = isl_map_set_tuple_name(map, type, s);
	} else {
		unsigned dim = isl_map_dim(map, type);
		map = isl_map_add_dims(map, type, max_dim - dim + 1);
		for (int i = dim; i < max_dim; ++i)
			isl_map_fix_si(map, type, i, 0);
		isl_map_fix_si(map, type, max_dim, nr);
	}
	return map;
}

/* Compute memory based dependences from the given access to any
 * other access and add them to "graph".
 *
 * If there are any data-dependent accesses, i.e., with nested
 * accesses, then the nested access values are projected out.
 */
isl_union_map *add_dependences_from(isl_union_map *graph, PDG *pdg,
	pdg::node *from_node, pdg::access *from, int max_dim,
	struct options *options)
{
	isl_ctx *ctx = pdg->get_isl_ctx();

	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *to_node = pdg->nodes[i];
		pdg::statement *s = pdg->nodes[i]->statement;
		int level = precedes_level_nodes(from_node, to_node);
		int before = level % 2;
		level >>= 1;

		for (int j = 0; j < s->accesses.size(); ++j) {
			pdg::access *to = s->accesses[j];
			isl_map *r_from;
			isl_map *r_to;
			isl_map *dep;
			isl_map *order;
			int from_space, to_dim;

			if (to->array != from->array)
				continue;
			if (to->type == pdg::access::read &&
			    from->type == pdg::access::read)
				continue;

			r_from = from->map->get_isl_map(ctx);
			r_to = to->map->get_isl_map(ctx);

			from_space = from_node->source->dim();
			r_from = isl_map_project_out(r_from,
				    isl_dim_in, from_space,
				    from->map->input() - from_space);
			to_dim = to_node->source->dim();
			r_to = isl_map_project_out(r_to,
				    isl_dim_in, to_dim,
				    to->map->input() - to_dim);

			r_from = isl_map_intersect_domain(r_from,
					from_node->source->get_isl_set(ctx));
			r_to = isl_map_intersect_domain(r_to,
					to_node->source->get_isl_set(ctx));

			dep = isl_map_apply_range(r_from, isl_map_reverse(r_to));
			if (before)
				order = isl_map_lex_le_first(
					isl_map_get_space(dep), level);
			else
				order = isl_map_lex_lt_first(
					isl_map_get_space(dep), level);
			dep = isl_map_intersect(dep, order);
			dep = normalize_dim(dep, isl_dim_in,
						from_node->nr, max_dim, options);
			dep = normalize_dim(dep, isl_dim_out,
						to_node->nr, max_dim, options);

			graph = isl_union_map_add_map(graph, dep);
		}
	}

	return graph;
}

/* Compute all memory based dependences and add them to "graph".
 */
isl_union_map *add_mem_dependences(isl_union_map *graph, PDG *pdg, int max_dim,
	struct options *options)
{
	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		pdg::statement *s = pdg->nodes[i]->statement;

		for (int j = 0; j < s->accesses.size(); ++j) {
			pdg::access *access = s->accesses[j];
			graph = add_dependences_from(graph, pdg, node, access,
							max_dim, options);
		}
	}

	return graph;
}

/* Compute all value based flow dependences and add them to "graph".
 *
 * If there are any data-dependent accesses, i.e., with nested
 * accesses, then the nested access values are projected out.
 */
isl_union_map *add_val_dependences(isl_union_map *graph, PDG *pdg, int max_dim,
	struct options *options)
{
	isl_ctx *ctx = pdg->get_isl_ctx();

	for (int i = 0; i < pdg->arrays.size(); ++i)
		find_deps(pdg, pdg->arrays[i], flow);
	if (options->type != TYPE_FLOW) {
		for (int i = 0; i < pdg->arrays.size(); ++i)
			find_deps(pdg, pdg->arrays[i], anti);
		for (int i = 0; i < pdg->arrays.size(); ++i)
			find_deps(pdg, pdg->arrays[i], output);
	}

	for (int i = 0; i < pdg->dependences.size(); ++i) {
		int from_space, to_dim;
		pdg::dependence *dep = pdg->dependences[i];
		if (dep->type == pdg::dependence::uninitialized)
			continue;

		isl_map *map = dep->relation->get_isl_map(ctx);
		from_space = dep->from->source->dim();
		to_dim = dep->to->source->dim();
		map = isl_map_project_out(map, isl_dim_in, from_space,
				dep->relation->input() - from_space);
		map = isl_map_project_out(map, isl_dim_out, to_dim,
				dep->relation->output() - to_dim);
		map = normalize_dim(map, isl_dim_in, dep->from->nr, max_dim,
					options);
		map = normalize_dim(map, isl_dim_out, dep->to->nr, max_dim,
					options);

		graph = isl_union_map_add_map(graph, map);
	}

	return graph;
}

int main(int argc, char *argv[])
{
	FILE *in = stdin;
	isl_ctx *ctx = isl_ctx_alloc();
	PDG *pdg;
	int nparam;
	int max_dim;
	isl_space *dim;
	isl_union_map *graph;
	isl_printer *p;
	struct options *options = options_new_with_defaults();

	argc = options_parse(options, argc, argv, ISL_ARG_ALL);

	if (options->input && strcmp(options->input, "-"))
		in = fopen(options->input, "r");
	assert(in);

	pdg = PDG::Load(in, ctx);

	max_dim = get_max_dim(pdg);

	nparam = pdg->params.size();
	dim = isl_space_alloc(pdg->get_isl_ctx(), nparam, max_dim + 1, max_dim + 1);
	isl_dim_set_parameter_names(dim, pdg->params);
	graph = isl_union_map_empty(dim);

	if (options->type == TYPE_MEM)
		graph = add_mem_dependences(graph, pdg, max_dim, options);
	else
		graph = add_val_dependences(graph, pdg, max_dim, options);

	p = isl_printer_to_file(pdg->get_isl_ctx(), stdout);
	p = isl_printer_set_output_format(p, ISL_FORMAT_ISL);
	p = isl_printer_print_union_map(p, graph);
	p = isl_printer_end_line(p);
	isl_printer_free(p);

	isl_union_map_free(graph);

	pdg->free();
	delete pdg;
	isl_ctx_free(ctx);

	return 0;
}
