#ifndef DOMAIN_H
#define DOMAIN_H

#include "suif_inc.h"
#include "srel.h"
#include "condition_tree.h"

#include <map>
#include <vector>

using std::map;
//using std::vector;

extern const char * k_static_root;
extern const char * k_iteration_domain;
extern const char * k_domain_symtab;
extern const char *k_array_bounds;

typedef std::vector<var_sym *> var_sym_v; // why is std:: needed ?

/* A variable that is either (always) a parameter or (sometimes) an iterator
 * or a scalar that is assigned to at least once.
 * The same variable may be used both as an iterator and a scalar that
 * is being written to, as long as the second use only occurs outside
 * the loops that the variables iterates over.
 */
struct Param_Iterator : public Global_Var_Decl {
    Param_Iterator(Const_String name) :
    	Global_Var_Decl(name), iterator(false), assigned(false), value(NULL),
	valid_for_read(true), lb_set(false), temp(false) {}
    bool iterator;
    bool assigned;	/* variable has been assigned a value at some point */
    bool valid_for_read;    /* current value may be unknown, but the variable
				was not last used as an iterator. */
    condition_tree *value; /* current value of the variable (NULL if none) */
    bool lb_set;
    int  lb;
    bool temp;		/* a dummy variable; not a data variable or parameter */
    ~Param_Iterator() {
	delete value;
    }
};

struct symtab_ann {
    global_symtab *	symtab;
    map<var_sym *, Param_Iterator *> vars;
    void add_var(var_sym *var);
    void add_var(named_lin_ineq *nlq);
    var_sym *create_tmp();
};

class domain_constructor;

class guard_constructor {
    int			verbose;
    symtab_ann	    	*sa;

public:
    SRelation		*s;
    domain_constructor		*dc;
    guard_constructor(symtab_ann *sa, int verbose, domain_constructor *dc) :
			sa(sa), verbose(verbose), dc(dc) {
	s = new SRelation(sa);
    }
    ~guard_constructor() {
	delete s;
    }

    bool construct_guard(instruction *ins);
    bool construct_guard(condition_tree *ct, F_And *f);
    bool fill_q(Constraint_Handle& h, condition_tree *ct, int mult);
};

struct context_info {
    symtab_ann *sa;
    SRelation  *ctx;
    tree_proc  *tp;

    context_info() : sa(new symtab_ann), ctx(NULL), tp(NULL) {}
};

struct for_ann {
    SRelation      	*bounds;
    var_sym		*iterator;
    tree_node_list	*body;

    for_ann(SRelation *bnd, var_sym *it, tree_node_list *bdy) :
	bounds(bnd), iterator(it), body(bdy) {}
    ~for_ann() {
	delete bounds;
    }
};

struct stmt;
struct stmt_node;
struct stmt_access;

class domain_constructor {
    SRelation * 	c;
    symtab_ann *	sa;
    int			level;
    int			plevel;
    var_sym_v		its;
    var_sym_v		out;
    instruction *	root;
    int			verbose;
    int			oldstyle;
    std::vector<int>	prefix;
    tree_proc *		tp;
    tree_node *		scope;

    SRelation * reorder(SRelation * c);
public:
    symtab_ann *get_symtab() { return sa; }
    SRelation *reorder(SRelation *c,
	    const std::vector<var_sym *> &nested_vars, int output);
    domain_constructor(tree_proc *tp, int verbose, int oldstyle) : 
	tp(tp), verbose(verbose), oldstyle(oldstyle), scope(NULL) {}
    bool is_broken_up_loop(tree_node *t1, tree_node *t2, tree_node_list **body,
				var_sym **iterator);
    void scan(context_info *ctx);
    boolean scan(tree_node_list * tp);
    boolean scan(tree_node * tp);
    void apply_value_bounds(annote *in);
    void check_pragmas(instruction *in);
    void construct();
    void construct_broken_up_loop(for_ann *fa);
    void construct(tree_node_list * tp);
    void construct(tree_node * tp);
    void recurse(tree_node * tp);
    void extract_messy_index(SRelation *map,
		Constraint_Handle &h, instruction *in, int f);
    void extract_messy_index(SRelation *map,
		Constraint_Handle &h, operand op, int f);
    SRelation *extract_messy_index(operand index, var_sym *out);
    stmt *extract_stmt(tree_instr *i);
    stmt_node *extract_stmt_node(instruction *i);
    stmt_node *extract_stmt_node(operand &op, bool write,
				 instruction *parent = NULL);
    stmt_node *extract_stmt_node_cal(in_cal *cal);
    stmt_access *create_access(dep_instr_annote *di, bool write);
    stmt_access *check_access(instruction *in, bool write);
    void visit(tree_node *t);
    stmt_node *read_stmt(instruction *i);
    stmt_access *data_dependent_test_cond(SRelation *c1, SRelation *c0);
    bool is_data_dependent_assignment(tree_if *ti);
    boolean construct_guard(tree_if *ti, tree_node_list *tnl,
	bool allow_nested);
    void set_prefix_and_iterator(int prefix_value, var_sym *iterator = NULL);
    void add_iteration_domain(suif_object *obj, instruction *insn);
    void leave_scope(tree_node *old_scope);
};

struct stmt_node {
	var_sym *var; // variable, if any, associated to this stmt
	operand op;   // operand of nested nodes
	stmt_node() : var(NULL) {}
	virtual ~stmt_node() {}
	virtual bool is_access() = 0;
	virtual void reorder(domain_constructor *dc) {}
};

struct stmt_access : stmt_node {
	bool			 write;
	var_sym			*array;
	int		 	 output;
	SRelation		*map;
	bool			 reordered;

	stmt_access(bool w, SRelation *map, var_sym *a, int o) :
		write(w), map(map), array(a), output(o), reordered(false) {}
	virtual bool is_access() { return true; }
	~stmt_access() {
		delete map;
	}
	virtual void reorder(domain_constructor *dc) {
		if (reordered)
			return;
		std::vector<var_sym *> nested_vars;
		for (int i = 0; i < map->nested.size(); ++i) {
			map->nested[i]->reorder(dc);
			nested_vars.push_back(map->nested[i]->var);
		}
		map = dc->reorder(map, nested_vars, output);
		reordered = true;
	}
};

struct stmt_call : stmt_node {
	const char *name;
	std::vector<stmt_node *>	arg;

	stmt_call(const char *name) : name(name) {}
	virtual bool is_access() { return false; }

	~stmt_call() {
		for (int i = 0; i < arg.size(); ++i)
			delete arg[i];
	}
	virtual void reorder(domain_constructor *dc) {
		for (int i = 0; i < arg.size(); ++i)
			arg[i]->reorder(dc);
	}
};

struct stmt {
	std::vector<stmt_access *>	out;
	stmt_node			*tree;

	stmt() {
		tree = NULL;
	}
	~stmt() {
		for (int i = 0; i < out.size(); ++i)
			delete out[i];
		delete tree;
	}
	void reorder(domain_constructor *dc) {
		for (int i = 0; i < out.size(); ++i)
			out[i]->reorder(dc);
		tree->reorder(dc);
	}
};

struct domain_ann {
    instruction	   	*insn;
    SRelation      	*domain;
    int			 level;
    int			 plevel;
    var_sym_v		 its;
    std::vector<int>	 prefix;
    stmt		*statement;
    int			 line;

    domain_ann(instruction *i, SRelation *d, int l, int pl, var_sym_v it,
		std::vector<int> p, stmt *s, int line):
	insn(i), domain(d), level(l), plevel(pl), its(it), prefix(p),
	statement(s), line(line) {}
};

struct access_ann {
    boolean		 write;
    named_lin_ineq      *nlq;
    var_sym		*array;
    int		 	 output;
    int			 arg_pos; /* argument position in function call or -1 */

    access_ann(boolean w, named_lin_ineq *n, var_sym *a, int o, int pos) :
	write(w), nlq(n), array(a), output(o), arg_pos(pos) {}
};

void construct_domains(tree_proc * tp, int verbose, int oldstyle, context_info *ctx);
void init(int &argc, char *argv[]);

#endif
