#include "suif_inc.h"
#include <map>
#include <sstream>
#include "convert.h"
#include "dump.h"
#include <isa/pdg.h>

#ifndef HAVE_STRNDUP
extern "C" char *strndup(const char *s, size_t n);
#endif

using pdg::PDG;
using pdg::array;

void
dumper::visit(tree_node *t)
{
    if (t->peek_annote(k_static_root)) {
	/* Only dump a single domain */
	if (root == 0)
	    root = t->number();
	else
	    root = -1;
    }

    if (root <= 0)
	return;

    domain_ann *d;
    if (t->kind() == TREE_INSTR) {
	instruction *in = ((tree_instr*)t)->instr();
	d = (domain_ann *)in->peek_annote(k_iteration_domain);
    } else
	d = (domain_ann *)t->peek_annote(k_iteration_domain);
    if (d == NULL)
	return;

    nodes.push_back(d);
}

dumper::~dumper()
{
}

string c_type(type_node *tn)
{
    std::ostringstream strm;

    switch (tn->op()) {
    case TYPE_INT: {
        base_type *bt = (base_type *) tn;

        if (!bt->is_signed())
	    strm << "unsigned ";
	switch (c_int_type(tn)) {
	case C_char:
	    if (bt->is_signed())
		strm << "signed ";
	    strm << "char";
	    break;
	case C_int:
	    strm << "int";
	    break;
	case C_short:
	    strm << "short";
	    break;
	case C_long:
	    strm << "long";
	    break;
	case C_longlong:
	    strm << "long long";
	    break;
	default:
	    assert(0);
	}
	break;
    }
    case TYPE_FLOAT: {
	switch (c_float_type(tn)) {
	case C_float:
	    strm << "float";
	    break;
	case C_double:
	    strm << "double";
	    break;
	case C_longdouble:
	    strm << "long double";
	    break;
	default:
	    assert(0);
	}
	break;
    }
    case TYPE_STRUCT: {
        struct_type *st = (struct_type *) tn;
	strm << "struct " << st->name();
	break;
    }
    default:
	assert(0);
    }

    return strm.str();
}

static int compute_total_read(struct stmt_node *node, int c)
{
	if (node->is_access()) {
		stmt_access *sa = static_cast<stmt_access *>(node);
		return c + !sa->write;
	}

	stmt_call *sc = static_cast<stmt_call *>(node);

	for (int i = 0; i < sc->arg.size(); ++i)
		c = compute_total_read(sc->arg[i], c);

	return c;
}

static int compute_total_read(struct stmt *s)
{
	return compute_total_read(s->tree, 0);
}

static type_node *dereference_type(type_node *tn, array *a)
{
	while (tn->is_array() || tn->is_ptr()) {
		if (tn->is_array()) {
			array_type *an = dynamic_cast<array_type*>(tn);
			assert(an->lower_bound().is_constant());
			assert(an->upper_bound().is_constant());
			tn = an->elem_type();
			a->dims.push_back(an->upper_bound().constant() -
					  an->lower_bound().constant() + 1);
		} else {
			ptr_type *an = dynamic_cast<ptr_type*>(tn);
			tn = an->ref_type();
			a->dims.push_back(-1);
		}
	}
	return tn;
}

pdg::access *dumper::add_access(pdg::node *pn, stmt_access *sa)
{
	isl_ctx *ctx = pdg->get_isl_ctx();
	pdg::access *a = new pdg::access;
	for (int i = 0; i < sa->map->nested.size(); ++i)
		a->nested.push_back(scan_stmt(pn, sa->map->nested[i]));
	var_sym *av = sa->array;
	if (av && arrays.find(av) == arrays.end()) {
		array *a = new array();
		a->id = av->sym_id();
		a->name = new str(av->name());
		type_node *tn = av->type();
		tn = dereference_type(tn, a);
		a->element_type = new str(c_type(tn));
		SRelation *bounds;
		bounds = (SRelation *)av->peek_annote(k_array_bounds);
		if (bounds) {
			isl_set *set = relation2set(ctx, *bounds);
			set = isl_set_align_params(set, isl_space_copy(dim));
			a->value_bounds = new pdg::IslSet(set);
		}
		pdg->arrays.push_back(a);
		arrays[av] = a;
	}
	a->array = av ? arrays[av] : NULL;
	bool write = sa->write;
	a->type = write ? pdg::access::write : pdg::access::read;
	isl_map *map = relation2map(ctx, *sa->map);
	map = isl_map_align_params(map, isl_space_copy(dim));
	if (sa->map->nested.size() > 0) {
		isl_space *dim = isl_space_domain(isl_map_get_space(map));
		isl_map *flatten;
		int in = isl_space_dim(dim, isl_dim_set) - sa->map->nested.size();
		dim = isl_space_from_domain(dim);
		dim = isl_space_move_dims(dim, isl_dim_out, 0,
				isl_dim_in, in, sa->map->nested.size());
		dim = isl_space_wrap(dim);
		flatten = isl_set_flatten_map(isl_set_universe(dim));
		map = isl_map_apply_domain(map, isl_map_reverse(flatten));
	}
	a->map = new pdg::IslMap(map);
	if (write)
		a->nr = totalread + nwrite++;
	else
		a->nr = nread++;
	pn->statement->accesses.push_back(a);
	return a;
}

pdg::call_or_access *dumper::scan_stmt(pdg::node *pn, stmt_node *n)
{
	pdg::call_or_access *coa = new pdg::call_or_access;

	if (n->is_access()) {
		coa->type = pdg::call_or_access::t_access;
		coa->access = add_access(pn, static_cast<stmt_access *>(n));
	} else {
		stmt_call *sc = static_cast<stmt_call *>(n);
		coa->type = pdg::call_or_access::t_call;
		coa->call = new pdg::function_call;
		coa->call->name = new str(string(sc->name));

		for (int i = 0; i < sc->arg.size(); ++i) {
			pdg::call_or_access *child;
			child = scan_stmt(pn, sc->arg[i]);
			coa->call->arguments.push_back(child);
		}
	}

	return coa;
}

/* Convert stmt to pdg::function_call and add all accesses to ps */
pdg::function_call *dumper::scan_stmt(pdg::node *pn, stmt *s)
{
	totalread = compute_total_read(s);
	nread = 0;
	nwrite = 0;

	pdg::function_call *call;
	pdg::call_or_access *coa = scan_stmt(pn, s->tree);

	for (int i = 0; i < s->out.size(); ++i)
		pn->statement->top_outputs.push_back(add_access(pn, s->out[i]));

	if (coa->type == pdg::call_or_access::t_access) {
		call = new pdg::function_call;
		call->arguments.push_back(coa);
		call->name = new str(string(""));
	} else {
		call = coa->call;
		delete coa;
	}
	return call;
}

void dumper::dump(int argc, char * argv[])
{
    isl_ctx *isl_ctx;

    if (root <= 0)
	return;

    isl_ctx = isl_ctx_alloc();
    pdg = new PDG(isl_ctx);
    pdg->placement = new str(string("original"));
    pdg->root = root;

    const char *name = tp->proc()->file()->name();
    const char *dot = strrchr(name, '.');
    if (!dot)
	pdg->name = new str(string(name));
    else {
	char *basename = strndup(name, dot-name);
	pdg->name = new str(string(basename));
	free(basename);
    }

    map<var_sym *, Param_Iterator *>::iterator gi;
    for (gi = sa->vars.begin(); gi != sa->vars.end(); ++gi) {
	if ((*gi).second->iterator)
	    continue;
	if ((*gi).second->assigned)
	    continue;
	if ((*gi).second->temp)
	    continue;
	pdg::parameter *p = new pdg::parameter;
	p->id = (*gi).first->sym_id();
	p->name = new str((*gi).second->base_name());
	if ((*gi).second->lb_set)
	    p->value = new integer((*gi).second->lb);

	file_symtab *symtab = tp->proc()->file()->symtab();
	var_def_list_iter j(symtab->var_defs());
	while (!j.is_empty()) {
	    var_def *def = j.step();
	    if (def->variable() != (*gi).first)
		continue;
	    annote_list_iter i(def->annotes());
	    while (!i.is_empty()) {
		annote *an = i.step();
		if (an->name() != k_repeat_init)
		    continue;
		immed_list *ims = an->immeds();
		if ((*ims)[2].kind() != im_int)
		    continue;
		p->value = new integer((*ims)[2].integer());
	    }
	}

	pdg->params.push_back(p);
    }

    dim = isl_space_alloc(isl_ctx, pdg->params.size(), 0, 0);
    isl_dim_set_parameter_names(dim, pdg->params);

    if (ctx->ctx) {
	isl_set *set = relation2set(isl_ctx, *ctx->ctx);
	set = isl_set_align_params(set, isl_space_copy(dim));
	set = isl_set_params(set);
	pdg->context = new pdg::IslSet(set);
    }

    int nodenr = 0;
    for (int i = 0; i < nodes.size(); ++i) {
	domain_ann *n = nodes[i];
	if (!n->domain->is_satisfiable())
	    continue;
	pdg::node * pn = new pdg::node;
	pdg::statement * ps = new pdg::statement;
	pdg->nodes.push_back(pn);
	pn->nr = nodenr++;
	pn->statement = ps;
	ps->operation = n->insn ? n->insn->number() : -1;
	ps->line = n->line;
	isl_set *set = relation2set(isl_ctx, *n->domain);
	set = isl_set_align_params(set, isl_space_copy(dim));
	pn->source = new pdg::IslSet(set);
	for (int j = 0; j < n->plevel; ++j)
	    pn->prefix.push_back(n->prefix[j]);
	ps->top_function = scan_stmt(pn, n->statement);
    }

    pdg->add_history_line("pers", argc, argv);
    pdg->Dump();
    pdg->free();
    delete pdg;
    isl_space_free(dim);
    isl_ctx_free(isl_ctx);
}

static void
visit (tree_node *t, void *d)
{
    ((dumper *)d)->visit(t);
}

void dumper::scan()
{
    root = 0;
    tp->map(::visit, (void *) this);
}

void dump(tree_proc *tp, context_info *ctx, int argc, char * argv[])
{
    dumper d(tp, ctx);
    d.scan();
    d.dump(argc, argv);
}
