#include <isl/map.h>
#include <isl/set.h>
#include <isl/union_map.h>
#include <isl/union_set.h>

#include <isa/yaml.h>
#include <isa/pdg.h>

using namespace pdg;

__isl_give isl_map *prefix2schedule(isl_ctx *ctx, pdg::node *node,
	int sched_dim)
{
	isl_space *dim;
	isl_map *schedule;

	dim = node->source->get_dim();
	dim = isl_space_map_from_set(dim);
	schedule = isl_map_identity(dim);

	for (int i = 0; i < node->prefix.size(); ++i) {
		int v = node->prefix[i];
		if (v == -1)
			continue;
		schedule = isl_map_insert_dims(schedule, isl_dim_out, i, 1);
		schedule = isl_map_fix_si(schedule, isl_dim_out, i, v);
	}
	schedule = isl_map_add_dims(schedule, isl_dim_out,
				    sched_dim - node->prefix.size());
	for (int i = node->prefix.size(); i < sched_dim; ++i)
		schedule = isl_map_fix_si(schedule, isl_dim_out, i, 0);

	return schedule;
}

int main(int argc, char * argv[])
{
	PDG *pdg;
	isl_ctx *ctx = isl_ctx_alloc();
	pdg = PDG::Load(stdin, ctx);
	isl_space *dim;
	isl_union_set *domain;
	isl_union_map *read;
	isl_union_map *write;
	isl_union_map *schedule;
	isl_printer *p;
	char name[40];
	int sched_dim = 0;

	dim = isl_space_set_alloc(ctx, pdg->params.size(), 0);
	isl_dim_set_parameter_names(dim, pdg->params);
	domain = isl_union_set_empty(dim);

	dim = isl_space_alloc(ctx, pdg->params.size(), 0, 0);
	isl_dim_set_parameter_names(dim, pdg->params);
	schedule = isl_union_map_empty(isl_space_copy(dim));
	write = isl_union_map_empty(isl_space_copy(dim));
	read = isl_union_map_empty(dim);

	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		if (node->prefix.size() > sched_dim)
			sched_dim = node->prefix.size();
	}

	for (int i = 0; i < pdg->nodes.size(); ++i) {
		pdg::node *node = pdg->nodes[i];
		pdg::statement *s = node->statement;
		isl_set *domain_i;
		isl_map *access_i, *schedule_i;

		snprintf(name, sizeof(name), "S%d", i);

		domain_i = node->source->get_isl_set(ctx);
		domain_i = isl_set_set_tuple_name(domain_i, name);
		domain = isl_union_set_add_set(domain, domain_i);

		for (int j = 0; j < s->accesses.size(); ++j) {
			pdg::access *access = s->accesses[j];

			if (!access->array)
				continue;

			access_i = access->map->get_isl_map(ctx);
			access_i = isl_map_set_tuple_name(access_i,
					isl_dim_in, name);
			access_i = isl_map_set_tuple_name(access_i,
					isl_dim_out,
					access->array->name->s.c_str());
			if (access->type == pdg::access::read)
				read = isl_union_map_add_map(read, access_i);
			else
				write = isl_union_map_add_map(write, access_i);
		}

		if (node->schedule)
			schedule_i = node->schedule->get_isl_map(ctx);
		else
			schedule_i = prefix2schedule(ctx, node, sched_dim);
		schedule_i = isl_map_set_tuple_name(schedule_i,
						    isl_dim_in, name);
		schedule = isl_union_map_add_map(schedule, schedule_i);
	}

	p = isl_printer_to_file(ctx, stdout);

	p = isl_printer_print_str(p, "D := ");
	p = isl_printer_print_union_set(p, domain);
	p = isl_printer_print_str(p, ";");
	p = isl_printer_end_line(p);

	p = isl_printer_print_str(p, "R := ");
	p = isl_printer_print_union_map(p, read);
	p = isl_printer_print_str(p, " * D;");
	p = isl_printer_end_line(p);

	p = isl_printer_print_str(p, "W := ");
	p = isl_printer_print_union_map(p, write);
	p = isl_printer_print_str(p, " * D;");
	p = isl_printer_end_line(p);

	p = isl_printer_print_str(p, "S := ");
	p = isl_printer_print_union_map(p, schedule);
	p = isl_printer_print_str(p, ";");
	p = isl_printer_end_line(p);

	isl_printer_free(p);

	isl_union_set_free(domain);
	isl_union_map_free(read);
	isl_union_map_free(write);
	isl_union_map_free(schedule);

	pdg->free();
	delete pdg;
	isl_ctx_free(ctx);

	return 0;
}
