#include <assert.h>
#include <stdio.h>
#include <yaml.h>
#include <isl/arg.h>
#include <isl/union_set.h>

struct options {
	char *eqv1;
	char *eqv2;
};

ISL_ARGS_START(struct options, options_args)
ISL_ARG_ARG(struct options, eqv1, "eqv1", NULL)
ISL_ARG_ARG(struct options, eqv2, "eqv2", NULL)
ISL_ARGS_END

ISL_ARG_DEF(options, struct options, options_args)

struct eqv {
	isl_union_set *only1;
	isl_union_set *only2;
	isl_union_set *proved;
	isl_union_set *not_proved;
};

static void *eqv_free(struct eqv *eqv)
{
	if (!eqv)
		return NULL;

	isl_union_set_free(eqv->only1);
	isl_union_set_free(eqv->only2);
	isl_union_set_free(eqv->proved);
	isl_union_set_free(eqv->not_proved);

	free(eqv);
	return NULL;
}

static __isl_give isl_union_set *extract_set(isl_ctx *ctx,
	yaml_document_t *document, yaml_node_t *node)
{
	if (node->type != YAML_SCALAR_NODE)
		isl_die(ctx, isl_error_invalid, "expecting scalar node",
			return NULL);

	return isl_union_set_read_from_str(ctx,
					    (char *) node->data.scalar.value);
}

static struct eqv *extract_eqv(isl_ctx *ctx, yaml_document_t *document,
	yaml_node_t *node)
{
	struct eqv *eqv;
	yaml_node_pair_t * pair;

	if (!node)
		return NULL;

	if (node->type != YAML_MAPPING_NODE)
		isl_die(ctx, isl_error_invalid, "expecting mapping",
			return NULL);

	eqv = isl_calloc_type(ctx, struct eqv);
	if (!eqv)
		return NULL;

	for (pair = node->data.mapping.pairs.start;
	     pair < node->data.mapping.pairs.top; ++pair) {
		yaml_node_t *key, *value;

		key = yaml_document_get_node(document, pair->key);
		value = yaml_document_get_node(document, pair->value);

		if (key->type != YAML_SCALAR_NODE)
			isl_die(ctx, isl_error_invalid, "expecting scalar key",
				return eqv_free(eqv));
		if (!strcmp((char *) key->data.scalar.value,
				"Equivalence proved"))
			eqv->proved = extract_set(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value,
				"Equivalence NOT proved"))
			eqv->not_proved = extract_set(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value,
				"Only in program 1"))
			eqv->only1 = extract_set(ctx, document, value);
		if (!strcmp((char *) key->data.scalar.value,
				"Only in program 2"))
			eqv->only2 = extract_set(ctx, document, value);
	}

	return eqv;
}

static struct eqv *eqv_parse(isl_ctx *ctx, FILE *in)
{
	struct eqv *eqv = NULL;
	yaml_parser_t parser;
	yaml_node_t *root;
	yaml_document_t document = { 0 };

	yaml_parser_initialize(&parser);

	yaml_parser_set_input_file(&parser, in);

	if (!yaml_parser_load(&parser, &document))
		goto error;

	root = yaml_document_get_root_node(&document);

	eqv = extract_eqv(ctx, &document, root);

	yaml_document_delete(&document);

	yaml_parser_delete(&parser);

	return eqv;
error:
	yaml_parser_delete(&parser);
	eqv_free(eqv);
	return NULL;
}

static int eqv_is_equal(struct eqv *eqv1, struct eqv *eqv2)
{
	if (!eqv1 || !eqv2)
		return 0;
	
	if (!eqv1->only1 != !eqv2->only1)
		return 0;
	if (!isl_union_set_is_equal(eqv1->only1, eqv2->only1)) {
		isl_union_set_dump(eqv1->only1);
		isl_union_set_dump(eqv2->only1);
		return 0;
	}
	
	if (!eqv1->only2 != !eqv2->only2)
		return 0;
	if (!isl_union_set_is_equal(eqv1->only2, eqv2->only2)) {
		isl_union_set_dump(eqv1->only2);
		isl_union_set_dump(eqv2->only2);
		return 0;
	}
	
	if (!eqv1->proved != !eqv2->proved)
		return 0;
	if (!isl_union_set_is_equal(eqv1->proved, eqv2->proved)) {
		isl_union_set_dump(eqv1->proved);
		isl_union_set_dump(eqv2->proved);
		return 0;
	}
	
	if (!eqv1->not_proved != !eqv2->not_proved)
		return 0;
	if (!isl_union_set_is_equal(eqv1->not_proved, eqv2->not_proved)) {
		isl_union_set_dump(eqv1->not_proved);
		isl_union_set_dump(eqv2->not_proved);
		return 0;
	}

	return 1;
}

int main(int argc, char **argv)
{
	isl_ctx *ctx;
	struct options *options;
	struct eqv *eqv1, *eqv2;
	FILE *file1, *file2;
	int equal;

	options = options_new_with_defaults();
	assert(options);
	argc = options_parse(options, argc, argv, ISL_ARG_ALL);
	ctx = isl_ctx_alloc_with_options(&options_args, options);

	file1 = fopen(options->eqv1, "r");
	assert(file1);
	file2 = fopen(options->eqv2, "r");
	assert(file2);

	eqv1 = eqv_parse(ctx, file1);
	eqv2 = eqv_parse(ctx, file2);

	equal = eqv_is_equal(eqv1, eqv2);

	eqv_free(eqv1);
	eqv_free(eqv2);

	fclose(file2);
	fclose(file1);
	isl_ctx_free(ctx);

	return equal ? 0 : 1;
}
