#ifndef ISA_ADG_H
#define ISA_ADG_H

#include <stdio.h>
#include <vector>
#include <yaml.h>
#include <isl/id.h>
#include <isl/aff.h>
#include <isl/set.h>
#include <isl/map.h>

/* Represents a (possibly indexed) variable of a certain type.
 *
 * "expr" contains a pointer to an allocated buffer created by get_expr().
 */
struct adg_var {
	isl_pw_multi_aff *access;
	isl_id *type;

	adg_var() : access(NULL), type(NULL), expr(NULL) {}
	adg_var(const adg_var &var) : expr(NULL) {
		access = isl_pw_multi_aff_copy(var.access);
		type = isl_id_copy(var.type);
	}
	~adg_var() {
		isl_pw_multi_aff_free(access);
		isl_id_free(type);
		free(expr);
	}

	__isl_keep const char *get_expr();
	void print(FILE *out = stdout);
	int print(yaml_emitter_t *emitter);
	int is_equal(adg_var *var2);
private:
	char *expr;
};

/* Represents a named piecewise affine expression.
 */
struct adg_expr {
	isl_id *name;
	isl_pw_aff *expr;

	adg_expr() : name(NULL), expr(NULL) {}
	adg_expr(const adg_expr &copy) {
		name = isl_id_copy(copy.name);
		expr = isl_pw_aff_copy(copy.expr);
	}
	~adg_expr() {
		isl_id_free(name);
		isl_pw_aff_free(expr);
	}

	int print(yaml_emitter_t *emitter);
	int is_equal(adg_expr *expr2);
};

/* Represents a domain.  The set "bounds" should have no existentially
 * quantified variables.  Instead, the domain may be nested (possibly
 * multiple times) with extra control variables in the "local" output
 * of the nested map.  Each of these extra variables is represented
 * by an adg_expr in the controls vector.
 * Additionally, the domain may also be nested with extra filter
 * variables in the input of the nested map.  Each of these filters
 * is represented by an adg_var in the filters vector.
 * The different types of nestings may be interleaved.
 */
struct adg_domain {
	isl_set *bounds;
	std::vector<adg_expr *> controls;
	std::vector<adg_var *> filters;

	adg_domain() : bounds(NULL) {}
	adg_domain(__isl_take isl_set *bounds) : bounds(bounds) {}
	~adg_domain() {
		for (int i = 0; i < controls.size(); ++i)
			delete controls[i];
		for (int i = 0; i < filters.size(); ++i)
			delete filters[i];
		isl_set_free(bounds);
	}

	int print(yaml_emitter_t *emitter);
	int is_equal(adg_domain *domain2);
};

/* Represents a port with a name, belonging to a node called node_name
 * and connected to an edge called edge_name.
 * For each iteration in "domain" each of adg_vars in "vars"
 * is either read from the edge or written to the edge.
 *
 * edge_nr is the number of the (non-control) edge to which
 *	port is connected or related.
 * arg_pos reflects the argument position of the variable accessed
 *	in this port.
 */
struct adg_port {
	isl_id *name;
	isl_id *node_name;
	isl_id *edge_name;
	std::vector<adg_var *> vars;
	adg_domain *domain;
	int edge_nr;
	int arg_pos;

	adg_port() : name(NULL), node_name(NULL), edge_name(NULL),
		domain(NULL), edge_nr(-1), arg_pos(-1) {}
	~adg_port() {
		isl_id_free(name);
		isl_id_free(node_name);
		isl_id_free(edge_name);
		for (int i = 0; i < vars.size(); ++i)
			delete vars[i];
		delete domain;
	}

	int print(yaml_emitter_t *emitter);
	int is_equal(adg_port *port2);
};

/* Represents a gate called "name", belonging to a node calles "node_name"
 * that copies the value of "in" to "out" for each iteration in "domain".
 */
struct adg_gate {
	isl_id *name;
	isl_id *node_name;
	adg_var *in;
	adg_var *out;
	adg_domain *domain;

	adg_gate() : name(NULL), node_name(NULL),
		in(NULL), out(NULL), domain(NULL) {}
	~adg_gate() {
		isl_id_free(name);
		isl_id_free(node_name);
		delete in;
		delete out;
		delete domain;
	}

	int print(yaml_emitter_t *emitter);
	int is_equal(adg_gate *gate2);
};

enum adg_arg_type {
	adg_arg_value,
	adg_arg_reference,
	adg_arg_return,
	adg_arg_unknown
};

/* Represents an argument to a function.
 * It contains the variable that is read or written and the way
 * it is passed to/from the function.
 */
struct adg_arg {
	adg_var *var;
	enum adg_arg_type type;

	adg_arg() : var(NULL), type(adg_arg_unknown) {}
	~adg_arg() {
		delete var;
	}

	int print(yaml_emitter_t *emitter);
	int is_equal(adg_arg *arg2);
};

/* Represents the execution of a function called "name" for each
 * iteration of "domain".
 * The "ctrl" vector contains references to variables that should
 * be written on each iteration of the function.  The value that
 * should be written is given by the expr field of the adg_expr.
 */
struct adg_function {
	isl_id *name;
	adg_domain *domain;

	std::vector<adg_arg *> in;
	std::vector<adg_arg *> out;
	std::vector<adg_expr *> ctrl;

	adg_function() : name(NULL), domain(NULL) {}
	~adg_function() {
		isl_id_free(name);
		delete domain;
		for (int i = 0; i < in.size(); ++i)
			delete in[i];
		for (int i = 0; i < out.size(); ++i)
			delete out[i];
		for (int i = 0; i < ctrl.size(); ++i)
			delete ctrl[i];
	}

	int print(yaml_emitter_t *emitter);
	int is_equal(adg_function *function2);
};

enum adg_edge_type {
	adg_edge_fifo,
	adg_edge_sticky_fifo,
	adg_edge_shift_register,
	adg_edge_broadcast,
	adg_edge_generic,
	adg_edge_error
};

/* Represents an edge called "name" connecting the port "from_port_name"
 * in node "from_node_name" to the port "to_port_name" in node "to_node_name".
 *
 * "map" maps an iteration of the "to_port_name" to the corresponding
 * iteration of the "from_port_name", i.e., to the iteration that wrote
 * the values that is being read in the given iteration.
 *
 * If each parameter is assigned its "val" values, then a buffer size
 * of "value_size" on this edge is sufficient to allow for a deadlock
 * free execution.
 */
struct adg_edge {
	isl_id *name;
	isl_id *from_node_name;
	isl_id *from_port_name;
	isl_id *to_node_name;
	isl_id *to_port_name;

	enum adg_edge_type type;

	isl_multi_aff *map;
	isl_int value_size;

	adg_edge() : name(NULL), from_node_name(NULL), from_port_name(NULL),
		    to_node_name(NULL), to_port_name(NULL), map(NULL),
		    type(adg_edge_fifo) {
		isl_int_init(value_size);
		isl_int_set_si(value_size, -1);
	}
	~adg_edge() {
		isl_id_free(name);
		isl_id_free(from_node_name);
		isl_id_free(from_port_name);
		isl_id_free(to_node_name);
		isl_id_free(to_port_name);
		isl_multi_aff_free(map);
		isl_int_clear(value_size);
	}

	int print(yaml_emitter_t *emitter);
	int is_equal(adg_edge *edge2);
};

/* Represents a node called "name" with input ports, gates, expressions,
 * output ports and function.
 *
 * "domain" is the iteration domain.
 * "schedule" is the schedule for this node.
 * "lifting" is a mapping that has been used to remove
 * existentially quantified from the domain of this node and that
 * should be applied to the domains of the subelements to ensure
 * that they have the same initial (static) control variables.
 */
struct adg_node {
	isl_id *name;

	std::vector<adg_port *> input_ports;
	std::vector<adg_gate *> gates;
	std::vector<adg_expr *> expressions;
	std::vector<adg_port *> output_ports;

	adg_function *function;

	adg_domain *domain;
	isl_map *schedule;
	isl_basic_map *lifting;

	adg_node() : name(NULL), function(NULL), domain(NULL), schedule(NULL),
		lifting(NULL) {}
	~adg_node() {
		isl_id_free(name);
		for (int i = 0; i < input_ports.size(); ++i)
			delete input_ports[i];
		for (int i = 0; i < gates.size(); ++i)
			delete gates[i];
		for (int i = 0; i < expressions.size(); ++i)
			delete expressions[i];
		for (int i = 0; i < output_ports.size(); ++i)
			delete output_ports[i];
		if (function)
			delete function;
		isl_map_free(schedule);
		isl_basic_map_free(lifting);
		delete domain;
	}
	int print(yaml_emitter_t *emitter);
	int is_equal(adg_node *node2);
};

/* Represents a parameters called "name" with lower bound "lb", upper
 * bound "ub" and default value "val".
 */
struct adg_param {
	isl_id *name;
	isl_aff *lb;
	isl_aff *ub;
	isl_aff *val;

	adg_param() : name(NULL), lb(NULL), ub(NULL), val(NULL) {}
	~adg_param() {
		isl_id_free(name);
		isl_aff_free(lb);
		isl_aff_free(ub);
		isl_aff_free(val);
	}

	int print(yaml_emitter_t *emitter);
	int is_equal(adg_param *param2);
};

/* Represents an approximate dependence graph called "name", with parameters,
 * nodes and edges.
 *
 * "context" represents the constraints on the parameters.
 * "iterators" represents the iterators used in the domains of the adg.
 * "all_iterators" represents the iterators used in the schedules of the adg.
 * "iterator_map" maps (some) elements of "all_iterators" to the
 * corresponding element in "iterators".
 * "controls" collects all static control variables so that we can
 *	reuse them in to_control.
 */
struct adg {
	isl_id *name;
	isl_set *context;
	isl_map *iterator_map;

	std::vector<adg_param *> params;
	std::vector<adg_node *> nodes;
	std::vector<adg_edge *> edges;
	std::vector<isl_id *> iterators;
	std::vector<isl_id *> all_iterators;
	std::vector<adg_expr *> controls;

	adg() : name(NULL), context(NULL), iterator_map(NULL) {}
	~adg() {
		isl_id_free(name);
		isl_set_free(context);
		isl_map_free(iterator_map);
		for (int i = 0; i < params.size(); ++i)
			delete params[i];
		for (int i = 0; i < nodes.size(); ++i)
			delete nodes[i];
		for (int i = 0; i < edges.size(); ++i)
			delete edges[i];
		for (int i = 0; i < all_iterators.size(); ++i)
			isl_id_free(all_iterators[i]);
		for (int i = 0; i < iterators.size(); ++i)
			isl_id_free(iterators[i]);
		for (int i = 0; i < controls.size(); ++i)
			delete controls[i];
	}
	adg_node *node_by_id(__isl_keep isl_id *id);
	void print(FILE *out = stdout);
	int print(yaml_emitter_t *emitter);
	int is_equal(adg *adg2);
	adg_expr *to_control(__isl_take isl_aff *expr);
};

#endif
