#ifndef PDG_H
#define PDG_H

#include <isa/yaml.h>
#include <gmp.h>

#include <isl/ctx.h>
#include <isl/set.h>
#include <isl/map.h>

using namespace yaml;

namespace pdg
{

class Matrix : public structure {
    static serialize *create(void *user) { return new Matrix(); }
    static serialize *convert(anchor_map *, serialize *, const type_info *);

public:
    std::vector< std::vector<int> > el;
    static void register_type();
    void init(SyckParser *p, SyckNode *n);
    void dump();
    void dump(emitter& e);
    Matrix() {};
    isl_basic_set *get_isl_basic_set(isl_space *dim) const;
    isl_basic_map *get_isl_basic_map(isl_space *dim) const;
};

class parameter : public structure {
    static serialize *create(void *user) { return new parameter(); }
public:
    int	    id;
    str	    *name;
    integer *value;

    static void register_type();
    parameter () : id(-1), name(NULL), value(NULL) {}
};

class IslSet : public structure {
	static serialize *create(void *user) { return new IslSet(); }
	static serialize *convert(anchor_map *, serialize *, const type_info *);
public:
	isl_set *set;

	IslSet() : set(NULL) {}
	IslSet(__isl_take isl_set *set) : set(set) {}
	~IslSet() { isl_set_free(set); }
	__isl_give isl_set *get_isl_set(void *unused = NULL) const {
		return isl_set_copy(set);
	}

	int dim() const {
		return isl_set_dim(set, isl_dim_set);
	}
	__isl_give isl_space *get_dim() const {
		return isl_set_get_space(set);
	}

	static void register_type();
	void dump(emitter &e);
};

class IslMap : public structure {
	static serialize *create(void *user) { return new IslMap(); }
	static serialize *convert(anchor_map *, serialize *, const type_info *);
public:
	isl_map *map;

	IslMap() : map(NULL) {}
	IslMap(__isl_take isl_map *map) : map(map) {}
	~IslMap() { isl_map_free(map); }
	__isl_give isl_map *get_isl_map(void *unused = NULL) const {
		return isl_map_copy(map);
	}

	int input() const {
		return isl_map_dim(map, isl_dim_in);
	}
	int output() const {
		return isl_map_dim(map, isl_dim_out);
	}
	__isl_give isl_space *get_dim() const {
		return isl_map_get_space(map);
	}

	static void register_type();
	void dump(emitter &e);
};

void isl_dim_set_parameter_names(isl_space *dim, const seq<parameter>& params);
isl_set *isl_set_set_parameter_names(isl_set *set, const seq<parameter>& params);

class UnionSet : public structure {
    static serialize *create(void *user) { return new UnionSet(); }

public:
    int	    	    dim;
    seq<Matrix>	    constraints;
    seq<parameter>  params;

    static void register_type();
    void dump(emitter& e);
    UnionSet () : dim(-1) {}
    UnionSet(isl_set *set, seq<parameter>& p);
    isl_set *get_isl_set(isl_ctx *ctx) const;
    void fill_fields(anchor_map *);
};

class UnionMap : public structure {
    static serialize *create(void *user) { return new UnionMap(); }

public:
    int	    input;
    int	    output;
    seq<Matrix>	constraints;
    seq<parameter>  params;

    static void register_type();
    void dump(emitter& e);
    UnionMap (int in = -1, int out = -1) 
	: input(in), output(out) {}
    UnionMap(isl_map *map, seq<parameter>& p);
    isl_map *get_isl_map(isl_ctx *ctx) const;
    void fill_fields(anchor_map *);
};

class dependence_type;

/* if uniquely_defined is set then the array is written by a single access
 * such that any element that is ever read
 * is known to be assigned exactly once before the read
 */
class array : public structure {
    static serialize *create(void *user) { return new array(); }
public:
    int	    id;
    str	    *name;
    seq<int> dims;
    str	    *element_type;
    enum type {
	unset = -1,
	input,
	output,
	temp,
	last
    } type;
    seq<dependence_type>	analysis_performed;
    IslSet			*value_bounds;
    int				uniquely_defined;

    static void register_type();
    array () : id(-1), name(NULL), element_type(NULL), type(unset),
		value_bounds(NULL), uniquely_defined(0) {}
};

class access;
class function_call;

class call_or_access : public structure {
    static serialize *create(void *user) { return new call_or_access(); }
public:
    enum type {
	unset = -1,
	t_access,
	t_call,
	last,
    } type;
    pdg::access		*access;
    pdg::function_call	*call;

    static void register_type();
    call_or_access() : type(unset), access(NULL), call(NULL) {}
};

class access : public structure {
    static serialize *create(void *user) { return new access(); }
public:
    pdg::array *	array;
    IslMap		*map;
    enum type {
	unset = -1,
	write,
	read,
	last,
    } type;
    int		nr;
    /* if not NULL, maps domain of node to extended domain,
     * such that the number of extra dimensions is equal to the number
     * of dimensions missing from the range of "map".
     */
    IslMap		*extension;
    /* If not NULL, maps extended domain to array domain.
     */
    IslMap		*extended_map;
    /* A list of function calls of accesses nested inside the index
     * expression of this access.  The final input dimensions of
     * the access map correspond to the values attained by these
     * nested constructs.
     */
    seq<call_or_access>	 nested;
    /* If non-empty, sources contains exact and complete dependences
     * from wrapped write accesses to a wrapped copy of this read access.
     */
    seq<IslMap>		 sources;

    static void register_type();
    access () : array(NULL), type(unset), map(NULL), nr(-1), extension(NULL),
	extended_map(NULL) {}

    __isl_give isl_map *extract_access_map(void);
};


class function_call : public structure {
    static serialize *create(void *user) { return new function_call(); }
public:
    str			*name;
    seq<call_or_access>	 arguments;

    static void register_type();
    function_call() : name(NULL) {}
};

class statement : public structure {
    static serialize *create(void *user) { return new statement(); }
public:
    seq<access> accesses;
    int	    	operation;
    int		line;
    function_call	*top_function;
    seq<access>		 top_outputs;

    static void register_type();
    statement () : operation(-1), top_function(NULL) {}
};

class dependence;

/* "source" is the original iteration domain.  It is a wrapped map
 *	if filters contains any elements.  In particular, it maps
 *	the iteration domain to the values of the filters
 *	for which the node is executed.
 * "filter_depth" represents the number of outer loops whose
 *	last iteration depends on filters.  It caches the value
 *	computed by get_filter_depth() and is initialized to -1.
 */
class node : public structure {
    static serialize *create(void *user) { return new node(); }
public:
    int	    	nr;
    str		*name;
    IslSet   	*source;
    IslMap  	*schedule;
    pdg::statement	*statement;
    seq<int>		 prefix;
    seq<call_or_access>	 filters;
    int			 filter_depth;

    static void register_type();
    node() : nr(-1), name(NULL), source(NULL), statement(NULL),
		schedule(NULL), filter_depth(-1) {}
    int get_filter_depth();
};

class enumerator : public structure {
    static serialize *create(void *user) { return new enumerator(); }
    static serialize *convert(anchor_map *, serialize *, const type_info *);
public:
    int		value;
    string     *s;

    static void register_type();
    void init(SyckParser *p, SyckNode *n);
    void dump(emitter& e);
    void operator= (int i);
    enumerator() : value(-1), s(NULL) {}
    enumerator(int v) : value(v), s(NULL) {}
    enumerator(string *s) : value(-1), s(s) {}
    ~enumerator();
};

/* If "controlled_relation" is not NULL, then it contains the exact
 * dependence relation parametrized on the last iterations of one or
 * more sources.
 * "relation" is then the result of projecting out these parameters.
 *
 * "from_controls" and "to_controls" contain the names of the control variables
 *	for control dependences.  The first sequence contrains the names
 *	of those variables in the from_access while the second contains
 *	the names in the to_access.  The names may be different in case
 *	of reuse.  In particular, the last iteration of an access F before
 *	access A may be equal to the last iteration of F before some
 *	other access B.  The values of these control variables can therefore
 *	be reused (i.e., transmitted over this dependence), but the names
 *	of the control variables may be different on either end.
 */
class dependence : public structure {
    static serialize *create(void *user) { return new dependence(); }
public:
    IslMap *relation;
    IslMap *extended_relation;
    IslMap *controlled_relation;
    node    *from;
    node    *to;
    access    *from_access;
    access    *to_access;
    enum type {
	unset = -1,
	flow,
	anti,
	pn,
	pn_union,
	pn_part,
	pn_wire,
	pn_hold,
	pn_shift,
	reuse,
	reuse_pair,
	output,
	uninitialized,
	last,
    } type;
    pdg::array *array;
    seq<str>	from_controls;
    seq<str>	to_controls;
    int		reordering;
    int		multiplicity;
    pdg::enumerator *	size;
    pdg::dependence *	container;
    integer *	value_size;

    static void register_type();
    dependence () : relation(NULL), extended_relation(NULL),
		    controlled_relation(NULL),
		    from(NULL), to(NULL), 
		    from_access(NULL), to_access(NULL), type(unset),
		    array(NULL), reordering(-1), multiplicity(-1),
		    size(NULL), container(NULL),
		    value_size(NULL) {}
};

class dependence_type : public structure {
public:
    static serialize *create(void *user) { return new dependence_type(); }
    dependence_type() : type(pdg::dependence::unset) {}
    dependence_type(enum pdg::dependence::type type) : type(type) {}
    enum pdg::dependence::type	type;
};

class history_entry : public structure {
    static serialize *create(void *user) { return new history_entry(); }
public:
    str	    *tool;
    str	    *version;
    seq<str>	arguments;

    static void register_type();
    history_entry(): tool(NULL), version(NULL) {}
};

class PDG : public structure {
    static serialize *create(void *user) { return new PDG((isl_ctx *)user); }
public:
    int	    dimension;
    int	    cacheline;
    int	    root;
    str	    *placement;
    seq<node>  nodes;
    seq<array> arrays;
    seq<dependence> dependences;
    seq<parameter>  params;
    IslSet   	*context;
    seq<history_entry>	history;
    str		*name;
    seq<int>	 statement_dimensions;

    /* Not serialized */
    isl_ctx	*ctx;
    bool	 ctx_created;

    static PDG *Load(char *str, isl_ctx *ctx);
    static PDG *Load(FILE *fp, isl_ctx *ctx);
    static void register_type();
    PDG(isl_ctx *ctx) : dimension(0), root(-1), cacheline(-1),
		placement(NULL), context(NULL), name(NULL), ctx(ctx),
		ctx_created(false) {}

    ~PDG();

    isl_ctx *get_isl_ctx();
    isl_set *get_context_isl_set();
    void add_history_line(const char *tool, int argc, char * argv[]);
};

__isl_give isl_map *schedule_dependence(PDG *pdg, pdg::dependence *dep,
	__isl_take isl_map *map);
isl_map *scatter_dependence(PDG *pdg, pdg::dependence *dep);

class HMatrix : public structure {
    static serialize *create(void *user) { return new Matrix(); }

public:
    static void register_type();
    void init(SyckParser *p, SyckNode *n) {};
};

}

#endif
