#include "isl_util.h"
#include <isl/set.h>
#include <isl/map.h>

/* For any two pairs of corresponding writes and reads,
 * do the writes occur in a different order than the reads ?
 * If so, then the dependence is "reordering".
 */
int isl_map_is_reordering(struct isl_map *dep)
{
	isl_map *read_before;
	isl_map *write_after;
	int r;

	read_before = isl_map_lex_lt(isl_space_range(isl_map_get_space(dep)));
	write_after = isl_map_lex_gt(isl_space_domain(isl_map_get_space(dep)));
	write_after = isl_map_apply_domain(write_after, isl_map_copy(dep));
	write_after = isl_map_apply_range(write_after, isl_map_copy(dep));
	write_after = isl_map_intersect(write_after, read_before);

	r = !isl_map_is_empty(write_after);

	isl_map_free(write_after);

	return r;
}

/* For reads from dep1 and dep2 from the same iteration,
 * does the corresponding write from dep1 occur before the write from dep2 ?
 *
 * The answer is no if we can find a write from dep1 that occurs
 * after a write from dep2 such that both writes correspond to the same read.
 */
int isl_is_inorder_accesses(isl_map *dep1, isl_map *dep2)
{
	isl_space *dim;
	isl_map *read_same;
	isl_map *write_after;
	int in_order;

	dim = isl_space_map_from_set(isl_space_range(isl_map_get_space(dep1)));
	read_same = isl_map_identity(dim);
	write_after = isl_map_lex_gt(isl_space_domain(isl_map_get_space(dep1)));
	write_after = isl_map_apply_domain(write_after, isl_map_copy(dep1));
	write_after = isl_map_apply_range(write_after, isl_map_copy(dep2));
	write_after = isl_map_intersect(write_after, read_same);

	in_order = isl_map_is_empty(write_after);

	isl_map_free(write_after);

	return in_order;
}

/* For any pair of corresponding write and read,
 * is there any other read in between ?
 * If so, then the loop is not "consecutive".
 * If two reads occur "simultaneously", we also do not consider 
 * the loop to be consecutive.
 */
int isl_is_consecutive_loop(isl_map *dep)
{
	isl_map *write_before;
	isl_map *read_after;
	int consecutive;

	/* Check for simultaneous reads */
	if (!isl_is_inorder_accesses(dep, dep))
		return 0;

	/* Check for read strictly in between write and read */
	write_before = isl_map_lex_lt(isl_space_range(isl_map_get_space(dep)));
	read_after = isl_map_lex_gt(isl_space_range(isl_map_get_space(dep)));
	write_before = isl_map_intersect_range(write_before,
				isl_map_range(isl_map_copy(dep)));
	write_before = isl_map_apply_domain(write_before, isl_map_copy(dep));
	write_before = isl_map_intersect(write_before, read_after);

	consecutive = isl_map_is_empty(write_before);

	isl_map_free(write_before);

	return consecutive;
}

/* For any pair of corresponding write and read of dep2,
 * is there any read from dep1 in between ?
 * If so, then dep2 is not "consecutive".
 * If two reads occur "simultaneously", we also do not consider 
 * the loop to be consecutive.
 */
int isl_is_consecutive_dep(isl_map *dep2, isl_map *dep1)
{
	isl_map *write_before;
	isl_map *read_after;
	int consecutive;

	write_before = isl_map_lex_lt(isl_space_range(isl_map_get_space(dep2)));
	read_after = isl_map_lex_ge(isl_space_range(isl_map_get_space(dep2)));
	write_before = isl_map_intersect_range(write_before,
				isl_map_range(isl_map_copy(dep1)));
	write_before = isl_map_apply_domain(write_before, isl_map_copy(dep2));
	write_before = isl_map_intersect(write_before, read_after);

	consecutive = isl_map_is_empty(write_before);

	isl_map_free(write_before);

	return consecutive;
}

/* Check whether there is a pair of distinct reads, one from dep1 and dep2,
 * that correspond to the same write.
 */
int isl_map_has_multiplicity(struct isl_map *dep)
{
	isl_space *dim;
	isl_map *read_before;
	isl_map *write_same;
	int mult;

	read_before = isl_map_lex_lt(isl_space_range(isl_map_get_space(dep)));
	dim = isl_space_map_from_set(isl_space_domain(isl_map_get_space(dep)));
	write_same = isl_map_identity(dim);
	write_same = isl_map_apply_domain(write_same, isl_map_copy(dep));
	write_same = isl_map_apply_range(write_same, isl_map_copy(dep));
	write_same = isl_map_intersect(write_same, read_before);

	mult = !isl_map_is_empty(write_same);

	isl_map_free(write_same);

	return mult;
}
