#include "suif_inc.h"
#include "for_normalize.h"
#include "domain.h"
#include "dump.h"

#define ARRAY_SIZE(a) (sizeof(a)/sizeof(*a))

static char **funcs;
static int verbose;
static int oldstyle;
static cmd_line_option my_options[] = {
	{ CLO_MULTI_STRING,   	"-func",	"", 	&funcs },
	{ CLO_NOARG,   		"-verbose",	"", 	&verbose },
	{ CLO_NOARG,   		"-oldstyle",	"", 	&oldstyle },
};

static int m_argc;
static char **m_argv;

static void extract_context(tree_instr *ti, context_info *ci)
{
    instruction *in = ti->instr();
    operand cpy = in->src_op(0);
    assert(cpy.is_expr());
    operand op = cpy.instr()->src_op(0);

    guard_constructor gc(ci->sa, 1, NULL);
    assert(op.kind() == OPER_INSTR);
    instruction * ins = op.instr();
    gc.construct_guard(ins);

    ci->ctx = new SRelation(*gc.s, ci->sa);
}

static context_info *handle_context(annote *ca, context_info *ctx)
{
    immed_list *ims = ca->immeds();
    if (ims->count() != 2)
	return ctx;
    immed val = (*ims)[1];
    if (val.kind() != im_symbol)
	return ctx;
    sym_node *sym = val.symbol();
    if (sym->kind() != SYM_PROC)
	return ctx;

    assert(ctx->ctx == NULL);

    proc_sym *proc = static_cast<proc_sym *>(sym);
    if (!proc->is_in_memory())
	proc->read_proc();
    ctx->tp = proc->block();
    tree_node_list *tnl = ctx->tp->body();
    tree_node_list_iter tnli(tnl);
    for (int i = 0; !tnli.is_empty(); ++i) {
	tree_node *tn = tnli.step();
	switch (tn->kind()) {
	case TREE_INSTR: {
	    tree_instr *in = (tree_instr*)tn;
	    switch (in->instr()->opcode()) {
	    case io_ret:
		extract_context(in, ctx);
		break;
	    case io_mrk:
		break;
	    default:
		assert(0);
	    }
	    break;
	}
	default:
	    assert(0);
	}
	if (ctx->ctx)
	    break;
    }
    return ctx;
}

static context_info *handle_parameter(annote *an, context_info *ctx)
{
    immed_list *ims = an->immeds();
    if (ims->count() != 3 && ims->count() != 4)
	return ctx;
    bool have_ub = ims->count() == 4;
    immed val = (*ims)[1];
    if (val.kind() != im_symbol)
	return ctx;
    sym_node *sym = val.symbol();
    if (sym->kind() != SYM_VAR)
	return ctx;

    if ((*ims)[2].kind() != im_int)
	return ctx;
    int lb = (*ims)[2].integer();

    if (have_ub && (*ims)[3].kind() != im_int)
	return ctx;

    var_sym* var = static_cast<var_sym*>(sym);
    ctx->sa->add_var(var);
    ctx->sa->vars[var]->lb = lb;
    ctx->sa->vars[var]->lb_set = true;
    Relation r(0);
    GEQ_Handle e1 = r.and_with_GEQ();
    e1.update_coef(r.get_local(ctx->sa->vars[var]), 1);
    e1.update_const(-lb);
    if (have_ub) {
	int ub = (*ims)[3].integer();
	GEQ_Handle e2 = r.and_with_GEQ();
	e2.update_coef(r.get_local(ctx->sa->vars[var]), -1);
	e2.update_const(ub);
    }
    SRelation *sr = new SRelation(r, ctx->sa);
    if (!ctx->ctx)
	ctx->ctx = sr;
    else {
	SRelation *r = ctx->ctx;
	ctx->ctx = r->Intersection(sr);
	delete sr;
	delete r;
    }

    return ctx;
}

static context_info *handle_pragmas(tree_proc *tp)
{
    context_info *ctx = new context_info;
    const char *k_C_pragma = lexicon->enter("C pragma")->sp;;
    annote_list_iter anli(tp->proc()->file()->annotes());
    while (!anli.is_empty()) {
	annote *an = anli.step();
	if (an->name() != k_C_pragma)
	    continue;
	immed_list *ims = an->immeds();
	if (!(*ims)[0].is_string())
	    continue;
	if (strcmp((*ims)[0].string(), "context") == 0)
	    ctx = handle_context(an, ctx);
	if (strcmp((*ims)[0].string(), "parameter") == 0)
	    ctx = handle_parameter(an, ctx);
    }
    return ctx;
}

/* Ensure for loops have step +1.
 *
 * We use a modified version of SUIF's normalize_fors_on_proc,
 * since we don't mind that the lower bound may not be zero.
 *
 * The normalization will introduce an assignment to the old
 * iterator.  We have to propagate this assignment to all
 * the accesses so that they no longer reference the old iterator.
 * Afterward, this assignment needs to be removed, but we can't
 * just use proc_dead_code since that will remove all dead code,
 * including assignments to scalars in a program that we want
 * to analyze.
 */
static void normalize_for_loops(tree_proc *tp)
{
	instruction_list *normalized = normalized_fors_on_proc(tp);
	if (!normalized)
		return;
	forward_propagate(tp, FPK_ALL);
	fold_all_constants(tp);
	while (!normalized->is_empty()) {
		instruction *ass = normalized->pop();
		tree_instr *parent = ass->parent();
		parent->parent()->remove(parent->list_e());
		delete parent->list_e();
		delete parent;
	}
	delete normalized;
}

void do_proc(tree_proc * tp)
{
    static context_info *ctx = NULL;

    if (!ctx)
	ctx = handle_pragmas(tp);

    char **f;
    proc_sym * psym = tp->proc();
    //printf("=======%s======= \n", psym->name());
    for (f = funcs; **f; ++f) 
	if (!strcmp(*f, psym->name()))
	    break;
    if (f != funcs && **f == '\0')
	return;
    if (ctx && ctx->tp == tp)
	return;
    normalize_for_loops(tp);
    fill_in_access(tp);
    construct_domains(tp, verbose, oldstyle, ctx);
    //tp->print();
    dump(tp, ctx, m_argc, m_argv);
}

int main(int argc, char * argv[])
{
    m_argc = argc;
    m_argv = new char* [argc+1];
    for (int i = 0; i < argc; ++i)
	m_argv[i] = argv[i];
    m_argv[argc] = NULL;
    start_suif(argc, argv);
    init_dead_code();
    init_unused();
    funcs = new char *[argc];
    parse_cmd_line(argc, argv, my_options, ARRAY_SIZE(my_options));
    init(argc, argv);
    suif_proc_iter(argc, argv, do_proc, TRUE);
    delete [] funcs;
    delete [] m_argv;
    return 0;
}
