#include <barvinok/isl.h>
#include <isa/yaml.h>
#include <isa/pdg.h>
#include "da.h"
#include <cloog/isl/cloog.h>
#include "clast_scanner.h"
#include "mem_options.h"

#define ALLOC(type) (type*)malloc(sizeof(type))

using pdg::PDG;
using namespace da;

static void dump_program(PDG *pdg, int *output, struct options *options);
static void scan_program(PDG *pdg, const char *inputsize, int maxdim,
			 struct clast_stmt *stmt, CloogOptions *options);

int main(int argc, char * argv[])
{
    PDG *pdg;
    isl_ctx *ctx = isl_ctx_alloc();
    struct options *options = options_new_with_defaults();

    argc = options_parse(options, argc, argv, ISL_ARG_ALL);

    pdg = PDG::Load(stdin, ctx);
    int output[pdg->arrays.size()];

    for (int i = 0; i < pdg->arrays.size(); ++i) {
	int n = pdg->dependences.size();
	find_deps(pdg, pdg->arrays[i], data_reuse);
	output[i] = n == pdg->dependences.size();
    }

    dump_program(pdg, output, options);
    pdg->free();
    delete pdg;
    options_free(options);
    isl_ctx_free(ctx);

    return 0;
}

static CloogDomain *scatter(__isl_take isl_set *domain,
	pdg::node *node, pdg::access *ac)
{
    int fulldim = node->prefix.size() + 1;

    for (int i = 0; i < node->prefix.size(); ++i) {
	if (node->prefix[i] == -1)
	    continue;
	domain = isl_set_insert_dims(domain, isl_dim_set, i, 1);
	domain = isl_set_fix_si(domain, isl_dim_set, i, node->prefix[i]);
    }
    domain = isl_set_insert_dims(domain, isl_dim_set, fulldim - 1, 1);
    domain = isl_set_fix_si(domain, isl_dim_set, fulldim - 1, ac->nr);

    return cloog_domain_from_isl_set(domain);
}

/* Compute the number of array elements that are read but not
 * (previously) written.  These are assumed to be the input.
 */
static __isl_give isl_pw_qpolynomial *compute_input_size(PDG *pdg,
	struct options *options)
{
	unsigned nparam = pdg->params.size();
	isl_space *dim;
	isl_pw_qpolynomial *inputsize;

	dim = isl_space_set_alloc(pdg->get_isl_ctx(), nparam, 0);
	isl_dim_set_parameter_names(dim, pdg->params);
	inputsize = isl_pw_qpolynomial_zero(dim);

	if (!options->with_input)
		return inputsize;

	for (int i = 0; i < pdg->dependences.size(); ++i) {
		isl_map *map;
		isl_pw_qpolynomial *size;
		pdg::dependence *dep = pdg->dependences[i];

		if (dep->type != pdg::dependence::uninitialized)
			continue;

		map = dep->relation->get_isl_map(pdg->get_isl_ctx());
		size = isl_map_card(map);
		inputsize = isl_pw_qpolynomial_add(inputsize, size);
	}

	return inputsize;
}

static void dump_program(PDG *pdg, int *output, struct options *options)
{
    int nparam = pdg->params.size();
    isl_ctx *ctx = pdg->get_isl_ctx();
    isl_set *context = pdg->get_context_isl_set();
    CloogState *state = cloog_isl_state_malloc(ctx);
    CloogOptions *cloog_options = cloog_options_malloc(state);
    CloogDomain *cloog_context, *domain;
    CloogUnionDomain *ud;
    CloogInput *input;
    struct clast_stmt *stmt;
    int maxdim = 0;
    isl_pw_qpolynomial *inputsize;
    isl_printer *printer;
    char *s_inputsize;
    char name[MAX_NAME];

    cloog_options->compilable = 1;
    cloog_context = cloog_domain_from_isl_set(isl_set_copy(context));

    inputsize = compute_input_size(pdg, options);
    inputsize = isl_pw_qpolynomial_gist(inputsize, context);
    printer = isl_printer_to_str(pdg->get_isl_ctx());
    printer = isl_printer_set_output_format(printer, ISL_FORMAT_C);
    printer = isl_printer_print_pw_qpolynomial(printer, inputsize);
    s_inputsize = isl_printer_get_str(printer);
    isl_printer_free(printer);
    isl_pw_qpolynomial_free(inputsize);

    ud = cloog_union_domain_alloc(nparam);

    for (int i = 0; i < pdg->nodes.size(); ++i) {
	pdg::node *node = pdg->nodes[i];
	pdg::statement *s = pdg->nodes[i]->statement;

	if (node->prefix.size() + 1 > maxdim)
	    maxdim = node->prefix.size() + 1;
	for (int j = 0; j < pdg->dependences.size(); ++j) {
	    pdg::dependence *dep = pdg->dependences[j];
	    if (dep->to != pdg->nodes[i] && dep->from != pdg->nodes[i])
		continue;
	    if (!options->with_input &&
		    dep->type == pdg::dependence::uninitialized)
		continue;
	    isl_map *rel = dep->relation->get_isl_map(pdg->get_isl_ctx());
	    if (dep->to == pdg->nodes[i]) {
		    isl_set *read = isl_map_range(isl_map_copy(rel));
		    domain = scatter(read, dep->to, dep->to_access);
		    ud = cloog_union_domain_add_domain(ud, NULL, domain, NULL,
							NULL);
	    }
	    if (dep->from == pdg->nodes[i]) {
		    isl_set *write = isl_map_domain(isl_map_copy(rel));
		    domain = scatter(write, dep->from, dep->from_access);
		    ud = cloog_union_domain_add_domain(ud, NULL, domain, NULL,
							(void *)1);
	    }
	    isl_map_free(rel);
	}
	for (int j = 0; j < s->accesses.size(); ++j) {
	    pdg::access *access = s->accesses[j];
	    if (access->type == pdg::access::read)
		continue;
	    int k;
	    for (k = 0; k < pdg->arrays.size(); ++k)
		if (pdg->arrays[k] == access->array)
		    break;
	    if (!output[k])
		continue;
	    isl_set *source = node->source->get_isl_set(pdg->get_isl_ctx());
	    domain = scatter(source, node, access);
	    ud = cloog_union_domain_add_domain(ud, NULL, domain, NULL, (void *)1);
	}
    }

    for (int i = 0; i < nparam; ++i) {
	const char *s = strdup(pdg->params[i]->name->s.c_str());
	ud = cloog_union_domain_set_name(ud, CLOOG_PARAM, i, s);
    }
    for (int i = 0; i < maxdim; ++i) {
	sprintf(name, "i%d", i);
	ud = cloog_union_domain_set_name(ud, CLOOG_ITER, i, name);
    }

    input = cloog_input_alloc(cloog_context, ud);
    stmt = cloog_clast_create_from_input(input, cloog_options);

    scan_program(pdg, s_inputsize, maxdim, stmt, cloog_options);
    cloog_clast_free(stmt);
    cloog_options_free(cloog_options);
    cloog_state_free(state);
    free(s_inputsize);
}

struct mem_counter : public clast_scanner {
    char tmpfile[L_tmpnam];
    char buffer[L_tmpnam+40];

    int count();
    void print_user(clast_user_stmt *s);
    void print_full(PDG *pdg, const char *inputsize, int maxdim, clast_stmt *s);
};

void mem_counter::print_full(PDG *pdg, const char *inputsize,
			     int maxdim, clast_stmt *s)
{
    dst = stdout;
    print_header();
    fprintf(dst, "void main() {\n");
    indent += 4;
    print_indent();
    if (maxdim <= 0)
	return;
    fprintf(dst, "int ");
    for (int i = 0; i < maxdim; ++i) {
	if (i)
	    fprintf(dst, ", ");
	fprintf(dst, "i%d", i);
    }
    fprintf(dst, ";\n");
    print_indent();
    fprintf(dst, "long count = %s;\n", inputsize);
    print_indent();
    fprintf(dst, "long max = count;\n");

    print(s);

    print_indent();
    fprintf(dst, "printf(\"final: %%ld\\n\", count);\n");
    fprintf(dst, "printf(\"max: %%ld\\n\", max);\n");
    fprintf(dst, "}\n");
}

void mem_counter::print_user(clast_user_stmt *u)
{
    print_indent();
    if (u->statement->usr == (void *)1) {
	fprintf(dst, "if (++count > max)\n");
	print_indent();
	fprintf(dst, "    max = count;\n");
    } else
	fprintf(dst, "--count;\n");
}

static void scan_program(PDG *pdg, const char *inputsize,
	int maxdim, struct clast_stmt *stmt, CloogOptions *options)
{
    mem_counter mc;
    mc.print_full(pdg, inputsize, maxdim, stmt);
}
