#include <vector>
#include <isl/set.h>
#include <isl/map.h>
#include <isl/constraint.h>
#include <isl/polynomial.h>
#include <barvinok/isl.h>
#include <isa/yaml.h>
#include <isa/pdg.h>
#include "da.h"
#include "translation.h"
#include "mem_options.h"

using std::vector;
using pdg::PDG;
using namespace da;
using std::cout;
using std::cerr;
using std::endl;

/* Compute the number of array elements that are read but not
 * (previously) written.  These are assumed to be the input.
 */
static __isl_give isl_pw_qpolynomial *compute_input_size(PDG *pdg,
	struct options *options)
{
	unsigned nparam = pdg->params.size();
	isl_space *dim;
	isl_pw_qpolynomial *inputsize;

	dim = isl_space_set_alloc(pdg->get_isl_ctx(), nparam, 0);
	isl_dim_set_parameter_names(dim, pdg->params);
	inputsize = isl_pw_qpolynomial_zero(dim);

	if (!options->with_input)
		return inputsize;

	for (int i = 0; i < pdg->dependences.size(); ++i) {
		isl_map *map;
		isl_pw_qpolynomial *size;
		pdg::dependence *dep = pdg->dependences[i];

		if (dep->type != pdg::dependence::uninitialized)
			continue;

		map = dep->relation->get_isl_map(pdg->get_isl_ctx());
		size = isl_map_card(map);
		inputsize = isl_pw_qpolynomial_add(inputsize, size);
	}

	return inputsize;
}

/* Compute the number of elements in domain for a fixed value
 * of the "fixed_loops" outer loop iterators and return this value
 * as a function of these loop iterators and the parameters.
 */
static __isl_give isl_pw_qpolynomial *compute_accesses_at(
	__isl_take isl_set *domain, int fixed_loops)
{
	isl_map *map;
	map = isl_map_from_range(domain);
	map = isl_map_move_dims(map, isl_dim_in, 0, isl_dim_out, 0, fixed_loops);
	return isl_map_card(map);
}

static __isl_give isl_map *less_than(__isl_take isl_space *dim,
	enum isl_dim_type type1, unsigned pos1,
	enum isl_dim_type type2, unsigned pos2)
{
	isl_int v;
	struct isl_constraint *c;
	struct isl_basic_map *bmap;

	isl_int_init(v);

	bmap = isl_basic_map_universe(isl_space_copy(dim));
	c = isl_inequality_alloc(isl_local_space_from_space(dim));
	isl_int_set_si(v, 1);
	isl_constraint_set_coefficient(c, type2, pos2, v);
	isl_int_set_si(v, -1);
	isl_constraint_set_coefficient(c, type1, pos1, v);
	isl_constraint_set_constant(c, v);
	bmap = isl_basic_map_add_constraint(bmap, c);

	isl_int_clear(v);

	return isl_map_from_basic_map(bmap);
}

/* Compute the number of elements in domain for a fixed value
 * of the "level" outer loop iterators and for all values of the
 * loop iterator at level "level" smaller than a fixed value
 * and return this value as a function of these (level+1) loop iterators
 * and the parameters.
 */
static __isl_give isl_pw_qpolynomial *compute_accesses_before(
	__isl_take isl_set *domain, int level)
{
	isl_map *map;
	map = isl_map_from_range(domain);
	map = isl_map_move_dims(map, isl_dim_in, 0, isl_dim_out, 0, level);
	map = isl_map_add_dims(map, isl_dim_in, 1);
	map = isl_map_intersect(map,
		less_than(isl_map_get_space(map), isl_dim_out, 0,
						isl_dim_in, level));
	return isl_map_card(map);
}

static int cmp_prefix_stat_dims(pdg::node *node, int *stat_dims, int d)
{
    for (int j = 0; j < d; ++j)
	if (node->prefix[2*j] != stat_dims[j])
	    return node->prefix[2*j] - stat_dims[j];

    return 0;
}

/*
 * Returns true if dependence is independent of first n_loops loops.
 */
static int is_loop_independent(isl_ctx *ctx,
	pdg::IslMap *relation, int n_loops)
{
	isl_map *rel;
	isl_map *test;
	assert(relation->input() >= n_loops);
	assert(relation->output() >= n_loops);
	int indep;

	rel = relation->get_isl_map(ctx);
	test = isl_map_universe(isl_map_get_space(rel));
	for (int i = 0; i < n_loops; ++i)
		test = isl_map_equate(test, isl_dim_in, i, isl_dim_out, i);
	indep = isl_map_is_subset(rel, test);
	isl_map_free(rel);
	isl_map_free(test);

	return indep;
}

struct contribution {
    unsigned	is_read : 1;
    unsigned	is_written : 1;
};

/*
 * Determines the net contribution of "dep" within statements
 * with statement level dimensions prefixed by stat_dims
 * and within a fixed iteration of the first n_loop loops.
 */
static struct contribution net_contribution(isl_ctx *ctx, pdg::dependence *dep,
					    int *stat_dims, int d, int n_loop,
					    struct options *options)
{
    struct contribution c;
    c.is_read = 0;
    c.is_written = 0;

    if (dep->type == pdg::dependence::uninitialized) {
	if (options->with_input &&
		cmp_prefix_stat_dims(dep->to, stat_dims, d) == 0)
	    c.is_read = 1;
    } else {
	int is_written = cmp_prefix_stat_dims(dep->from, stat_dims, d) == 0;
	int is_read = cmp_prefix_stat_dims(dep->to, stat_dims, d) == 0;
	int is_local = is_written && is_read &&
				is_loop_independent(ctx, dep->relation, n_loop);
	if (!is_local) {
	    c.is_read = is_read;
	    c.is_written = is_written;
	}
    }

    return c;
}

struct access {
    isl_set	*read;
    isl_set	*write;
};

/*
 * Compute increment in statement/loop identified by stat_dims
 */
static __isl_give isl_pw_qpolynomial *compute_live_increment(PDG *pdg,
	int *stat_dims, int d, struct access *accesses, struct options *options)
{
    unsigned nparam = pdg->params.size();
    isl_space *dim;
    isl_pw_qpolynomial *sum;

    dim = isl_space_set_alloc(pdg->get_isl_ctx(), nparam, d - 1);
    isl_dim_set_parameter_names(dim, pdg->params);
    sum = isl_pw_qpolynomial_zero(dim);
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	struct contribution c;

	c = net_contribution(pdg->get_isl_ctx(),
				dep, stat_dims, d, d-1, options);

	if (c.is_written) {
	    isl_pw_qpolynomial *inc;
	    inc = compute_accesses_at(isl_set_copy(accesses[i].write), d - 1);
	    sum = isl_pw_qpolynomial_add(sum, inc);
	}
	if (c.is_read) {
	    isl_pw_qpolynomial *dec;
	    dec = compute_accesses_at(isl_set_copy(accesses[i].read), d - 1);
	    sum = isl_pw_qpolynomial_sub(sum, dec);
	}
    }
    return sum;
}

/*
 * Compute increment in previous iterations of loop at given level.
 */
static __isl_give isl_pw_qpolynomial *compute_live_increment_in_loop(PDG *pdg,
	int *stat_dims, int level, struct access *accesses,
	struct options *options)
{
    unsigned nparam = pdg->params.size();
    isl_space *dim;
    isl_pw_qpolynomial *sum;

    dim = isl_space_set_alloc(pdg->get_isl_ctx(), nparam, level + 1);
    isl_dim_set_parameter_names(dim, pdg->params);
    sum = isl_pw_qpolynomial_zero(dim);
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	struct contribution c;

	c = net_contribution(pdg->get_isl_ctx(),
				dep, stat_dims, level+1, level+1, options);

	if (c.is_written) {
	    isl_pw_qpolynomial *inc;
	    inc = compute_accesses_before(isl_set_copy(accesses[i].write), level);
	    sum = isl_pw_qpolynomial_add(sum, inc);
	}
	if (c.is_read) {
	    isl_pw_qpolynomial *dec;
	    dec = compute_accesses_before(isl_set_copy(accesses[i].read), level);
	    sum = isl_pw_qpolynomial_sub(sum, dec);
	}
    }
    return sum;
}

struct tree;

struct tree {
    int d;
    int *stat_dims;
    vector<struct tree*> children;
    isl_set	 	*domain;
};

/*
 * Compute bound on number of live array elements in the
 * given (sub)tree.
 *
 * live_before: number of live elements before the loop/statement
 *		in terms of the outer loop iterators
 */
static __isl_give isl_pw_qpolynomial_fold *compute_bound(PDG *pdg,
					__isl_take isl_pw_qpolynomial *live,
					struct tree *tree,
					struct access *accesses,
					struct options *options)
{
    /* If we are in a leaf, we actually compute the bound. */
    if (tree->children.size() == 0) {
	isl_set *domain;
	isl_pw_qpolynomial_fold *bound;

	domain = isl_set_copy(tree->domain);
	live = isl_pw_qpolynomial_intersect_domain(live, domain);
	bound = isl_pw_qpolynomial_bound(live, isl_fold_max, NULL);

	return bound;
    }

    isl_pw_qpolynomial_fold *pwf;

    /* If we are not in the outer dummy loop, add contribution of
     * previous iterations.
     */
    if (tree->d > 0) {
	isl_space *dim;
	isl_pw_qpolynomial *inc;

	live = isl_pw_qpolynomial_add_dims(live, isl_dim_set, 1);
	inc = compute_live_increment_in_loop(pdg, tree->stat_dims, tree->d-1,
					    accesses, options);
	live = isl_pw_qpolynomial_add(live, inc);
    }

    pwf = compute_bound(pdg, isl_pw_qpolynomial_copy(live),
			tree->children[0], accesses, options);
    for (int i = 1; i < tree->children.size(); ++i) {
	isl_pw_qpolynomial_fold *pwf_i;
	isl_pw_qpolynomial *inc;
	struct tree *child = tree->children[i-1];
	inc = compute_live_increment(pdg, child->stat_dims, child->d,
				    accesses, options);
	live = isl_pw_qpolynomial_add(live, inc);
	pwf_i = compute_bound(pdg, isl_pw_qpolynomial_copy(live),
				tree->children[i], accesses, options);
	pwf = isl_pw_qpolynomial_fold_fold(pwf, pwf_i);
    }
    isl_pw_qpolynomial_free(live);
    return pwf;
}

/*
 * Construct a tree for all statements whose statement dimensions
 * start with the d first elements of stat_dims.
 * The first one of those statemens is statement n.
 */
static struct tree *compute_tree(PDG *pdg, int *stat_dims, int d, int n)
{
    struct tree *tree = new struct tree;
    tree->domain = NULL;
    tree->d = d;
    tree->stat_dims = new int[d];
    for (int i = 0; i < d; ++i)
	tree->stat_dims[i] = stat_dims[i];

    if (pdg->nodes[n]->prefix.size() == 2*d-1) {
	tree->domain = pdg->nodes[n]->source->get_isl_set(pdg->get_isl_ctx());
	return tree;
    }

    for (int i = n; i < pdg->nodes.size(); ++i) {
	if (cmp_prefix_stat_dims(pdg->nodes[i], stat_dims, d) > 0)
	    break;

	stat_dims[d] = pdg->nodes[i]->prefix[2*d];

	struct tree *child = compute_tree(pdg, stat_dims, d+1, i);
	tree->children.push_back(child);

	/* Skip over nodes with the same prefix at current level */
	for (; i+1 < pdg->nodes.size(); ++i)
	    if (cmp_prefix_stat_dims(pdg->nodes[i+1], stat_dims, d+1) != 0)
		break;
    }
    return tree;
}

static struct tree *compute_tree(PDG *pdg)
{
    int n_stat_dims = normalize_prefix(pdg);
    int stat_dims[n_stat_dims];

    return compute_tree(pdg, stat_dims, 0, 0);
}

static void free_tree(struct tree *tree)
{
    for (int i = 0; i < tree->children.size(); ++i)
	free_tree(tree->children[i]);
    delete [] tree->stat_dims;
    isl_set_free(tree->domain);
    delete tree;
}

static void compute_bound(PDG *pdg, struct options *options)
{
    struct tree *tree = compute_tree(pdg);

    struct access accesses[pdg->dependences.size()];
    for (int i = 0; i < pdg->dependences.size(); ++i) {
	pdg::dependence *dep = pdg->dependences[i];
	isl_map *rel = dep->relation->get_isl_map(pdg->get_isl_ctx());

	accesses[i].write = isl_map_domain(isl_map_copy(rel));
	accesses[i].read = isl_map_range(rel);
    }

    isl_pw_qpolynomial *inputsize;
    inputsize = compute_input_size(pdg, options);

    isl_pw_qpolynomial_fold *pwf;
    pwf = compute_bound(pdg, inputsize, tree, accesses, options);
    free_tree(tree);
    isl_set *context = pdg->get_context_isl_set();
    pwf = isl_pw_qpolynomial_fold_gist(pwf, context);

    isl_printer *printer;
    printer = isl_printer_to_file(pdg->get_isl_ctx(), stdout);
    printer = isl_printer_set_output_format(printer, ISL_FORMAT_C);
    printer = isl_printer_print_pw_qpolynomial_fold(printer, pwf);
    printer = isl_printer_end_line(printer);
    isl_printer_free(printer);

    isl_pw_qpolynomial_fold_free(pwf);

    for (int i = 0; i < pdg->dependences.size(); ++i) {
	isl_set_free(accesses[i].write);
	isl_set_free(accesses[i].read);
    }
}

int main(int argc, char * argv[])
{
    PDG *pdg;
    isl_ctx *ctx = isl_ctx_alloc();
    struct options *options = options_new_with_defaults();

    argc = options_parse(options, argc, argv, ISL_ARG_ALL);


    pdg = PDG::Load(stdin, ctx);
    int nparam = pdg->params.size();

    for (int i = 0; i < pdg->arrays.size(); ++i)
	find_deps(pdg, pdg->arrays[i], data_reuse);

    compute_bound(pdg, options);

    pdg->free();
    delete pdg;
    options_free(options);
    isl_ctx_free(ctx);

    return 0;
}
