#include <sstream>
#include <stdlib.h>
#include <unistd.h>
#include <gmp.h>
#include <isl/set.h>
#include <isl/map.h>
#include <barvinok/isl.h>
#include "size.h"
#include "size_bernstein.h"
#include "size_cloog.h"
#include "size_options.h"
#include "config.h"

namespace size {

using std::cerr;
using std::endl;
using std::ostream;

#ifndef USE_SIZE_CLOOG
int selfloop_size_cloog(__isl_take isl_set *write, __isl_take isl_set *read)
{
	isl_set_free(write);
	isl_set_free(read);
	return -1;
}
#endif

static bool is_rectangular_dependence(__isl_keep isl_map *dep)
{
	int ok;
	isl_set *dom;

	ok = isl_map_is_translation(dep);
	assert(ok >= 0);
	if (!ok)
		return false;

	dom = isl_map_domain(isl_map_copy(dep));
	ok = isl_set_is_box(dom);
	isl_set_free(dom);
	assert(ok >= 0);

	return ok;
}

/* Compute the number of tokens in the FIFO at the reference point(s) ref.
 * That is, the number of dependences that have started but not finished
 * before any point in ref.  Note that the same read iteration may
 * depend on several write iterations, so we cannot simply count the
 * number of writes and reads and take the differences, as the total
 * number of reads may be smaller than the total number of writes.
 * Instead, we set up two mappings between the reference point
 * and the dependences, one where the write happens before the
 * reference point and one where the read happens before the reference
 * point.  We then take the set difference and then count the number
 * of elements in this difference.
 */
isl_pw_qpolynomial *selfloop_tokens_in_fifo(__isl_take isl_map *dep,
	__isl_take isl_set *ref)
{
	int d;
	isl_set *dep_set;
	isl_map *read_before_ref;
	isl_map *write_before_ref;

	dep = isl_map_make_disjoint(dep);
	ref = isl_set_make_disjoint(ref);

	d = isl_map_dim(dep, isl_dim_in);
	dep = isl_map_move_dims(dep, isl_dim_in, d, isl_dim_out, 0, d);
	dep_set = isl_map_domain(dep);

	read_before_ref = isl_map_lex_gt(isl_set_get_space(ref));
	read_before_ref = isl_map_intersect_domain(read_before_ref, ref);
	write_before_ref = isl_map_copy(read_before_ref);

	read_before_ref = isl_map_insert_dims(read_before_ref, isl_dim_out, 0, d);
	write_before_ref = isl_map_add_dims(write_before_ref, isl_dim_out, d);

	read_before_ref = isl_map_intersect_range(read_before_ref,
						isl_set_copy(dep_set));
	write_before_ref = isl_map_intersect_range(write_before_ref, dep_set);

	write_before_ref = isl_map_subtract(write_before_ref, read_before_ref);

	return isl_map_card(write_before_ref);
}

/* Compute the maximum of the difference in number of writes and number
 * of reads before any read.  The difference is equal to the number
 * of token in the FIFO and the maximum is the buffer size of the FIFO.
 * The procedure assumes that the dependence is non-parametric and
 * may actually iterate over all reads to compute the maximum.
 */
static int selfloop_size_full(__isl_take isl_map *dep,
	__isl_take isl_set *ref)
{
	isl_pw_qpolynomial *qp_diff;
	isl_qpolynomial *qp_max;
	isl_int n, d;
	int is_cst;
	int size;

	qp_diff = selfloop_tokens_in_fifo(dep, ref);

	qp_max = isl_pw_qpolynomial_max(qp_diff);

	isl_int_init(n);
	isl_int_init(d);
	is_cst = isl_qpolynomial_is_cst(qp_max, &n, &d);

	assert(is_cst);
	assert(isl_int_is_one(d));

	size = isl_int_get_si(n);

	isl_int_clear(n);
	isl_int_clear(d);

	isl_qpolynomial_free(qp_max);

	return size;
}

enumerator::operator pdg::enumerator *() const
{
    switch (type) {
    case fixed:
	return new pdg::enumerator(v);
    case qp:
    case fold: {
	isl_printer *printer;
	char *str;
	string *s;
	printer = isl_printer_to_str(pdg->get_isl_ctx());
	printer = isl_printer_set_output_format(printer, ISL_FORMAT_C);
	if (type == qp)
		printer = isl_printer_print_pw_qpolynomial(printer, pwqp);
	else
		printer = isl_printer_print_pw_qpolynomial_fold(printer, pwf);
	str = isl_printer_get_str(printer);
	s = new string(str);
	isl_printer_free(printer);
	free(str);
	return new pdg::enumerator(s);
    }
    default:
	assert(0);
    }
}

integer *enumerator::evaluate() const
{
    switch (type) {
    case fixed:
	return new integer(v);
    case qp:
    case fold: {
	isl_space *dim;
	isl_point *pnt;
	isl_qpolynomial *qp;
	isl_int n, d;
	int is_cst;
	int v;

	isl_int_init(n);
	isl_int_init(d);
	dim = isl_space_params_alloc(pdg->get_isl_ctx(), pdg->params.size());
	isl_dim_set_parameter_names(dim, pdg->params);
	pnt = isl_point_zero(dim);
	for (int i = 0; i < pdg->params.size(); ++i) {
		isl_int_set_si(n, pdg->params[i]->value->v);
		pnt = isl_point_set_coordinate(pnt, isl_dim_param, i, n);
	}
	if (type == enumerator::qp)
		qp = isl_pw_qpolynomial_eval(
			isl_pw_qpolynomial_copy(pwqp), pnt);
	else
		qp = isl_pw_qpolynomial_fold_eval(
			isl_pw_qpolynomial_fold_copy(pwf), pnt);
	is_cst = isl_qpolynomial_is_cst(qp, &n, &d);
	assert(is_cst);
	isl_int_fdiv_q(n, n, d);
	v = isl_int_get_si(n);
	isl_qpolynomial_free(qp);
	isl_int_clear(n);
	isl_int_clear(d);
	return new integer(v);
    }
    default:
	assert(0);
    }
    return NULL;
}

size::enumerator *selfloop_size(pdg::PDG *pdg, __isl_take isl_map *dep,
			       size_options *options)
{
    int nparam = isl_map_dim(dep, isl_dim_param);
    isl_set *read;

    assert(isl_map_dim(dep, isl_dim_in) == isl_map_dim(dep, isl_dim_out));

    read = isl_map_range(isl_map_copy(dep));

    bool rect = !nparam && is_rectangular_dependence(dep);

    size::enumerator *e = NULL;
    if (!nparam && options->flags != SIZE_BERNSTEIN) {
	int maxsize;
	if (options->flags == SIZE_SIMULATE) {
	    isl_set *write = isl_map_domain(dep);
	    maxsize = selfloop_size_cloog(write, read);
	} else if (!rect)
	    maxsize = selfloop_size_full(dep, read);
	else {
	    isl_set *ref = isl_set_lexmin(read);
	    maxsize = selfloop_size_full(dep, ref);
	}
	e = new enumerator(maxsize);
    } else {
	e = selfloop_size_bernstein(pdg, dep, read); 
    }

    return e;
}

}
