#include "suif_inc.h"
#include <stdlib.h>
#include <assert.h>
#include "domain.h"
#include "build.h"

#include <isa/yaml.h>
#include <isa/pdg.h>
#include <cloog/isl/cloog.h>

using pdg::PDG;

static CloogInput *read_cloog_input(isl_ctx *ctx, char *in);

#define ARRAY_SIZE(a) (sizeof(a)/sizeof(*a))

static CloogInput *cloog_input;
static CloogOptions *options;
static CloogState *state;

static int oldstyle;

static void do_proc(tree_proc * tp)
{
    static context_info *ctx = new context_info;
    proc_sym * psym = tp->proc();
    fill_in_access(tp);
    construct_domains(tp, 0, oldstyle, ctx);
    build(tp, cloog_input, options);
}

int
main(int argc, char * argv[])
{
    isl_ctx *ctx = isl_ctx_alloc();
    char *domains = NULL;

    state = cloog_state_malloc();
    options = cloog_options_malloc(state);
    options->sh = 1;
    cmd_line_option my_options[] = {
	{ CLO_STRING,   "-domains",	"", 	&domains },
	{ CLO_INT,	"-f",		"1", 	&options->f },
	{ CLO_INT,	"-strides",	"0", 	&options->strides },
	{ CLO_NOARG,  	"-oldstyle",	"", 	&oldstyle },
    };
    
    start_suif(argc, argv);
    init_dead_code();
    init_unused();
    parse_cmd_line(argc, argv, my_options, ARRAY_SIZE(my_options));
    init(argc, argv);

    cloog_input = read_cloog_input(ctx, domains);
    if (!cloog_input) {
	fprintf(stderr, "Empty input\n");
	return 1;
    }

    suif_proc_iter(argc, argv, do_proc, TRUE);

    if (domains)
	delete [] domains;

    cloog_state_free(state);
    isl_ctx_free(ctx);

    return 0;
}

static CloogInput *construct_cloog_input(CloogState *state,
	pdg::PDG *pdg, CloogOptions *options)
{
	int nscat = 0;
	int niter = 0;
	int nparam = pdg->params.size();
	isl_ctx *ctx = pdg->get_isl_ctx();
	CloogDomain *context, *domain;
	CloogScattering *scat;
	CloogUnionDomain *ud;

	context = cloog_domain_from_isl_set(pdg->get_context_isl_set());

	ud = cloog_union_domain_alloc(nparam);

	for (int i = 0; i < pdg->nodes.size(); ++i) {
		isl_set *source;
		isl_map *map;

		if (!pdg->nodes[i]->schedule) {
			fprintf(stderr, "Missing schedule\n");
			exit(-1);
		}

		source = pdg->nodes[i]->source->get_isl_set(ctx);
		domain = cloog_domain_from_isl_set(source);

		map = pdg->nodes[i]->schedule->get_isl_map(ctx);
		scat = cloog_scattering_from_isl_map(map);

		if (!nscat)
			nscat = cloog_scattering_dimension(scat, domain);

		ud = cloog_union_domain_add_domain(ud, NULL, domain, scat,
							pdg->nodes[i]);
	}

	for (int i = 0; i < nparam; ++i) {
		const char *s = strdup(pdg->params[i]->name->s.c_str());
		ud = cloog_union_domain_set_name(ud, CLOOG_PARAM, i, s);
	}
	for (int i = 0; i < nscat; ++i) {
		char name[MAX_NAME];
		sprintf(name, "c%d", i);
		ud = cloog_union_domain_set_name(ud, CLOOG_SCAT, i, name);
	}

	return cloog_input_alloc(context, ud);
}

static CloogInput *read_cloog_input(isl_ctx *ctx, char *in)
{
    FILE *input = fopen(in, "r");
    assert(input);

    PDG *pdg = PDG::Load(input, ctx);
    fclose(input);
    if (pdg == NULL)
	return NULL;

    return construct_cloog_input(state, pdg, options);
}
