#include "srel.h"
#include "domain.h"

#include <vector>
#include <algorithm>

typedef std::vector<Variable_ID> varvector;

SRelation::SRelation (int in, int out, std::vector< std::vector<int> > & c, 
    symtab_ann *a) : Relation(in, out), sa(a)
{
    assert(c.size() >= 1);
    int	dim = c[0].size();
    varvector vars;

    for (int j = 1; j <= n_inp(); ++j)
	vars.push_back(input_var(j));
    for (int j = 1; j <= n_out(); ++j)
	vars.push_back(output_var(j));
    assert(in + out + 2 == dim);

    F_And *base = add_and();
    for (int i = 0; i < c.size(); ++i) {
	Constraint_Handle h;
	if (c[i][0]) 
	    h = base->add_GEQ();
	else
	    h = base->add_EQ();
	for (int j = 1; j < dim-1; ++j)
	    h.update_coef(vars[j-1], c[i][j]);
	h.update_const(c[i][dim-1]);
    }
    finalize();
}

void SRelation::fix(var_sym *var, int val)
{
    sa->add_var(var);
    EQ_Handle e = and_with_EQ();
    e.update_coef(get_local(sa->vars[var]), -1);
    e.update_const(val);
}

/* Add constraint "var >= 0" */
void SRelation::set_infinite(var_sym *var)
{
    sa->add_var(var);
    GEQ_Handle e = and_with_GEQ();
    e.update_coef(get_local(sa->vars[var]), 1);
}

void SRelation::intersect(named_lin_ineq *nlq)
{
    for (int i = 1; i < nlq->n(); ++i)
	sa->add_var(nlq->names()[i].var());

    integer_matrix lq = nlq->ineqs();
    for (int i = 0; i < lq.m(); ++i) {
	GEQ_Handle g = and_with_GEQ();
	g.update_const(lq[i][0]);
	for (int j = 1; j < lq.n(); ++j) {
	    g.update_coef(get_local(sa->vars[nlq->names()[j].var()]), lq[i][j]);
	}
    }
    delete nlq;
}

static void print_constraint(Constraint_Handle c, varvector& vv, int geq)
{
    printf(" %6d", geq);
    for (int i = 0; i < vv.size(); ++i)
	printf(" " coef_fmt, c.get_coef(vv[i]));
    printf(" " coef_fmt, c.get_const());
    printf("\n");
}

static bool equal(operand op1, operand op2)
{
    if (op1.kind() != op2.kind())
	return false;
    switch (op1.kind()) {
    case OPER_INSTR: {
	if (op1.instr()->opcode() != op2.instr()->opcode())
	    return false;
	if (op1.instr()->opcode() == io_ldc) {
	    in_ldc *in1 = dynamic_cast<in_ldc *>(op1.instr());
	    in_ldc *in2 = dynamic_cast<in_ldc *>(op2.instr());
	    if (in1->value() != in2->value())
		return false;
	}
	for (unsigned n = 0; n < op1.instr()->num_srcs(); n++) {
	    if (!equal(op1.instr()->src_op(n), op2.instr()->src_op(n)))
		return false;
	}
	break;
    }
    case OPER_SYM:
	return op1.symbol() == op2.symbol();
    case OPER_NULL:
	return true;
    default:
	return false;
    }
    return true;
}

/* Set nested accesses of d to the union of those in a and b,
 * unifying accesses that have identical operations.
 */
static SRelation *set_nested(SRelation *d, SRelation *a, SRelation *b)
{
	for (int i = 0; i < a->nested.size(); ++i)
		d->nested.push_back(a->nested[i]);
	for (int i = 0; i < b->nested.size(); ++i) {
		int j;
		for (j = 0; j < d->nested.size(); ++j) {
			if (b->nested[i] == d->nested[j])
				break;
			if (b->nested[i]->op.is_null())
				continue;
			if (d->nested[j]->op.is_null())
				continue;
			if (equal(b->nested[i]->op, d->nested[j]->op)) {
				Relation r = copy(*(Relation*)d);
				EQ_Handle e = r.and_with_EQ();
				e.update_coef(r.get_local(d->sa->vars[b->nested[i]->var]), -1);
				e.update_coef(r.get_local(d->sa->vars[d->nested[j]->var]), 1);
				r = Project(r, d->sa->vars[b->nested[i]->var]);
				SRelation *p = new SRelation(r, d->sa, d->nested);
				delete d;
				d = p;
				break;
			}
		}
		if (j == d->nested.size())
			d->nested.push_back(b->nested[i]);
	}

	return d;
}

SRelation * SRelation::Intersection(SRelation *b)
{
    Relation I = ::Intersection(copy(*(Relation*) this), copy(*(Relation*) b));
    SRelation *c = new SRelation(I, sa);
    c = set_nested(c, this, b);
    return c;
}

SRelation * SRelation::Union(SRelation *b)
{
    Relation I = ::Union(copy(*(Relation*) this), copy(*(Relation*) b));
    SRelation *u = new SRelation(I, sa);
    u = set_nested(u, this, b);
    return u;
}

SRelation * SRelation::Difference(SRelation *b)
{
    Relation I = ::Difference(copy(*(Relation*) this), copy(*(Relation*) b));
    SRelation *d = new SRelation(I, sa);
    d = set_nested(d, this, b);
    return d;
}

SRelation * SRelation::Composition(SRelation *b)
{
    Relation I = ::Composition(copy(*(Relation*) this), copy(*(Relation*) b));
    return new SRelation(I, sa);
}

SRelation * SRelation::Complement()
{
    Relation I = ::Complement(copy(*(Relation*) this));
    return new SRelation(I, sa, nested);
}
